/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Keeps track of articles which are unavailable, in case they become available
 * later */


#include "backfill.h"
#include "common.h"
#include "today.h"

/* The full name of a backfill file for a given server (on heap) */
static char *backfill_get_filename(const char *server_id)
{
     return XSTRDUPCAT(RC_DIR "/backfill.", server_id);
}

static BackFillArticle *backfill_add_article(BackFillGroup *bfgroup,
	long int artnum, long int day, BOOL just_added, BOOL no_xover)
{
    BackFillArticle bfart;

    bfart.artnum = artnum;
    bfart.day = day;
    bfart.just_added = just_added;
    bfart.no_xover = no_xover;

    return &bfgroup->vec_data[VECTORP_APPEND(bfgroup, bfart) - 1];
}

BackFillServer *backfill_load(const char *server_id)
{
    char buffer[200];
    BackFillServer *bfserv = XMALLOC(sizeof(BackFillServer));
    char *filename = backfill_get_filename(server_id);
    FILE *fp = fopen(filename, "r");
    BOOL no_xover = 2;

    bfserv->groups = NULL;
    bfserv->server_id = XSTRDUP(server_id);
    free(filename);
    if (!fp)
    {
	deblog_warning(
  _("Can't read backfill file for %s, probably hasn't been created yet: %s\n"),
	    server_id, strerror(errno));
	return bfserv;
    }

    /* For each group name */
    while (fscanf(fp, "%s", buffer) == 1)
    {
	BackFillGroup *bfgroup;
	char *group_name = XSTRDUP(buffer);

	/* Now we should find a '{' */
	if (fscanf(fp, "%s", buffer) != 1 || buffer[0] != '{' || buffer[1])
	{
	    deblog_critical(_("Missing/broken '{' in %s's backfill file\n"),
		    server_id);
	    fclose(fp);
	    return bfserv;
	}

	deblog_debug("Loading backfill articles for %s on %s\n",
		group_name, server_id);
	bfgroup = backfill_add_group(bfserv, group_name);
	free(group_name);

	/* Should have list of artnum+day pairs followed by '}' */
	while (fgets(buffer, sizeof(buffer), fp) && buffer[0] != '}')
	{
	    char *endptr;
	    char *midstr;
	    long int artnum = strtol(buffer, &endptr, 10);
	    long int day = today;

	    /* May need to skip \n after { */
	    if (buffer[0] == '\n')
		continue;
	    deblog_debug("backfill line: %s\n", buffer);
	    if (!buffer[0] || endptr == buffer || !isspace(*endptr))
	    {
		deblog_critical(
			_("Broken article number in %s's backfill file\n"),
			server_id);
		deblog_critical(
			"buffer[0] = %d, endptr %p, buffer %p, *endptr %d\n",
			buffer[0], endptr, buffer, *endptr);
		fclose(fp);
		return bfserv;
	    }

	    midstr = endptr;
	    /*while (isspace(*++midstr));*/
	    day = strtol(midstr, &endptr, 10);
	    if (midstr == endptr || 
		    (!isspace(*endptr) && *endptr != '\n' && *endptr))
	    {
		deblog_critical(
			_("Broken day code in %s's backfill file\n"),
			server_id);
		fclose(fp);
		return bfserv;
	    }

	    midstr = endptr;
	    /*while (isspace(*++midstr));*/
	    no_xover = strtol(midstr, &endptr, 10);
	    if (midstr == endptr ||
		    (!isspace(*endptr) && *endptr != '\n' && *endptr))
	    {
		deblog_debug(
			_("Broken/missing xover code in %s's backfill file\n"),
			server_id);
		no_xover = 2;
	    }

	    deblog_debug("Article %ld day %ld no_xover %d\n", artnum, day,
		    no_xover);
	    backfill_add_article(bfgroup, artnum, day, FALSE, no_xover);
	}
    }
    fclose(fp);
    return bfserv;
}

/* Saves above list(s) */
void backfill_save(const BackFillServer *bfserv, long int expiry)
{
    char *filename = backfill_get_filename(bfserv->server_id);
    FILE *fp = fopen(filename, "w");
    BackFillGroup *bfgroup;

    free(filename);
    if (!fp)
    {
	deblog_warning(_("Can't write backfill file for %s: %s\n"),
	    bfserv->server_id, strerror(errno));
	return;
    }
    for (bfgroup = bfserv->groups; bfgroup; bfgroup = bfgroup->next)
    {
	size_t i;

	if (!bfgroup->vec_used_members)
	    continue;
	fprintf(fp, "%s {\n", bfgroup->group_name);
	for (i = 0; i < bfgroup->vec_used_members; ++i)
	{
	    BackFillArticle *bfart = &bfgroup->vec_data[i];
	    
	    if (bfart->artnum >= 0 && today - bfart->day < expiry)
	    {
		fprintf(fp, "  %ld %ld %d\n", bfart->artnum, bfart->day,
			bfart->no_xover);
	    }
	}
	fprintf(fp, "}\n");
    }
    fclose(fp);
}

BackFillGroup *backfill_get_group(BackFillServer *bfserv, const char
	*group_name)
{
    BackFillGroup *bfgroup;

    for (bfgroup = bfserv->groups; bfgroup; bfgroup = bfgroup->next)
    {
	if (!strcmp(bfgroup->group_name, group_name))
	    return bfgroup;
    }
    return NULL;
}

BackFillGroup *backfill_add_group(BackFillServer *bfserv,
	const char *group_name)
{
    BackFillGroup *bfgroup = XMALLOC(sizeof(BackFillGroup));

    bfgroup->next = bfserv->groups;
    bfserv->groups = bfgroup;
    bfgroup->num_arts = 0;
    bfgroup->group_name = XSTRDUP(group_name);
    VECTORP_INIT(bfgroup, BackFillArticle);
    bfgroup->current = 0;

    return bfgroup;
}

BOOL backfill_add_unavailable(BackFillGroup *bfgroup, long int artnum,
	BOOL no_xover)
{
    size_t i;

    for (i = 0; i < bfgroup->vec_used_members; ++i)
    {
	BackFillArticle *bfart = &bfgroup->vec_data[i];

	if (bfart->artnum == artnum)
	{
	    return TRUE;
	}
    }
    backfill_add_article(bfgroup, artnum, today, TRUE, no_xover);
    return FALSE;
}

/* Removes a previously unavailable article from the list when it has become
 * available */
void backfill_remove_available(BackFillGroup *bfgroup, long int artnum)
{
    size_t i;

    for (i = 0; i < bfgroup->vec_used_members; ++i)
    {
	BackFillArticle *bfart = &bfgroup->vec_data[i];

	if (bfart->artnum == artnum)
	{
	    deblog_debug(_("Backfill %p/%d became available: no_xover %d\n"),
			bfgroup, bfart->artnum, bfart->no_xover);
	    bfart->artnum = -1;
	    return;
	}
    }
}

void backfill_remove_old(BackFillGroup *bfgroup, long int artnum)
{
    size_t i;

    for (i = 0; i < bfgroup->vec_used_members; ++i)
    {
	BackFillArticle *bfart = &bfgroup->vec_data[i];

	if (bfart->artnum < artnum)
	{
	    deblog_debug(_("Backfill %p/%d obsolete\n"),
		    bfgroup, bfart->artnum);
	    bfart->artnum = -1;
	}
    }
}

