/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"
#include <assert.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#if USE_PTHREADS
#include <pthread.h>
#include "main.h"
#endif

#if ENABLE_ICONV
#include <iconv.h>
#include "vector.h"
#endif

#include "child.h"
#include "nntp.h"

#if !USE_PTHREADS
static BOOL child_shouldnt_exit = TRUE;
static ChildData current_child_data = CHILD_NULL;

static void child_exit(void)
{
	if (child_shouldnt_exit) 
	{
		fprintf(stderr,
			_("<%d> Warning: Child process exiting unexpectedly\n"),
			current_child_data.thread_idx);
	}
}
#else /* USE_PTHREADS */

static int child_nthreads = 0;
ChildData **child_data_table = NULL;
pthread_t *child_thread_table = NULL;

#endif /* USE_PTHREADS */

void child_send_command(ChildData *child_data, 
		intercom_Command code, const char *data, ...)
{
	va_list ap;

	va_start(ap, data);
	intercom_send_vcommand(child_data->out_q, code, child_data->thread_idx,
			data, ap);
	va_end(ap);
}

static void child_shutdown(ChildData *child_data, BOOL error)
{
	/* Try to let server know we're leaving as gracefully as possible, but
	 * don't hang around for response or anything if there's an error */
	deblog_info(_("Thread shutting down %s\n"), error ? _("due to error") :
			_("naturally"));
	if (child_data->sbuf)
	{
		if (!child_data->sent_quit)
		{
			child_data->sent_quit = TRUE;
			rw_write_line(child_data->sbuf, "QUIT");
		}
		if (!error)
		{
			nntp_check_response(child_data->sbuf, child_data->server,
				_("Bad response to QUIT"), "2", deblog_Level_Critical);
		}
		/* Reset in case this thread services another server */
		child_data->sent_quit = FALSE;
		rw_close_and_free(child_data->sbuf);
		child_data->sbuf = NULL;
		deblog_info(_("Closed socket OK\n"));
	}
	if (child_data->intercom)
	{
		child_send_command(child_data,
				error ? command_QuietError : command_AllDone, "");
	}
	free(child_data->server);
	child_data->server = NULL;
#if !USE_PTHREADS
	child_shouldnt_exit = FALSE;
#endif /* !USE_PTHREADS */
}

#if !USE_PTHREADS
static void child_signal_handler(int sig)
{
	static BOOL recursive_exit = FALSE;
	
	if (!recursive_exit)
	{
		recursive_exit = TRUE;
		if (!current_child_data.sent_quit)
		{
			current_child_data.sent_quit = TRUE;
			if (current_child_data.sbuf)
				rw_write_line(current_child_data.sbuf, "QUIT");
		}
		fprintf(stderr, _("Child exiting due to signal %d\n"), sig);
		fflush(stderr);
		exit(1);
	}
	else
	{
		abort();
	}
}
#endif /* !USE_PTHREADS */

static int nibble_from_hex(char c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	else if (c >= 'A' && c <= 'F')
		return c - 'A' + 10;
	else if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	else
		return -1;
}

static size_t hex_to_addr(const char *hex, struct sockaddr **p_addr)
{
	char *data;
	size_t l, n;
	
	l = strlen(hex) / 2;
	data = XMALLOC(l);
	for (n = 0; n < l; ++n)
	{
		int byte = 16 * nibble_from_hex(hex[n * 2]) +
				nibble_from_hex(hex[n * 2 + 1]);
				
		if (byte < 0)
		{
			free(data);
			*p_addr = NULL;
			return 0;
		}
		data[n] = byte;
	}
	*p_addr = (struct sockaddr *) data;
	return l;
}

static void child_connect(ChildData *child_data, const char *comdata)
{
	char *path = XMALLOC(strlen(comdata));
	char *servid = XMALLOC(strlen(comdata));
	char *addr_s = XMALLOC(strlen(comdata));
	struct sockaddr *addr;
	size_t addr_len;
	int proto;
	BOOL ssl;

#if !USE_PTHREADS
	child_shouldnt_exit = TRUE;
#endif /* !USE_PTHREADS */
	child_data->pipeline_depth = 0;
	sscanf(comdata, "%s %s %d %s %d %d %d", servid, addr_s, &proto, path,
			&child_data->pipeline, &child_data->ignore_436, &ssl);
			
	deblog_info(_("Thread %d connecting to server %s\n"),
			child_data->thread_idx, servid);
	child_data->server = servid;
	free(child_data->path);
	if (strcmp(path, "!"))
	{
		child_data->path = path;
	}
	else
	{
		free(path);
		child_data->path = NULL;
	}
	
	addr_len = hex_to_addr(addr_s, &addr);
	free(addr_s);
	if (!addr_len)
	{
		deblog_critical(_("Intercom has corrupted socket address data\n"));
		child_shutdown(child_data, TRUE);
		return;
	}
	
	deblog_debug(_("Connecting to %s, pipelining to depth %d %s\n"), servid,
			PIPELINE_DEPTH,
			child_data->pipeline ? _("enabled") : _("disabled"));
	if (child_data->ignore_436)
	{
		deblog_debug(_("Ignoring 441 436 <Message-ID> POST rejections\n"));
	}
	child_data->sbuf = nntp_connect(child_data->server, addr,
			addr_len, proto, child_data->thread_idx, ssl);
	free(addr);
	if (!child_data->sbuf)
	{
		child_shutdown(child_data, TRUE);
		return;
	}
	child_send_command(child_data, command_Connected, "");
}

static void child_authenticate(ChildData *child_data, const char *usrnpwd)
{
	char *user, *passwd;

	assert(child_data->sbuf);
	user = XMALLOC(strlen(usrnpwd) + 1);
	passwd = XMALLOC(strlen(usrnpwd) + 1);
	sscanf(usrnpwd, "%s %s", user, passwd);
	if (nntp_authenticate(child_data->sbuf, child_data->server, user, passwd))
	{
		child_shutdown(child_data, TRUE);
	}
	else
	{
		child_send_command(child_data, command_Authenticated, "");
	}
	free(user);
	free(passwd);
}

static void child_request_article(ChildData *child_data)
{
	child_send_command(child_data, command_GetNextArticle, "%ld",
			child_data->rq_group_ix);
}

static void child_get_group_stats(ChildData *child_data, const char *data)
{
	char *groupname = XMALLOC(strlen(data) + 1);
	long int first, last;

	sscanf(data, "%ld %s", &child_data->rq_group_ix, groupname);
	deblog_debug(_("Entering group %ld %s to get stats\n"),
			child_data->rq_group_ix, groupname);
	if (nntp_enter_group(child_data->sbuf, child_data->server, groupname,
			&first, &last, TRUE))
	{
		child_send_command(child_data, command_BogusGroup, "%ld",
			child_data->rq_group_ix);
	}
	else
	{
		deblog_debug(_("%s: %ld-%ld\n"), groupname, first, last);
		child_send_command(child_data, command_GroupArticleStats,
				"%ld %ld %ld", child_data->rq_group_ix, first, last);
	}
	free(groupname);
}

static void child_issue_xhdr(ChildData *child_data,
		const char *data, BOOL xover)
{
	const char *line;
	long int group_ix;
	char *range = XMALLOC(strlen(data));

	sscanf(data, "%ld %s", &group_ix, range);
	rw_writef(child_data->sbuf, "%s %s\r\n",
			xover ? "XOVER" : "XHDR Message-ID", range);
	deblog_debug("Issuing %s %s\r\n",
			xover ? "XOVER" : "XHDR Message-ID", range);
	free(range);
	if (!(line = nntp_check_response(child_data->sbuf, child_data->server,
			_("Bad response to XHDR/XOVER"), "2", deblog_Level_Critical)))
	{
		child_shutdown(child_data, TRUE);
		return;
	}
	deblog_debug(_("Response to XOVER/XHDR command: %s\n"), line);
	while ((line = rw_read_line(child_data->sbuf)) != NULL && line[0] != '.')
	{
		deblog_debug(_("XOVER/XHDR line: %s\n"), line);
		child_send_command(child_data,
				xover ? command_XoverData : command_GroupContains,
				"%ld %s", group_ix, line);
	}
	child_send_command(child_data, command_GroupComplete, "%ld", group_ix);
}

static const char *child_path_filter(const char *line, void *data)
{
	ChildData *child_data = data;
	size_t len;

	if (child_data->path_done)
		return line;
	if (!line[0])
	{
		child_data->path_done = TRUE;
		return line;
	}
	if (strncmp(line, "Path: ", 6))
		return line;
	len = strlen(line) + strlen(child_data->path) + 2;
	child_data->path_filter_line = XREALLOC(child_data->path_filter_line, len);
	sprintf(child_data->path_filter_line,
			"Path: %s!%s", child_data->path, line + 6);
	child_data->path_done = TRUE;
	return child_data->path_filter_line;
}

static void child_shift_pipeline(ChildData *child_data)
{
	int i;

	for (i = 1; i < child_data->pipeline_depth; ++i)
	{
		child_data->group_ix[i - 1] = child_data->group_ix[i];
		child_data->artnum[i - 1] = child_data->artnum[i];
		child_data->change_group[i - 1] = child_data->change_group[i];
		child_data->backfill[i - 1] = child_data->backfill[i];
	}
	--child_data->pipeline_depth;
}

#if ENABLE_ICONV

static char *set_charset_case(const char *charset_in,
		int (*convertor)(int), BOOL strip_dashes)
{
	char *charset_out = XMALLOC(strlen(charset_in) + 1);
	int n, m;

	for (n = 0, m = 0; n < strlen(charset_in); ++n)
	{
		char c = charset_in[n];

		if ((!strip_dashes || (c != '-' && c != '_')) && c != '"')
			charset_out[m++] = convertor(c);
	}
	charset_out[m] = 0;
	return charset_out;
}

inline static char *lower_charset(const char *charset_in, BOOL strip_dashes)
{
	return set_charset_case(charset_in, tolower, strip_dashes);
}

inline static char *upper_charset(const char *charset_in, BOOL strip_dashes)
{
	return set_charset_case(charset_in, toupper, strip_dashes);
}

static int compare_charsets(const char *c1, const char *c2)
{
	char *c1a = lower_charset(c1, TRUE);
	char *c2a = lower_charset(c2, TRUE);
	int result = strcmp(c1a, c2a);

	free(c1a);
	free(c2a);
	return result;
}

/* Extends line_out and moves out_p, which is a pointer to within line_out,
 * if size of line_out isn't enough for l more chars after out_p */
static void check_extend_line_out(char **p_line_out, char **p_out_p,
		size_t *p_out_size, size_t l)
{
	long int shortage = l - (*p_out_size - (*p_out_p - *p_line_out));

	if (shortage > 0)
	{
		long int out_used = *p_out_p - *p_line_out;
		*p_out_size += shortage;
		*p_line_out = XREALLOC(*p_line_out, *p_out_size + 1);
		*p_out_p = *p_line_out + out_used;
	}
}

static char hex_digit(char hex, const char *msgid)
{
	if (isdigit(hex))
		return hex - '0';
	else if (hex >= 'A' && hex <= 'F')
		return hex - 'A' + 10;
	else if (hex >= 'a' && hex <= 'f')
		return hex - 'a' + 10;
	deblog_warning(_("Invalid hex digit '%c' in MIME sequence in %s\n"),
			hex, msgid);
	/* Weird return value results in valid ASCII (") if happens twice */
	return 2;
}

static iconv_t open_convertor(const char *to_charset, const char *from_charset)
{
	char *fc = upper_charset(from_charset, FALSE);
	char *tc = upper_charset(to_charset, FALSE);
	char *iconv_from = XSTRDUPCAT(fc, "//TRANSLIT");
	iconv_t cd = iconv_open(tc, iconv_from);

	free(iconv_from);
	if (cd == (iconv_t) -1)
	{
		static BOOL warned_translit = FALSE;

		if (!warned_translit)
		{
			warned_translit = TRUE;
			deblog_info(_("//TRANSLIT extension to iconv_open failed, "
					"trying plain '%s'\n"), from_charset);
		}
		cd = iconv_open(tc, fc);
		if (cd == (iconv_t) -1)
		{
			deblog_warning(_("Unable to open convertor from %s to %s"),
				from_charset, to_charset);
		}
	}
	free(fc);
	free(tc);
	return cd;
}

static char *demime_q(const char *start, const char *end, const char *msgid)
{
	char *demimed = XMALLOC(end - start + 1);
	const char *in_p = start;
	char *out_p = demimed;

	while (in_p != end)
	{
		switch (*in_p)
		{
			case '=':
			if (end - in_p < 3)
			{
				deblog_warning(_("Invalid MIME = sequence in %s\n"), msgid);
				*out_p = 0;
				return demimed;
			}
			*(out_p++) = hex_digit(*(in_p + 1), msgid) * 16 +
					hex_digit(*(in_p + 2), msgid);
			in_p += 3;
			break;
			case '_':
			*(out_p++) = ' ';
			++in_p;
			break;
			default:
			*(out_p++) = *in_p++;
		}
	}
	*out_p = 0;
	return demimed;
}

static char *demime_b(const char *start, const char *end, const char *msgid)
{
	char *demimed;
	char *out_p;
	const char *in_p = start;
	int padding = 0;
	long int word;
	int shift = 18;

	if (end[-1] == '=')
	{
		++padding;
		if (end[-2] == '=')
			++padding;
	}
	/* Input should be multiple of 4, but round up in case it isn't */
	demimed = XMALLOC(((end - start + 3) / 4) * 3 + 1);
	out_p = demimed;
	if ((end - start) % 4)
	{
		deblog_warning(_("%s contains base64 sequence with length != 4\n"),
			msgid);
		/* Treats the discrepancy as surplus characters which will be ignored */
	}
	word = 0;
	while (in_p != end)
	{
		int i;

		i = *(in_p++);
		if (i >= 'A' && i <= 'Z')
			i -= 'A';
		else if (i >= 'a' && i <= 'z')
			i = i - 'a' + 26;
		else if (i >= '0' && i <= '9')
			i = i - '0' + 52;
		else if (i == '+')
			i = 62;
		else if (i == '/')
			i = 63;
		else if (i == '=')
			i = 0;
		else
		{
			deblog_warning(_("%s contains invalid character '%c' in base64\n"),
			msgid, i);
			i = 0;
		}
		word |= i << shift;
		if (shift)
		{
			shift -= 6;
		}
		else
		{
			int out_left;

			if (in_p == end)
			out_left = 3 - padding;
			else
			out_left = 3;
			for (shift = 16; out_left && (shift >= 0); shift -= 8)
			{
			*out_p++ = (word >> shift) & 0xff;
			--out_left;
			}
			shift = 18;
			word = 0;
		}
	}
	*out_p = 0;
	return demimed;
}

char *remime_q(const char *charset, const char *raw)
{
	/* If we use (signed) char for c, %x *printf conversion treats it as
	 * negative */
	unsigned char c;
	char *remimed = XMALLOC(8 + strlen(raw) * 3 + strlen(charset));
	const char *in_p = raw;
	char *out_p = remimed + sprintf(remimed, "=?%s?Q?", charset);

	while ((c = *(in_p++)) != 0)
	{
		if (strchr(" _?=", c) || (c & 0x80) || iscntrl(c))
			out_p += sprintf(out_p, "=%2.2x", c);
		else
			*(out_p++) = c;
	}
	strcpy(out_p, "?=");
	return remimed;
}

/* See comment at full definition below */
static int get_converted_line(iconv_t cd, const char *line_in,
		char **pline_out);

/* This returns the input string without copying if it doesn't contain MIME
 * so beware. */
static char *convert_mime_header(const char *to_charset, const char *line_in,
		const char *msgid)
{
	char *line_out;
	const char *in_p;
	char *out_p;
	const char *token;
	size_t out_size;
	BOOL added_conversion = FALSE;
	
	token = strstr(line_in, "=?");
	if (!token)
		return (char *) line_in;
	/* Remember result may be longer than input. out_size excludes terminator*/
	out_size = strlen(line_in);
	line_out = XMALLOC(out_size + 1);
	in_p = line_in;
	out_p = line_out;
	while (*in_p)
	{
		char *from_charset = NULL;
		const char *start_seq;
		char menc;
		size_t len_plain;

		if (token)
			len_plain = token - in_p;
		else
			len_plain = strlen(in_p);
		check_extend_line_out(&line_out, &out_p, &out_size, len_plain);
		strncpy(out_p, in_p, len_plain);
		out_p += len_plain;
		if (!token)
			break;
		in_p = token + 2;
		start_seq = token;
		token = strchr(in_p, '?');
		/* Got separator after encoding, isolate encoding */
		if (token)
		{
			from_charset = XMALLOC(token - in_p + 1);
			strncpy(from_charset, in_p, token - in_p);
			from_charset[token - in_p] = 0;
			in_p = token + 1;
			menc = toupper(*(in_p++));
			/* Point token to ?= at end of sequence */
			token = strstr(in_p + 1, "?=");
			if (*(in_p++) != '?' || (menc != 'Q' && menc != 'B'))
			token = NULL;
		}
		if (!token)
		{
			deblog_warning(_("%s contains invalid MIME sequences in header\n"),
				msgid);
			check_extend_line_out(&line_out, &out_p, &out_size, strlen(in_p));
			strcpy(out_p, start_seq);
			free(from_charset);
			return line_out;
		}
		if (compare_charsets(from_charset, to_charset))
		{
			iconv_t cd;
			char *raw_encoded = NULL;
			char *remimed = NULL;
			char *demimed = (menc == 'Q') ?
			demime_q(in_p, token, msgid) : demime_b(in_p, token, msgid);

			/* demimed is raw string in from_charset; convert to to_charset */
			cd = open_convertor(to_charset, from_charset);
			if (cd != (iconv_t) -1)
			{
				if (get_converted_line(cd, demimed, &raw_encoded) < 0)
				{
					deblog_warning(_("Conversion of MIME sequence in %s "
							"from %s to %s failed"), msgid,
							from_charset, to_charset);
				}
				iconv_close(cd);
			}
			free(demimed);
			if (raw_encoded)
			{
				size_t l;
	
				remimed = remime_q(to_charset, raw_encoded);
				l = strlen(remimed);
				check_extend_line_out(&line_out, &out_p, &out_size, l);
				strcpy(out_p, remimed);
				out_p += l;
				free(raw_encoded);
				free(remimed);
				added_conversion = TRUE;
			}
		}
		if (!added_conversion)
		{
			long int l = (token - start_seq) + 2;
			check_extend_line_out(&line_out, &out_p, &out_size, l);
			strncpy(out_p, start_seq, l);
			out_p += l;
		}
		free(from_charset);
		from_charset = NULL;
		/* Move in_p past what we just (hopefully converted and) added */
		in_p = token + 2;
		token = strstr(in_p, "=?");
	}
	*out_p = 0;
	return line_out;
}

/*  Returns -1 for error. */
static int get_converted_line(iconv_t cd, const char *line_in, char **pline_out)
{
	char *iconv_in = (char *) line_in;
	size_t in_left = strlen(line_in);
	char *line_out = XMALLOC(in_left + 1);
	char *iconv_out = line_out;
	size_t out_size = in_left;
	size_t out_left = in_left;

	while (in_left > 0)
	{
		if (iconv(cd, &iconv_in, &in_left, &iconv_out, &out_left) ==
			(size_t) -1 && errno != E2BIG)
		{
			free(line_out);
			*pline_out = NULL;
			return -1;
		}
		if (in_left > 0)
		{
			line_out = XREALLOC(line_out, out_size + in_left + 1);
			iconv_out = line_out + out_size - out_left;
			out_size += in_left;
			out_left += in_left;
		}
	}
	line_out[out_size - out_left] = 0;
	*pline_out = line_out;
	return 0;
}

typedef struct VECTOR(char *) vec_of_char;

static int child_fetch_article_iconv(ChildData *child_data,
		rw_buf *in_sbuf, rw_buf *out_fbuf, rw_modifier path_filter)
{
	int result = 0;
	int n;
	vec_of_char in_lines;
	vec_of_char out_lines;
	vec_of_char *output_lines = &in_lines;
	const char *message_id = NULL;
	iconv_t cd = (iconv_t) -1;
	int mime_version = -1;
	/* Content-Type may span multiple lines */
	int first_content_type = -1;
	int content_type = -1;
	int last_content_type = -1;
	int encoding = -1;
	int newsreader = -1;
	int charset_o = -1;		/* Offset of charset (after "charset=") in
					  Content-Type */
	const char *in_line;
	BOOL ascii = TRUE;
	char *from_charset = NULL;
	BOOL header;
	BOOL following_content_type = FALSE;

	VECTOR_INIT(in_lines, char *);
	VECTOR_INIT(out_lines, char *);
	while ((in_line = rw_read_line(in_sbuf)) != NULL)
	{
		if (in_line[0] == '.')
		{
			in_line = in_line + 1;
			if (!in_line[0])
			break;
		}
		if (path_filter)
			in_line = path_filter(in_line, child_data);
		in_line = XSTRDUP(in_line);
		VECTOR_APPEND(in_lines, in_line);
		if (!strncmp(in_line, "Message-I", 9))
			message_id = in_line + 12;
	}
	if (!message_id)
		message_id = _("<unknown>");
	if (!in_line && !rw_buf_eof(in_sbuf))
	{
		deblog_critical(_("child_fetch_article_iconv: NULL line from socket "
				"but not EOF\n"));
		result = -1;
		goto cfai_out;
	}

	deblog_debug(_("Considering %s for character set conversion\n"),
			message_id);

	/* If entirely 7-bit ASCII don't convert */
	for (n = 0; ascii && n < in_lines.vec_used_members; ++n)
	{
		char *c = in_lines.vec_data[n];

		deblog_debug("ASCII?: %s\n", c);
		while (*c)
		{
			if (*c & 0x80)
			{
			ascii = FALSE;
			break;
			}
			++c;
		}
	}
	if (ascii)
	{
		deblog_debug(_("Plain ASCII, not converting\n"));
		goto cfai_save;
	}
	else
	{
		deblog_debug(_("Non-ASCII characters found, may need conversion\n"));
	}

	/* Find important headers */
	for (n = 0; n < in_lines.vec_used_members; ++n)
	{
		char *line = in_lines.vec_data[n];

		if (!line[0])
			break;
		if (!strncmp(line, "Mime-Version:", 13))
			mime_version = n;
		else if (!strncmp(line, "Content-Type:", 13))
			first_content_type = content_type = n;
		else if (!strncmp(line, "Content-Transfer-Encoding:", 26))
			encoding = n;
		else if (!strncmp(line, "X-Newsreader:", 13))
			newsreader = n;
		if (n == content_type ||
			(content_type != -1 && following_content_type &&
			 isspace(line[0])))
		{
			char *substr = strstr(line, "charset=");

			if (substr)
			{
			charset_o = (substr - line) + 8;
			content_type = n;
			}
			following_content_type = TRUE;
		}
		else
		{
			if (following_content_type)
			last_content_type = n - 1;
			following_content_type = FALSE;
		}
	}
	if (last_content_type == -1)
		last_content_type = following_content_type;

	/* Better leave non plain messages alone unless header says all parts have
	 * same charset (is that possible?) */
	if (content_type != -1 && charset_o == -1 &&
			strstr(in_lines.vec_data[first_content_type], "text/"))
	{
		deblog_info(_("Content type apparently not simple text, "
				"not converting\n"));
		goto cfai_save;
	}

	/* Work out message's charset */
	if (charset_o != -1)
	{
		const char *orig_charset = in_lines.vec_data[content_type] + charset_o; 
		/* charset may be at end of Content-Type or have something else
		* after it separated by ; */
		const char *semi = strchr(orig_charset, ';');

		if (semi)
		{
			int l = semi - orig_charset;

			from_charset = XMALLOC(l + 1);
			strncpy(from_charset, orig_charset, l);
			from_charset[l] = 0;
		}
		else
		{
			from_charset = XSTRDUP(orig_charset);
		}
		deblog_debug(_("Article declares charset %s\n"), from_charset);
	}
	else if (newsreader != -1 && !strncmp(in_lines.vec_data[newsreader] + 14,
			"Microsoft Outlook", 17))
	{
		deblog_debug(_("Using MS OLE charset %s\n"), child_data->ole_charset);
		from_charset = child_data->ole_charset ?
			XSTRDUP(child_data->ole_charset) : NULL;
	}
	if (!from_charset)
	{
		deblog_debug(_("Using default charset %s\n"), child_data->from_charset);
		from_charset = child_data->from_charset ?
			XSTRDUP(child_data->from_charset) : NULL;
	}
	if (!from_charset)
	{
		deblog_debug(_("Can't determine source charset, not converting\n"));
		goto cfai_save;
	}

	/* No need to convert if charsets are same */
	if (!compare_charsets(from_charset, child_data->to_charset))
	{
		deblog_debug(_("Charsets %s and %s equivalent, not converting\n"),
			from_charset, child_data->to_charset);
		goto cfai_save;
	}

	deblog_info(_("Converting %s from %s to %s\n"), message_id,
		from_charset, child_data->to_charset);
	cd = open_convertor(child_data->to_charset, from_charset);
	if (cd == (iconv_t) -1)
		goto cfai_save;
	header = TRUE;
	for (n = 0; n < in_lines.vec_used_members; ++n)
	{
		char *line = NULL;

		if (get_converted_line(cd, in_lines.vec_data[n], &line) < 0)
		{
			deblog_warning(_("Conversion of %s from %s to %s failed\n"),
				message_id, from_charset, child_data->to_charset);
			deblog_info(_("Bad line was: %s\n"), in_lines.vec_data[n]);
			goto cfai_save;
		}
		if (header)
		{
			if (!line[0])
			{
				header = FALSE;
				/* End of header, make sure we add any we need that weren't in
				 * original */
				if (mime_version == -1)
				{
					char *s = XSTRDUP("Mime-Version: 1.0");
					VECTOR_APPEND(out_lines, s);
				}
				if (content_type == -1)
				{
					char *lowered_charset = lower_charset(
							child_data->to_charset, FALSE);
					char *s = XSTRDUPCAT("Content-Type: text/plain; charset=",
							lowered_charset);
	
					VECTOR_APPEND(out_lines, s);
					free(lowered_charset);
				}
				if (encoding == -1)
				{
					char *s = XSTRDUPCAT("Content-Transfer-Encoding: ",
							child_data->transfer_encoding);
					VECTOR_APPEND(out_lines, s);
				}
			}
			/*
			else if (n == mime_version)
			{
			}
			*/
			else if (n == content_type)
			{
				/* Need to correct with new content type */
				char *lowered_charset = lower_charset(
						child_data->to_charset, FALSE);
				char *old_line = line;
	
				if (charset_o != -1)
				{
					/* Conversion could have made charset_o invalid */
					char *substr = strstr(line, "charset=");
	
					if (substr)
					{
						charset_o = (substr - line) + 8;
						/* Is there anything after charset clause? */
						substr = strchr(substr, ';');
						/* Allocate enough space for old_line up to charset=
						* then new charset then any clauses following charset
						*/
						line = XMALLOC(charset_o + 1 + strlen(lowered_charset) +
							(substr ? strlen(substr) : 0));
						/* Copy line up to = of old charset */
						strncpy(line, old_line, charset_o);
						line[charset_o] = 0;
						/* Add new charset */
						strcat(line, lowered_charset);
						/* Add subsequent clause(s) */
						if (substr)
							strcat(line, substr);
						free(old_line);
					}
					else
					{
						charset_o = -1;
					}
				}
				if (charset_o == -1)
				{
					/* If there was a semicolon at the end of this line
					 * preserve it */
					BOOL semi = FALSE;
					int l = strlen(line);
	
					while (l && isspace(line[l]))
						--l;
					if (line[l] == ';')
						semi = TRUE;
					old_line = line;
					line = XSTRDUPCAT(old_line, "; charset=");
					free(old_line);
					old_line = line;
					line = XSTRDUPCAT(old_line, lowered_charset);
					free(old_line);
					if (semi)
					{
						old_line = line;
						line = XSTRDUPCAT(old_line, ";");
						free(old_line);
					}
				}
				free(lowered_charset);
			}
			else if (n == encoding)
			{
				/* Just replace this header with new version */
				free(line);
				line = XSTRDUPCAT("Content-Transfer-Encoding: ",
							child_data->transfer_encoding);
			}
		}
		VECTOR_APPEND(out_lines, line);
	}
	output_lines = &out_lines;

cfai_save:
	header = TRUE;
	for (n = 0; (result != -1) && (n < output_lines->vec_used_members); ++n)
	{
		char *line = output_lines->vec_data[n];

		if (!line[0])
		{
			header = FALSE;
		}
		else if (header)
		{
			line = convert_mime_header(child_data->to_charset,
					line, message_id);
		}
		if (rw_write_line(out_fbuf, line) < 0)
		{
			deblog_critical(_("child_fetch_article_iconv: "
				"Failed to write temporary file.\n"));
			result = -1;
		}
		if (line != output_lines->vec_data[n])
			free(line);
	}

cfai_out:
	free(from_charset);
	if (cd != (iconv_t) -1)
		iconv_close(cd);
	for (n = 0; n < in_lines.vec_used_members; ++n)
		free(in_lines.vec_data[n]);
	free(in_lines.vec_data);
	for (n = 0; n < out_lines.vec_used_members; ++n)
		free(out_lines.vec_data[n]);
	free(out_lines.vec_data);
	return result;
}

#define CHILD_FETCH_ARTICLE(child_data, in_sbuf, out_fbuf, filter) \
	(child_data->to_charset ? \
		child_fetch_article_iconv(child_data, in_sbuf, out_fbuf, filter) : \
		rw_stof(in_sbuf, out_fbuf, filter, child_data))

#else /* ! ENABLE_ICONV */

#define CHILD_FETCH_ARTICLE(child_data, in_sbuf, out_fbuf, filter) \
	rw_stof(in_sbuf, out_fbuf, filter, child_data)

#endif /* ! ENABLE_ICONV */

/* Returns TRUE if problem */
static BOOL child_pop_article(ChildData *child_data)
{
	BOOL result = FALSE;

	if (child_data->change_group[0])
	{
		deblog_info(
			_("Child now expecting to receive articles from group %ld\n"),
			child_data->group_ix[0]);
		if (!nntp_check_response(child_data->sbuf, child_data->server,
				_("Can't enter group"), "2", deblog_Level_Critical))
		{
			child_shutdown(child_data, TRUE);
			return TRUE;
		}
		child_send_command(child_data, command_EnteredGroup, "%ld",
			child_data->group_ix[0]);
	}
	else
	{
		int fd = -1;
		rw_buf *fbuf = NULL;
		char *filename = NULL;

		child_data->path_done = FALSE;
		deblog_debug(_("Expecting to receive %d/%d\n"), child_data->group_ix[0],
				child_data->artnum[0]);
		if (!nntp_check_response(child_data->sbuf, child_data->server,
				_("Bad response to ARTICLE"), "2",
				child_data->backfill[0] ?
				deblog_Level_Info : deblog_Level_Warning))
		{
			/* Don't need to report back if it's a backfill */
			if (!child_data->backfill[0])
			{
				child_send_command(child_data,
						command_ArticleUnavailable, "%d %ld",
						child_data->group_ix[0], child_data->artnum[0]);
			}
			result = TRUE;
		}
		if (!result)
		{
			fd = xmktmpindir(INCOMING_DIR, "Mesg", &filename, FALSE,
				_("Can't create temporary file to download message"));
			if (fd < 0)
			{
				child_shutdown(child_data, TRUE);
				result = TRUE;
			}
			else
			{
				deblog_debug(_("Saving %d/%d to %s\n"),
						child_data->group_ix[0], child_data->artnum[0],
						filename);
				fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
				if (CHILD_FETCH_ARTICLE(child_data, child_data->sbuf, fbuf,
						child_data->path ? child_path_filter : NULL) < 0)
				{
					deblog_critical(
							_("Can't download message to temporary file\n"));
					child_shutdown(child_data, TRUE);
					result = TRUE;
				}
			}
			if (fbuf)
				rw_free_buf(fbuf);
			if (fd != -1)
				close(fd);
			if (!result)
			{
				child_send_command(child_data,
						command_HaveDownloaded, "%ld %ld %d %s",
						child_data->group_ix[0], child_data->artnum[0],
						child_data->backfill[0], filename);
			}
		}
		free(filename);
	}
	child_shift_pipeline(child_data);
	return result;
}

static void child_get_article(ChildData *child_data, const char *data)
{
	sscanf(data, "%ld %ld %d",
			&child_data->group_ix[child_data->pipeline_depth],
			&child_data->artnum[child_data->pipeline_depth],
			&child_data->backfill[child_data->pipeline_depth]);
	child_data->change_group[child_data->pipeline_depth] = FALSE;
	rw_writef(child_data->sbuf, "ARTICLE %ld\r\n",
			child_data->artnum[child_data->pipeline_depth]);
	++child_data->pipeline_depth;
	deblog_debug(_("Increasing pipeline depth to %d\n"),
			child_data->pipeline_depth); 
	if (child_data->pipeline_depth >=
			(child_data->pipeline ? PIPELINE_DEPTH : 1))
	{
		deblog_debug(_("Pipeline full, popping\n"));
		child_pop_article(child_data);
	}
	child_request_article(child_data);
}

static void child_enter_group(ChildData *child_data, const char *data)
{
	char *groupname = XMALLOC(strlen(data) + 1);

	sscanf(data, "%ld %s", &child_data->rq_group_ix, groupname);
	deblog_debug(_("Entering group %ld %s to get articles\n"),
			child_data->rq_group_ix, groupname);
	child_data->group_ix[child_data->pipeline_depth] = child_data->rq_group_ix;
	child_data->change_group[child_data->pipeline_depth] = TRUE;
	nntp_enter_group(child_data->sbuf, child_data->server, groupname,
			NULL, NULL, FALSE);
	++child_data->pipeline_depth;
	deblog_debug(_("Increasing pipeline depth to %d\n"),
			child_data->pipeline_depth); 
	if (child_data->pipeline_depth >=
			(child_data->pipeline ? PIPELINE_DEPTH : 1))
	{
		deblog_debug(_("Pipeline full, popping\n"));
		child_pop_article(child_data);
	}
	deblog_debug(_("Entered %s, getting article\n"), groupname);
	child_request_article(child_data);
	free(groupname);
}

inline static char hex_digit_from_nibble(int nibble)
{
	if (nibble > 9)
		return nibble - 10 + 'A';
	else
		return nibble + '0';
}

static char *addr_to_hex(void *data, size_t len)
{
	size_t n;
	char *addr = XMALLOC(len * 2 + 1);
	
	for (n = 0; n < len; ++n)
	{
		addr[n * 2] = hex_digit_from_nibble((((char *) data)[n] >> 4) & 0xf);
		addr[n * 2 + 1] = hex_digit_from_nibble(((char *) data)[n] & 0xf);
	}
	addr[len * 2] = 0;
	return addr;
}

static void child_resolve(ChildData *child_data, const char *data)
{
	char *addr = XMALLOC(strlen(data));
	int port;
	struct sockaddr *s_addr;
	size_t addr_len;
	int proto;
	
	sscanf(data, "%s %d", addr, &port);
	if (nntp_resolve_address(addr, port, &s_addr, &addr_len, &proto))
	{
		free(addr);
		addr = addr_to_hex(s_addr, addr_len);
		child_send_command(child_data, command_Resolved, "%s %d", addr, proto);
	}
	else
	{
		child_shutdown(child_data, TRUE);
	}
	free(addr);
}

static void whinge_unlink(const char *pathname, int e)
{
	if (e == ENOENT)
	{
		deblog_info(_("Tried to delete %s after uploading,\n"), pathname);
		deblog_info(_("but it was already deleted by another thread.\n"));
	}
	else
	{
		deblog_critical(_("%s posted OK (or was duplicate),\n"), pathname);
		deblog_critical(_("but couldn't be removed from outgoing queue:\n"));
		deblog_critical("%s\n", strerror(e));
	}
}

static void child_post_article(ChildData *child_data, const char *pathname)
{
	rw_buf *fbuf;
	int fd;
	const char *response = NULL;
	int result;
	BOOL close_busy = FALSE;

	/* Only need to read file, but has to be opened writable for lockf */
	if ((fd = open(pathname, O_RDWR)) == -1)
	{
		if (errno != ENOENT)
		{
			int e = errno;
			char *reason = XSTRDUPCAT(_("Couldn't open for uploading: "),
				strerror(e));

			deblog_critical(_("Couldn't open %s for uploading:\n"), pathname);
			deblog_critical("%s\n", strerror(e));
			child_send_command(child_data, command_BadUpload, "%s %s %s",
					pathname, child_data->server, reason);
		}
		else
		{
			deblog_info(
				_("%s doesn't exist for uploading (already uploaded)\n"),
				pathname);
			child_send_command(child_data, command_LocalDuplicate, "");
		}
		return;
	}
	if (lockf(fd, F_TLOCK, 0))
	{
		int e = errno;

		if (e == EACCES || e == EAGAIN)
		{
			deblog_info(_("%s locked for uploading (already being uploaded)\n"),
				pathname);
			child_send_command(child_data, command_LockedDuplicate, "");
		}
		else
		{
			char *reason = XSTRDUPCAT(_("Couldn't lock for uploading: "),
				strerror(e));

			child_send_command(child_data, command_BadUpload, "%s %s %s",
					pathname, child_data->server, reason);
			free(reason);
		}
		close(fd);
		return;
	}
	deblog_info(_("Uploading %s\n"), pathname);
	fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
	rw_write_line(child_data->sbuf, "POST");
	if (!nntp_check_response(child_data->sbuf, child_data->server,
			_("Bad response to POST"), "3", deblog_Level_Critical))
	{
		rw_close_and_free(fbuf);
		child_send_command(child_data, command_NoUploads, "");
		return;
	}
	result = rw_ftos(fbuf, child_data->sbuf, NULL, NULL);
	if (!result)
	/* From now on result == 0 means posting was successful, or rejected
	 * because it's a duplicate, and response is reason for failure */
	{
		response = rw_read_line(child_data->sbuf);
		deblog_debug("%s\n", response ? response : _("No response"));
		if (!response)
		{
			child_send_command(child_data, command_Error,
					_("No response from server when posting\n"));
			return;
		}
		else if (!maincf_get_report_dups() &&
				(!strncmp(response, "441 435 ", 8)
				|| strstr(response, "Duplicate")
				|| strstr(response, "Already got")
				|| (child_data->ignore_436 &&
					!strncmp(response, "441 436 <", 9))))
		{
			deblog_info(_("Upload %d already known by server\n"), pathname);
			child_send_command(child_data, command_RemoteDuplicate, pathname);
		}
		else if (response[0] != '2')
		{
			result = -1;
		}
		else
		{
			deblog_info(_("%s uploaded successfully\n"), pathname);
			child_send_command(child_data, command_Uploaded, pathname);
		}
	}
	else
	{
		response = strerror(errno);
	}
	/* Lock means we should remove file before closing it, but that could be
	 * impossible on some systems */
	if (result)
	{
		deblog_info(_("Posting to %s failed: %s\n"),
			child_data->server, response);
		child_send_command(child_data, command_BadUpload, "%s %s %s",
			pathname, child_data->server, response);
	}
	else
	{
		if (unlink(pathname))
		{
			if (errno == EBUSY)
			{
				close_busy = TRUE;
				deblog_info(_("Can't delete %s after uploading "
							"while same thread still has a lock.\n"),
						pathname);
				deblog_info(_("This is a system limitation. "
							"Will retry after releasing lock.\n"));
			}
			else
			{
				whinge_unlink(pathname, errno);
			}
		}
	}
	rw_close_and_free(fbuf);
	if (close_busy && unlink(pathname) && errno != ENOENT)
		whinge_unlink(pathname, errno);
}

static void child_flush_pipeline(ChildData *child_data)
{
	deblog_info(_("Flushing pipelined articles prior to next task\n"));
	while (child_data->pipeline_depth)
		child_pop_article(child_data);
	if (child_data->all_requested)
	{
		deblog_message(_("All articles downloaded\n"));
		child_data->all_requested = FALSE;
		child_send_command(child_data, command_AllDownloaded, "");
	}
}

static void child_find_xref_in_xover(ChildData *child_data)
{
	int i = 0;
	const char *line;
	int ix = -1;
	BOOL includes_name = FALSE;
	BOOL is_newsgroups = FALSE;

	rw_write_line(child_data->sbuf, "LIST OVERVIEW.FMT");
	if (!nntp_check_response(child_data->sbuf, child_data->server,
			_("Can't get overview format"), "2", deblog_Level_Warning))
	{
		child_send_command(child_data, command_FoundXrefInXover, "-1 0 0");
		return;
	}
	for (i = 0; (line = rw_read_line(child_data->sbuf)) != NULL 
			&& strcmp(line, "."); ++i)
	{
		if (!strncmp(line, "Newsgroups:", 11))
		{
			is_newsgroups = TRUE;
		}
		else if (ix == -1 && !strncmp(line, "Xref:", 5))
		{
			is_newsgroups = FALSE;
		}
		else
		{
			/* Skip rest of loop body if not a match */
			continue;
		}
		ix = i;
		deblog_debug(_("Found %s in overview.fmt at ix %d\n"),
			line, ix);
		/* See if line ends in "full" */
		if ((line = strstr(line, "full")) != NULL && !strcmp(line, "full"))
		{
			deblog_debug(
				_("Overview for Newsgroups/Xref includes header name\n"));
			includes_name = TRUE;
		}
		else
		{
			deblog_debug(
				_("Overview for Newsgroups/Xref excludes header name\n"));
		}
	}
	child_send_command(child_data, command_FoundXrefInXover, "%d %d %d", ix,
			includes_name, is_newsgroups);
}

static void child_main_loop(ChildData *child_data)
{
	const intercom_data *command;

	for (;;)
	{
		command = intercom_get_command(child_data->intercom);
		switch (command->command)
		{
#if ENABLE_ICONV
			case command_ToCharset:
				child_data->to_charset = XSTRDUP(command->data);
				break;
			case command_FromCharset:
				child_data->from_charset = XSTRDUP(command->data);
				break;
			case command_OLECharset:
				child_data->ole_charset = XSTRDUP(command->data);
				break;
			case command_TransferEncoding:
				child_data->transfer_encoding = XSTRDUP(command->data);
				break;
#endif
			case command_Upload:
				child_flush_pipeline(child_data);
				child_post_article(child_data, command->data);
				break;
			case command_Resolve:
				child_resolve(child_data, command->data);
				break;
			case command_Connect:
				child_connect(child_data, command->data);
				break;
			case command_ModeReader:
				/*assert(child_data->sbuf);*/
				if (!child_data->sbuf)
				{
					deblog_error(_("Thread %d asked to issue MODE READER " \
							"when not connected\n"), child_data->thread_idx);
				}
				if (nntp_set_mode(child_data->sbuf,
						child_data->server, "READER"))
				{
					child_shutdown(child_data, TRUE);
				}
				else
				{
					child_send_command(child_data, command_Ready, "");
				}
				break;
			case command_Auth:
				child_authenticate(child_data, command->data);
				break;
			case command_FindXrefInXover:
				child_find_xref_in_xover(child_data);
				break;
			case command_GetGroupStats:
				child_flush_pipeline(child_data);
				child_get_group_stats(child_data, command->data);
				break;
			case command_EnterGroup:
				child_enter_group(child_data, command->data);
				break;
			case command_GetArticles:
				child_issue_xhdr(child_data, command->data, FALSE);
				break;
			case command_GetXover:
				child_issue_xhdr(child_data, command->data, TRUE);
				break;
			case command_GetArticle:
				child_get_article(child_data, command->data);
				break;
			case command_AllRequested:
				child_data->all_requested = TRUE;
				break;
			case command_Throttle:
				usleep(atoi(command->data) * 1000);
				child_request_article(child_data);
				break;
			case command_Shutdown:
				deblog_info(
					_("<%d> Received shutdown command, %d %s in pipeline\n"),
					child_data->thread_idx,
					child_data->pipeline_depth,
					child_data->pipeline_depth == 1 ? _("article") :
						_("articles"));
				child_data->sent_quit = TRUE;
				if (child_data->sbuf)
				{
					rw_write_line(child_data->sbuf, "QUIT");
					child_flush_pipeline(child_data);
				}
				child_shutdown(child_data, FALSE);
				break;
			case command_WaitPending:
				child_flush_pipeline(child_data);
				child_send_command(child_data, command_WaitAck, "");
				break;
			default:
				deblog_error(_("Child doesn't understand command code %d\n"),
						command->command);
				break;
		}
	}
}

void child_start(int thread_idx, IntercomQ *in_q, IntercomQ *out_q)
{
	ChildData *child_data;
#if !USE_PTHREADS
	static BOOL exit_handled = FALSE;

	deblog_debug("child_start: thread %d\n", thread_idx);
	if (!exit_handled)
	{
		atexit(child_exit);
		exit_handled = TRUE;
	}
	signal(SIGHUP, child_signal_handler);
	signal(SIGINT, SIG_IGN);
	signal(SIGPIPE, child_signal_handler);
	signal(SIGTERM, child_signal_handler);
	signal(SIGALRM, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
	signal(SIGUSR2, SIG_IGN);
	child_data = &current_child_data;
	child_shouldnt_exit = FALSE;
#else /* USE_PTHREADS */
	child_data = calloc(1, sizeof(ChildData));
	if (thread_idx >= child_nthreads)
	{
		int n;
		int nthreads = thread_idx + 1;
		static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
		
		pthread_mutex_lock(&mutex);
		child_data_table = realloc(child_data_table,
				nthreads * sizeof(ChildData *));
		child_thread_table = realloc(child_thread_table,
				nthreads * sizeof(pthread_t));
		for (n = child_nthreads; n < nthreads - 1; ++n)
		{
			child_data_table[n] = NULL;
			child_thread_table[n] = main_thread;
		}
		child_nthreads = nthreads;
		pthread_mutex_unlock(&mutex);
	}
	child_data_table[thread_idx] = child_data;
	child_thread_table[thread_idx] = pthread_self();
#endif /* USE_PTHREADS */
	child_data->in_q = in_q;
	child_data->out_q = out_q;
	child_data->intercom = intercom_new(child_data->in_q);
	child_data->thread_idx = thread_idx;
	child_data->pipeline_depth = 0;
	child_data->path = NULL;
#if ENABLE_ICONV
	child_data->to_charset = child_data->from_charset =
			child_data->ole_charset = NULL;
	child_data->transfer_encoding = "8bit";
#endif
#if !USE_PTHREADS
	deblog_use_pipe(child_data->out_q, thread_idx);
#endif
	child_main_loop(child_data);
}

ChildData *child_data_from_index(int thread_idx)
{
#if USE_PTHREADS
	assert(thread_idx >= 0 && thread_idx < child_nthreads);
	return child_data_table[thread_idx];
#else /* USE_PTHREADS */
	return &current_child_data;
#endif /* USE_PTHREADS */
}

#if USE_PTHREADS
int child_lookup_current_thread_idx(void)
{
	int n;
	pthread_t t = pthread_self();
	
	if (t == main_thread)
		return -1;
	for (n = 0; n < child_nthreads; ++n)
	{
		if (child_thread_table[n] == t)
			return n;
	}
	return -1;
}
#endif
