/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Logging for debugging */


#include "common.h"
#include "child.h"
#include "intercom.h"
#include "nscurses.h"
#include "stats.h"
#ifdef ENABLE_KEYED_LOG
#include "vector.h"
#endif

#if USE_PTHREADS

#include <pthread.h>
#include "main.h"

static int deblog_get_thread_idx(void)
{
	return child_lookup_current_thread_idx();
}

#else

static rw_buf *deblog_buf = NULL;

/* When other processes are forked this is updated to their thread index */
static int deblog_thread_idx = -1;

void deblog_use_pipe(rw_buf *buf, int thread_idx)
{
	deblog_buf = buf;
	deblog_thread_idx = thread_idx;
	deblog_debug(_("Ready to log via pipe\n"), thread_idx);
}

static int deblog_get_thread_idx(void)
{
	return deblog_thread_idx;
}

#endif /* !USE_PTHREADS */

BOOL deblog_be_brief = FALSE;

static deblog_mask_t deblog_mask = deblog_Mask_Default;

static BOOL deblog_need_lf_flag = FALSE;


static BOOL deblog_recursion_allowed = FALSE;
static BOOL deblog_merged_sinks = TRUE;
static BOOL deblog_stats_no_update = FALSE;

#ifdef ENABLE_KEYED_LOG
/* Using a vector could get horribly inefficient if there are loads of
 * different keys, but hopefully there shouldn't be enough to warrant the
 * overhead of something more sophisticated */
struct VECTOR(const char *) deblog_keys;

static void deblog_init_keys(void)
{
	static BOOL inited = FALSE;

	if (!inited)
	{
		inited = TRUE;
		VECTOR_INIT(deblog_keys, const char *);
	}
}

/* Returns pointer to found string in keys vector, or NULL */
static char const **deblog_find_key(const char *key)
{
	size_t k;

	deblog_init_keys();
	for (k = 0; k < deblog_keys.vec_used_members; ++k)
	{
		if (!strcmp(deblog_keys.vec_data[k], key))
		{
			return &deblog_keys.vec_data[k];
		}
	}
	return NULL;
}

#endif

/* Makes sure we haven't left cursor at end of stats line at exit */
static void deblog_clear_line(void)
{
	if (deblog_get_thread_idx() == -1 &&
					deblog_need_lf_flag && !nscurses_active())
	{
		putchar('\n');
		deblog_need_lf_flag = FALSE;
	}
}

void deblog_separate_sinks(void)
{
	deblog_merged_sinks = FALSE;
}

void deblog_allow_recursion(BOOL allowed)
{
	deblog_recursion_allowed = allowed;
}

static void deblog_print_to_stream(int thread_idx, FILE *sink, deblog_level_t
		level, const char *message)
{
	/* Don't print thread indexes to stdout in brief mode */
	if (!deblog_be_brief || sink == stderr)
	{
		if (thread_idx != -1)
		{
			fprintf(sink, _("<%d> "), thread_idx);
		}
		else
		{
			fputs(_("<M> "), sink);		/* M for Main/Master */
		}
	}
	switch (level)
	{
		case deblog_Level_Error:
			fputs(_("Error: "), sink);
			break;
		case deblog_Level_Critical:
			fputs(_("Critical: "), sink);
			break;
		case deblog_Level_Warning:
			fputs(_("Warning: "), sink);
			break;
		case deblog_Level_Brief:
			if (sink == stderr)
				fputs(_("Brief: "), sink);
			break;
		case deblog_Level_Message:
			if (sink == stderr)
				fputs(_("Message: "), sink);
			break;
		case deblog_Level_Info:
			fputs(_("Info: "), sink);
			break;
		case deblog_Level_Debug:
			fputs(_("Debug: "), sink);
			break;
#ifdef ENABLE_KEYED_LOG
		case deblog_Level_Keyed:
			fputs(_("Debug: "), sink);
			break;
#endif
		default:
			break;
	}
	fputs(message, sink);
	fflush(sink);
}

static void deblog_choose_stream_and_print(int thread_idx, deblog_level_t level,
		const char *message)
{
	FILE *sink = (level == deblog_Level_Error
			|| level == deblog_Level_Critical
			|| level == deblog_Level_Warning
			|| (level == deblog_Level_Brief && deblog_be_brief)
			|| (level == deblog_Level_Message && !deblog_be_brief)
			) ? stdout : stderr;

	/* First decide whether to use curses. Errors cause shutdown and are
	 * printed to stdout; info and debug are sent to stderr if using -s */
	if (nscurses_active())
	{
		if (level == deblog_Level_Error)
		{
			nscurses_stop();
			deblog_print_to_stream(thread_idx, stdout, level, message);
		}
		else if (sink == stdout || deblog_merged_sinks)
		{
			nscurses_deblog(thread_idx, level, message);
		}
		if (sink == stderr && !deblog_merged_sinks)
		{
			deblog_print_to_stream(thread_idx, stderr, level, message);
		}
	}
	else
	{
		/* stdout or stderr. First, do we need a newline to get us clear of
		 * stats display? */
		if (deblog_need_lf_flag && (deblog_merged_sinks || sink == stdout))
			stats_clear_line();
		deblog_print_to_stream(thread_idx, sink, level, message);
		if (sink == stdout || deblog_merged_sinks)
		{
			deblog_need_lf_flag = FALSE;
		}
	}
	/* Finally, whether curses or not, duplicate stdout messages on stderr
	 * if -s */
	if (sink == stdout && !deblog_merged_sinks)
	{
		deblog_print_to_stream(thread_idx, stderr, level, message);
	}
	if (!deblog_stats_no_update && level != deblog_Level_Error)
	{
		stats_update(sink == stdout);
	}
}

void deblog_print_thread(int thread_idx, deblog_level_t level,
		const char *message)
{
#if USE_PTHREADS
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
#else
	static BOOL recursion = FALSE;

	if (recursion && !deblog_recursion_allowed)
	{
		if (nscurses_active())
		{
			nscurses_emergency_stop();
		}
		fprintf(stderr, _("Error: Logging system called recursively\n"));
		exit(1);
	}
#endif /* !USE_PTHREADS */
	if (!(deblog_mask & (1 << level)))
		return;
#ifdef ENABLE_KEYED_LOG
	if (level == deblog_Level_Keyed)
	{
		char *key = XSTRDUP(message);
		char *key_end = strchr(key, ' ');
		char const **found = NULL;

		if (key_end)
			*key_end = 0;
		found = deblog_find_key(key);
		free(key);
		if (!found)
			return;
	}
#endif
	/* Send message to parent process if we're a thread; sometimes this gets
	 * called by main thread for a child thread so need to check real current
	 * thread rather than thread_idx. */
#if !USE_PTHREADS
	recursion = TRUE;
	if (deblog_buf)
	{
		size_t len;
		char *lftocr = XSTRDUP(message);

		/* Replace trailing LF with CR to avoid upsetting rw_read_line */
		len = strlen(lftocr) - 1;
		if (lftocr[len] == '\n')
			lftocr[len] = '\r';
		intercom_send_command(deblog_buf, command_Log, thread_idx,
				"%d %s", level, lftocr);
		free(lftocr);
	}
	else
	{
		deblog_choose_stream_and_print(thread_idx, level, message);
	}
	recursion = FALSE;
#else /* USE_PTHREADS */
	pthread_mutex_lock(&mutex);
	deblog_choose_stream_and_print(thread_idx, level, message);
	pthread_mutex_unlock(&mutex);
#endif /* USE_PTHREADS */
	if (level == deblog_Level_Error)
		exit(1);
}

void deblog_vprintf_thread(int thread_idx, deblog_level_t level,
		const char *format, va_list ap)
{
	char *formatted;

	if (vasprintf(&formatted, format, ap) <= 0)
	{
		fprintf(stderr, _("Unable to format logging message\n"));
		exit(1);
	}
	deblog_print_thread(thread_idx, level, formatted);
	free(formatted);
}

void deblog_printf_thread(int thread_idx, deblog_level_t level,
		const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, level, format, ap);
	va_end(ap);
}

void deblog_error_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Error, format, ap);
	va_end(ap);
}

void deblog_critical_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Critical, format, ap);
	va_end(ap);
}

void deblog_warning_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Warning, format, ap);
	va_end(ap);
}

void deblog_brief_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Brief, format, ap);
	va_end(ap);
}

void deblog_message_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Message, format, ap);
	va_end(ap);
}

void deblog_info_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Info, format, ap);
	va_end(ap);
}

void deblog_debug_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Debug, format, ap);
	va_end(ap);
}

void deblog_verror_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Error, format, ap);
}

void deblog_vcritical_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Critical, format, ap);
}

void deblog_vwarning_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Warning, format, ap);
}

void deblog_vbrief_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Brief, format, ap);
}

void deblog_vmessage_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Message, format, ap);
}

void deblog_vinfo_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Info, format, ap);
}

void deblog_vdebug_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Debug, format, ap);
}

void deblog_vprintf(deblog_level_t level, const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(), level, format, ap);
}

void deblog_printf(deblog_level_t level, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(), level, format, ap);
	va_end(ap);
}

void deblog_error(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Error, format, ap);
	va_end(ap);
}

void deblog_critical(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Critical, format, ap);
	va_end(ap);
}

void deblog_warning(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Warning, format, ap);
	va_end(ap);
}

void deblog_brief(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Brief, format, ap);
	va_end(ap);
}

void deblog_message(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Message, format, ap);
	va_end(ap);
}

void deblog_info(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Info, format, ap);
	va_end(ap);
}

void deblog_debug(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Debug, format, ap);
	va_end(ap);
}

void deblog_verror(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Error, format, ap);
}

void deblog_vcritical(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Critical, format, ap);
}

void deblog_vwarning(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Warning, format, ap);
}

void deblog_vbrief(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Brief, format, ap);
}

void deblog_vmessage(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Message, format, ap);
}

void deblog_vinfo(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Info, format, ap);
}

void deblog_vdebug(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Debug, format, ap);
}

void deblog_need_lf(void)
{
	static BOOL atexit_registered = FALSE;

	deblog_need_lf_flag = TRUE;
	if (!atexit_registered)
	{
		atexit_registered = TRUE;
		atexit(deblog_clear_line);
	}
}

void deblog_set_mask(deblog_mask_t mask)
{
	deblog_mask = mask;
}

void deblog_enable_level(deblog_level_t level)
{
	deblog_mask |= 1 << level;
}

void deblog_disable_level(deblog_level_t level)
{
	deblog_mask &= ~(1 << level);
}

void deblog_disable_stats_update(void)
{
	deblog_stats_no_update = TRUE;
}

#ifdef ENABLE_KEYED_LOG
void deblog_keyed(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Keyed, format, ap);
	va_end(ap);
}

void deblog_vkeyed(const char *format, va_list ap)
{
	deblog_vprintf_thread(deblog_get_thread_idx(),
					deblog_Level_Keyed, format, ap);
}

void deblog_keyed_thread(int thread_idx, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	deblog_vprintf_thread(thread_idx, deblog_Level_Keyed, format, ap);
	va_end(ap);
}

void deblog_vkeyed_thread(int thread_idx, const char *format, va_list ap)
{
	deblog_vprintf_thread(thread_idx, deblog_Level_Keyed, format, ap);
}

void deblog_enable_key(const char *key)
{
	char const **fkey = deblog_find_key(key);

	if (!fkey)
	{
		VECTOR_APPEND(deblog_keys, key);
	}
}

void deblog_disable_key(const char *key)
{
	char const **fkey = deblog_find_key(key);

	if (fkey)
	{
		*fkey = NULL;
	}
}

#endif
