#ifndef DEBLOG_H
#define DEBLOG_H

/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Logging for debugging */


#include <stdarg.h>

#include "rw.h"

/* NB, avoid \r at end of line, and avoid \n anywhere else, because of the way
 * messages may be piped from child to parent process */

/* Logging levels are similar to glib's; Message goes to stdout; rest go to
 * stderr */
typedef enum {
    deblog_Level_Error,		/* Fatal error, causes exit */
    deblog_Level_Critical,	/* Error, but can continue */
    deblog_Level_Warning,
    deblog_Level_Brief,		/* Special message used instead of Level_Message
				   in brief mode */
    deblog_Level_Message,
    deblog_Level_Info,		/* Mainly for debugging */
    deblog_Level_Debug,		/* Verbose debugging */
#ifdef ENABLE_KEYED_LOG
    deblog_Level_Keyed		/* First word of message is key; only printed
				   if key is enabled; key in message must be
				   followed by space (ASCII 0x20) */
#endif
} deblog_level_t;

/* Whether we're in brief mode */
extern BOOL deblog_be_brief;

/* These are the logging functions themselves, to be called printf-style */
void deblog_printf(deblog_level_t, const char *, ...);
void deblog_error(const char *, ...);
void deblog_critical(const char *, ...);
void deblog_warning(const char *, ...);
void deblog_brief(const char *, ...);
void deblog_message(const char *, ...);
void deblog_info(const char *, ...);
void deblog_debug(const char *, ...);

/* As above, called vprintf-style */
void deblog_vprintf(deblog_level_t, const char *, va_list);
void deblog_verror(const char *, va_list);
void deblog_vcritical(const char *, va_list);
void deblog_vwarning(const char *, va_list);
void deblog_vbrief(const char *, va_list);
void deblog_vmessage(const char *, va_list);
void deblog_vinfo(const char *, va_list);
void deblog_vdebug(const char *, va_list);

/* As above with a thread index */
void deblog_printf_thread(int thread_idx, deblog_level_t, const char *, ...);
void deblog_error_thread(int thread_idx, const char *, ...);
void deblog_critical_thread(int thread_idx, const char *, ...);
void deblog_warning_thread(int thread_idx, const char *, ...);
void deblog_brief_thread(int thread_idx, const char *, ...);
void deblog_message_thread(int thread_idx, const char *, ...);
void deblog_info_thread(int thread_idx, const char *, ...);
void deblog_debug_thread(int thread_idx, const char *, ...);

void deblog_vprintf_thread(int thread_idx, deblog_level_t, const char *,
	va_list);
void deblog_verror_thread(int thread_idx, const char *, va_list);
void deblog_vcritical_thread(int thread_idx, const char *, va_list);
void deblog_vwarning_thread(int thread_idx, const char *, va_list);
void deblog_vbrief_thread(int thread_idx, const char *, va_list);
void deblog_vmessage_thread(int thread_idx, const char *, va_list);
void deblog_vinfo_thread(int thread_idx, const char *, va_list);
void deblog_vdebug_thread(int thread_idx, const char *, va_list);

#ifdef ENABLE_KEYED_LOG
void deblog_keyed(const char *, ...);
void deblog_vkeyed(const char *, va_list);
void deblog_keyed_thread(int thread_idx, const char *, ...);
void deblog_vkeyed_thread(int thread_idx, const char *, va_list);

void deblog_enable_key(const char *key);
void deblog_disable_key(const char *key);
#endif

/* No printf-style formatting: this makes it safe to log messages without
 * arguments, but which may have %'s in them */
void deblog_print_thread(int thread_idx, deblog_level_t level, const char
	*message);

/* This tells deblog that the last line written to the screen didn't end with a
 * LF, so it must print one itself before the next message */
void deblog_need_lf(void);

typedef enum {
    deblog_Mask_Error = 1 << deblog_Level_Error,
    deblog_Mask_Critical = 1 << deblog_Level_Critical,
    deblog_Mask_Warning = 1 << deblog_Level_Warning,
    deblog_Mask_Brief = 1 << deblog_Level_Brief,
    deblog_Mask_Message = 1 << deblog_Level_Message,
    deblog_Mask_Info = 1 << deblog_Level_Info,
    deblog_Mask_Debug = 1 << deblog_Level_Debug,
#ifdef ENABLE_KEYED_LOG
    deblog_Mask_Keyed = 1 << deblog_Level_Keyed,
#endif

    deblog_Mask_Default = deblog_Mask_Error
	| deblog_Mask_Critical
	| deblog_Mask_Warning
	| deblog_Mask_Brief
	| deblog_Mask_Message,
    deblog_Mask_All = deblog_Mask_Default
	| deblog_Mask_Info
	| deblog_Mask_Debug
#ifdef ENABLE_KEYED_LOG
	| deblog_Mask_Keyed
#endif
} deblog_mask_t;

/* Sets mask to absolute value, OR'd from deblog_mask values. Default value is
 * all except Info and Debug */
void deblog_set_mask(deblog_mask_t);

/* Enables/disables an individual level */
void deblog_enable_level(deblog_level_t);
void deblog_disable_level(deblog_level_t);

/* We're a subthread, so log via pipe */
void deblog_use_pipe(rw_buf *, int thread_idx);

/* Allows deblog_vprintf to be called recursively under exceptional
 * circumstances eg if interrupted by a signal causing shutdown */
void deblog_allow_recursion(BOOL allowed);

/* Normally we assume that stdout/stderr are merged. This tells it they're
 * separate. It makes a difference because of the way the stats output
 * overwrites lines. */
void deblog_separate_sinks(void);

/* Normally, stats are forced updated after each message, use this to turn it
 * off at end of run */
void deblog_disable_stats_update(void);

#endif	/* ! DEBLOG_H */
