/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "histgdbm.h"
#include "today.h"

static BOOL use_reorganize = FALSE;

void histgdbm_use_reorganize(BOOL use)
{
    use_reorganize = use;
}

#ifdef HAVE_GDBM_H

#include <gdbm.h>

#include "history.h"
#include "main.h"
#include "maincf.h"

/* Keys are Message-IDs (without <>), values are days since 1900 */
static GDBM_FILE histgdbm_dbf = NULL;


/*
static void histgdbm_scan(void)
{
    datum key;
    datum oldkey;

    oldkey.dsize = 0;
    oldkey.dptr = NULL;
    deblog_debug(_("Scanning history database\n"));
    for (key = gdbm_firstkey(histgdbm_dbf); key.dptr; key =
	    gdbm_nextkey(histgdbm_dbf, key))
    {
	datum content = gdbm_fetch(histgdbm_dbf, key);

	deblog_debug(_("Found <%*s>, %ld stored at %p, %p\n"),
		key.dsize, key.dptr, *((long int *) content.dptr),
		key.dptr, content.dptr);
	free(oldkey.dptr);
	free(content.dptr);
	oldkey = key;
    }
    free(oldkey.dptr);
    deblog_debug(_("Scan complete\n"));
}
*/

static void histgdbm_expire(void)
{
    datum key;
    /* Deleting keys may change order, so have to build vector */
    size_t tbd_size = 0;
    size_t tbd_used = 0;
    size_t tbd;
    datum *to_be_deleted = NULL;
    long int total = 0;
    datum oldkey;

    oldkey.dsize = 0;
    oldkey.dptr = NULL;
    deblog_message(_("Expiring history database\n"));
    deblog_debug(_("Today is day %ld\n"), today);
    deblog_info(_("Scanning for expiry candidates\n"));
    for (key = gdbm_firstkey(histgdbm_dbf); key.dptr; key =
	    gdbm_nextkey(histgdbm_dbf, key))
    {
	datum content = gdbm_fetch(histgdbm_dbf, key);

	free(oldkey.dptr);
	oldkey.dptr = NULL;
	if (today - *((long int *) content.dptr) >
		maincf_get_gdbm_expiry())
	{
	    if (tbd_used >= tbd_size)
	    {
		tbd_size = tbd_size ? tbd_size * 2 : 128;
		to_be_deleted = realloc(to_be_deleted, tbd_size *
			sizeof(datum));
		if (!to_be_deleted)
		{
		    gdbm_close(histgdbm_dbf);
		    histgdbm_dbf = NULL;
 		    deblog_error(
			    _("Not enough memory to expire history database\n")
			    );
		}
	    }
	    to_be_deleted[tbd_used++] = key;
	    deblog_debug(_("<%s>, %ld to be expired\n"), key.dptr,
		    *((long int *) content.dptr));
	}
	else
	    oldkey = key;
	free(content.dptr);
	++total;
    }
    free(oldkey.dptr);
    deblog_info(_("Purging history database\n"));
    for (tbd = 0; tbd < tbd_used; ++tbd)
    {
	gdbm_delete(histgdbm_dbf, to_be_deleted[tbd]);
	free(to_be_deleted[tbd].dptr);
    }
    if (tbd_used && use_reorganize)
    {
	deblog_info(_("Reorganizing history database\n"));
	gdbm_reorganize(histgdbm_dbf);
    }
    deblog_message(_("Expired %ld/%ld history entries\n"), tbd_used, total);
}

static void histgdbm_close(void)
{
    if (histgdbm_dbf)
    {
	if (main_normal_exit)
	    histgdbm_expire();
	gdbm_close(histgdbm_dbf);
	histgdbm_dbf = NULL;
    }
}

/* Returns a new string (must be freed when done) containing Mesage-ID stripped
 * of <> */
static char *histgdbm_strip_brackets(const char *withb)
{
    size_t len = strlen(withb);
    char *withoutb = XMALLOC(len - 1);

    strncpy(withoutb, withb + 1, len - 2);
    withoutb[len - 2] = 0;
    return withoutb;
}

static BOOL histgdbm_check(const char *msgid)
{
    datum key;
    BOOL result;
    
    key.dptr = histgdbm_strip_brackets(msgid);
    key.dsize = strlen(key.dptr) + 1;
    result = gdbm_exists(histgdbm_dbf, key);
    free(key.dptr);
    return result;
}

static void histgdbm_add(const char *msgid)
{
    datum key;
    datum content;
    
    key.dptr = histgdbm_strip_brackets(msgid);
    key.dsize = strlen(key.dptr) + 1;
    deblog_debug(_("Adding <%s> to history\n"), key.dptr);
    content.dptr = (void *) &today;
    content.dsize = sizeof(long int);
    if (gdbm_store(histgdbm_dbf, key, content, GDBM_REPLACE))
    {
	deblog_critical(_("Could not add %s to history: %s\n"),
		msgid, gdbm_strerror(gdbm_errno));
    }
    free(key.dptr);
}

void histgdbm_init(void)
{
    /* gdbm_open lacks const for filename, so use temporary to avoid warning */
    char *filename = XSTRDUP(RC_DIR "/history");

    history_register_implementation(histgdbm_check, histgdbm_add);
    histgdbm_dbf = gdbm_open(filename, 0, GDBM_WRCREAT |
	    (maincf_get_gdbm_fast() ? GDBM_FAST : 0), 0644, NULL);
    free(filename);
    if (!histgdbm_dbf)
    {
	deblog_error(_("Could not open history database: %s\n"),
		gdbm_strerror(gdbm_errno));
    }
    atexit(histgdbm_close);
}

#endif /* HAVE_GDBM_H */

