/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* It all starts here */

#include "common.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#if USE_PTHREADS
#include <pthread.h>
#endif

#include "group.h"
#include "hashtree.h"
#include "histcheck.h"
#include "histgdbm.h"
#include "history.h"
#include "histsnews.h"
#include "intercom.h"
#include "kiddy.h"
#include "local.h"
#include "nntp.h"
#include "nscurses.h"
#include "nsthread.h"
#include "score.h"
#include "server.h"
#include "snstore.h"
#include "stats.h"
#include "today.h"
#include "uploads.h"

#if USE_PTHREADS
pthread_t main_thread;
#endif

static void main_enter_new_group(NSThread *thread);

static enum {
	PM_nntp,
	PM_rnews,
	PM_batch,
	PM_snstore
} main_post_method;

/* Allows rnews file to be repaired if something goes wrong in the middle of
 * adding to it */
static off_t main_rnews_size = -1;

static char *main_snstore = NULL;

/* Whether exit is normal or response to signal */
BOOL main_normal_exit = FALSE;

#define RNEWS_FILE (INCOMING_DIR "/rnews")
#define BATCH_FILE (INCOMING_DIR "/batch")

static void main_repair_rnews_atexit(void)
{
	static BOOL recursive_exit = FALSE;
	
	if (!recursive_exit && main_rnews_size != -1)
	{
		recursive_exit = TRUE;
		truncate(RNEWS_FILE, main_rnews_size);
		main_rnews_size = -1;
	}
}

/* For not too harmful signals we can try to shut down neatly, but can't use
 * deblog, because may recurse */
static void main_signal_handler(int sig)
{
	static BOOL recursive_exit = FALSE;
	
	if (!recursive_exit)
	{
		recursive_exit = TRUE;
		main_normal_exit = FALSE;
		signal(SIGCHLD, SIG_IGN);
		fprintf(stderr, _("Exiting due to signal %d\n"), sig);
		fflush(stderr);
		exit(1);
	}
	else
		abort();
}

/* For more drastic signals just try to avoid corrupted rnews file at all costs
 */
static void main_drastic_handler(int sig)
{
	static BOOL recursive_exit = FALSE;
	
	if (!recursive_exit)
	{
		recursive_exit = TRUE;
		main_normal_exit = FALSE;
		signal(SIGCHLD, SIG_IGN);
		nscurses_emergency_stop();
		fprintf(stderr, _("Aborting due to signal %d\n"), sig);
		fflush(stderr);
	}
	main_repair_rnews_atexit();
	if (sig != SIGABRT)
		abort();
}

enum {
	sigint_NotReceived,
	sigint_Received,
	sigint_Handled
} main_got_sigint = sigint_NotReceived;

static void main_sigint_handler(int sig)
{
	static BOOL recursive = FALSE;
	UNUSED(sig);
	
	if (!recursive)
	{
		recursive = TRUE;
		signal(SIGINT, main_signal_handler);
		main_got_sigint = sigint_Received;
	}
}

/* Starts as many new threads as necessary to fill quota; returns TRUE if no
 * more can be started */
static BOOL main_start_new_threads(void)
{
	static BOOL upload_first = FALSE;
	BOOL all_done = TRUE;
	server_GFT_Result sgr = server_GFT_OK;

	if (main_got_sigint)
		return TRUE;
	while (kiddy_nthreads < maincf_get_max_threads() && sgr == server_GFT_OK)
	{
		ServerData *serv;
		NSThread *thread;

		switch (sgr = server_get_for_threading(&serv, &upload_first))
		{
			case server_GFT_OK:
				all_done = FALSE;
				if (upload_first)
					serv->uploading = TRUE;
				if ((thread = nsthread_new(serv, upload_first)) != NULL)
				{
#if ENABLE_ICONV
					const char *charset;
#endif

					++serv->nthreads;
					++kiddy_nthreads;
					upload_first = !upload_first;
#if ENABLE_ICONV
					if ((charset = maincf_get_convert_to_charset()) != NULL)
					{
						nsthread_send_command(thread,
								command_ToCharset, charset);
					}
					if ((charset = maincf_get_convert_from_charset()) != NULL)
					{
						nsthread_send_command(thread,
								command_FromCharset, charset);
					}
					if ((charset = maincf_get_ole_charset()) != NULL)
					{
						nsthread_send_command(thread,
								command_OLECharset, charset);
					}
					if ((charset = maincf_get_transfer_encoding()) != NULL)
					{
						nsthread_send_command(thread,
								command_TransferEncoding, charset);
					}
#endif
				}
				else
					return FALSE;
			case server_GFT_AllBusy:
				all_done = FALSE;
				break;
			default:
				break;
		}
	}
	return all_done;
}

static void main_got_message_id(NSThread *thread, const char *data)
{
	long int group_ix;
	long int artnum;
	char *msgid = XMALLOC(strlen(data));

	sscanf(data, "%ld %ld %s", &group_ix, &artnum, msgid);
	if (!group_add_message_id(thread->serv->group[group_ix], artnum, msgid) &&
			thread->serv->min_worth_getting &&
			++thread->serv->total_to_download ==
			thread->serv->min_worth_getting)
	{
		main_start_new_threads();
	}
	free(msgid);
}

static void main_hdl_parsed(const char *filename, const char *msgid);

static void main_headers_only(NSThread *thread, const char *msgid,
		const char *headers, long int score)
{
	char *filename;
	int fd;
	FILE *fp;

	fd = xmktmpindir(INCOMING_DIR, "Hdrs", &filename, FALSE,
			_("Can't create temporary file for killed message's headers"));
	if (fd == -1)
	{
		return;
	}
	fp = fdopen(fd, "w");
	if (!fp)
	{
		close(fd);
		deblog_critical(_("Can't write headers for killed message: %s\n"),
				strerror(errno));
		return;
	}
	if (thread->serv->path)
	{
		fprintf(fp, "Path: %s\n", thread->serv->path);
	}
	fprintf(fp, "%s\nX-Newsstar-Score: %ld\nX-Newsstar-Server: %s\n\n",
			headers, score, thread->serv->identifier);
	fclose(fp);
	main_hdl_parsed(filename, msgid);
	free(filename);
}

static BOOL main_check_xover(NSThread *thread, const char *current, int field,
		long int group_ix, long int artnum, char *all_headers)
{
	static int bad_count = 0;

	if (!current)
	{
		GroupData *group = thread->serv->group[group_ix];

		deblog_critical(_("Bad XOVER field %d from %s\n"),
				field, thread->serv->identifier);
		group->artnums[artnum - group->first_available].status
			= artnum_Unavailable;
		if (++bad_count > 10)
		{
			deblog_error(_("Too many bad XOVERs, giving up\n"));
		}
		free(all_headers);
		return FALSE;
	}
	return TRUE;
}

enum {
	Xover_Field_Subject,
	Xover_Field_From,
	Xover_Field_Date,
	Xover_Field_MessageId,
	Xover_Field_References,
	Xover_Field_ByteCount,
	Xover_Field_LineCount,
	Xover_Field_Newsgroups
};

static const char *Xover_Header[] = {
	"Subject: ", "From: ", "Date: ", "Message-ID: ",
	"References: ", "Size: ", "Lines: ", "Newsgroups: " 
};
#define XOVER_HEADER_SIZE \
	(sizeof("Newsgroups") + sizeof("Subject") + sizeof("From") + \
	 sizeof("Date") +  sizeof("Message-ID") + sizeof("References") \
	 + sizeof("Size") + sizeof("Lines") + 14)

static void main_got_xover(NSThread *thread, const char *data)
{
	long int group_ix;
	long int artnum;
	char *xover = XSTRDUP(data);
	char *current, *next;
	const char *msgid = NULL;
	int field;
	char *all_headers;
	const char *groupname;
	long int score = 0;
	long int field_end = thread->serv->xref_in_xover + 1;
	
	if (!field_end)
		field_end = Xover_Field_Newsgroups;
	sscanf(data, "%ld %ld", &group_ix, &artnum);
	groupname = thread->serv->group[group_ix]->name;
	deblog_debug(_("Processing XOVER for %s/%ld\n"), groupname, artnum);
	all_headers = XMALLOC(XOVER_HEADER_SIZE + strlen(data) + strlen(groupname));
	all_headers[0] = 0;
	/* Can't use sscanf to read rest of data, because some fields may be empty
	 * and sscanf combines sequential whitespace. Remember artnum\t precedes
	 * rest of fields */
	next = strchr(xover, '\t');
	if (!main_check_xover(thread, next, 0, group_ix, artnum, all_headers))
		return;
	for (field = 0; field != field_end; ++field)
	{
		current = next + 1;
		next = strchr(current, '\t');
		if (field != field_end - 1 && 
				!main_check_xover(thread, next, field, group_ix, artnum,
					all_headers))
		{
			return;
		}
		if (next)
			*next = 0;
		if (field == Xover_Field_MessageId)
		{
			msgid = current;
		}
		/* Add header name if necessary */
		if (field <= Xover_Field_LineCount
				|| (field == thread->serv->xref_in_xover &&
					(!thread->serv->xref_is_newsgroups ||
					 !thread->serv->xref_includes_name)))
		{
			if (all_headers[0])
				strcat(all_headers, "\n");
			strcat(all_headers, Xover_Header[(field ==
						thread->serv->xref_in_xover) ? Xover_Field_Newsgroups
					: field]);
		}
		/* Add header body if it can be used as is */
		if (field <= Xover_Field_LineCount
				|| (field == thread->serv->xref_in_xover &&
					thread->serv->xref_is_newsgroups))
		{
			strcat(all_headers, current);
			/*deblog_debug("!XOVER! %s%s\n", Xover_Header[field], current);*/
		}
		/* Convert Xref to Newsgroups format */
		if (field == thread->serv->xref_in_xover &&
				!thread->serv->xref_is_newsgroups)
		{
			BOOL first_group;

			current = thread->serv->xref_includes_name ?
				strchr(current, ' ') : current;
			if (current)
			{
				while (*current && isspace(*current))
					++current;
				/* Should now be pointing at server name */
				while (*current && !isspace(*current))
					++current;
				while (*current && isspace(*current))
					++current;
				/* Should now be pointing at first newsgroup, check */
				if (!*current)
					current = NULL;
			}
			if (!main_check_xover(thread, current, field, group_ix, artnum,
					all_headers))
			{
				return;
			}
			first_group = TRUE;
			while (*current)
			{
				char *colon;

				/* Search forwards for colon */
				if (!main_check_xover(thread, colon = strchr(current, ':'),
							field, group_ix, artnum, all_headers))
				{
					return;
				}
				*colon = 0;
				if (!first_group)
					strcat(all_headers, ",");
				first_group = FALSE;
				strcat(all_headers, current);
				current = colon + 1;
				while (*current && !isspace(*current))
					++current;
				while (*current && isspace(*current))
					++current;
			}
		}
	}
	if (thread->serv->xref_in_xover == -1)
	{
		char *newsgroups = NULL;

		asprintf(&newsgroups, "\n%s%s",
				Xover_Header[Xover_Field_Newsgroups], groupname);
		strcat(all_headers, newsgroups);
		free(newsgroups);
	}
	deblog_debug(_("%s/%ld has ID %s\n"), groupname, artnum, msgid);
	deblog_debug("score_have_master() = %d, server's scores = %p\n",
			score_have_master(), thread->serv->scores);
	/* deblog_debug(_("Isolated Message-ID %s from XOVER\n"), msgid); */
	if ((score_have_master() || thread->serv->scores)
			&& (score = thread->serv->default_score +
				score_calculate(thread->serv->scores, all_headers)) < 0)
	{
		deblog_debug(_("%s/%ld %s has scored %ld\n"), groupname, artnum, msgid,
				score);
		if (group_kill_message_id(thread->serv->group[group_ix], artnum, msgid))
		{
			deblog_debug(_("%s/%ld %s is duplicate anyway\n"), groupname,
					artnum, msgid);
		}
		else if (score > thread->serv->kill_score)
		{
			deblog_debug(_("Saving headers only for %s/%ld %s\n"), groupname,
					artnum, msgid);
			main_headers_only(thread, msgid, all_headers, score);
		}
		else
		{
			deblog_debug(_("Killfiling %s/%ld %s\n"), groupname, artnum, msgid);
		}
	}
	else
	if (!group_add_message_id(thread->serv->group[group_ix], artnum, msgid) &&
			thread->serv->min_worth_getting &&
			++thread->serv->total_to_download ==
			thread->serv->min_worth_getting)
	{
		main_start_new_threads();
	}
	free(xover);
	free(all_headers);
}

/* How long a child should throttle for if it overtakes list of available
 * articles */
#ifndef UNKNOWN_THROTTLE
#define UNKNOWN_THROTTLE 1000
#endif

inline static void upload_next(NSThread *thread)
{
	nsthread_send_command(thread, command_Upload,
			thread->serv->outgoing[thread->serv->nuploaded++]);
}

/* Returns TRUE if uploads are initiated */
static BOOL main_start_uploads(NSThread *thread)
{
	nsthread_send_command(thread, command_AllRequested, "");
	thread->serv->status = server_AllDone;
	deblog_info_thread(thread->thread_idx, _("All articles requested\n"));
	if (thread->serv->nuploaded < thread->serv->noutgoing)
	{
		thread->serv->uploading = TRUE;
		thread->all_downloaded_msg = _("Uploading");
		upload_next(thread);
		return TRUE;
	}
	return FALSE;
}

static void main_get_next_article(NSThread *thread, long int group_ix)
{
	long int art;
	const char *msgid;
	GroupData *group = thread->serv->group[group_ix];
	BackFillGroup *bfgroup = group_get_backfill(group, FALSE);
	BOOL backfill = FALSE;
	BackFillArticle *bfart = NULL;

	deblog_debug_thread(thread->thread_idx, _("Another article wanted\n"));
	deblog_debug_thread(thread->thread_idx,
			_("Current group for fetching articles is %s\n"), group->name);
	/* Do backfills first */
	while (bfgroup && bfgroup->current < bfgroup->vec_used_members &&
			(bfart = &bfgroup->vec_data[bfgroup->current],
			 bfart->just_added || bfart->artnum < 0))
	{
		++bfgroup->current;
	}
	if (bfart && bfgroup->current < bfgroup->vec_used_members)
	{
		backfill = TRUE;
		art = bfart->artnum;
		++bfgroup->current;
		deblog_debug_thread(thread->thread_idx,
				_("Requesting backfill article %ld\n"), art);
	}
	else
	{
		art = group_get_next_article(group, &msgid);
	}
	deblog_debug_thread(thread->thread_idx,
			_("Next article from %s is %ld%s\n"), group->name, art,
			backfill ? _(" (backfill)") : "");
	switch (art)
	{
		case 0:
			main_enter_new_group(thread);
			break;
		case -1:
			deblog_info_thread(thread->thread_idx,
					_("Article requests caught up with data, throttling\n"));
			nsthread_send_command(thread, command_Throttle, "%d",
					UNKNOWN_THROTTLE);
			break;
		default:
			deblog_debug_thread(thread->thread_idx,
					_("Ordering to get article\n"));
			nsthread_send_command(thread, command_GetArticle, "%ld %ld %d",
					group_ix, art, backfill);
			break;
	}
}

static void main_get_next_article_str(NSThread *thread, const char *data)
{
	long int group_ix;

	sscanf(data, "%ld", &group_ix);
	main_get_next_article(thread, group_ix);
}

typedef enum {
	show_status_GroupStats,
	show_status_Download
} show_status;

static void main_show_group_status(const NSThread *thread, long int group_ix,
		show_status ss)
{
	const char *sord = (ss == show_status_Download) ? _("<") : _(">");
	char *msg = XMALLOC(strlen(sord) + 2 +
				strlen(thread->serv->group[group_ix]->name));

	/*deblog_message_thread(thread->thread_idx,
	   ss ? _("Downloading from %s\n") :
			_("Examining %s\n"), thread->serv->group[group_ix]->name);
	*/
	if (ss)
	{
		deblog_message_thread(thread->thread_idx, _("Downloading from %s\n"),
				thread->serv->group[group_ix]->name);
	}
	else
	{
		deblog_info_thread(thread->thread_idx, _("Examining %s\n"),
				thread->serv->group[group_ix]->name);
	}
	sprintf(msg, _("%s %s"), sord, thread->serv->group[group_ix]->name);
	nscurses_show_status(thread->thread_idx, msg);
	free(msg);
}

static void main_enter_new_group(NSThread *thread)
{
	BOOL waiting = FALSE;
	long int old_group_ix = thread->group_ix;

	thread->wait_state = nsthread_NotWaiting;
	thread->group_ix = server_get_best_group_articles(thread->serv,
			thread->group_ix, &waiting);
	deblog_info_thread(thread->thread_idx,
			_("Asking child to change fetching group to %ld %s\n"),
			thread->group_ix, thread->group_ix == -1 ? _("<finished>") :
			thread->serv->group[thread->group_ix]->name);
	if (thread->group_ix == -1)
	{
		if (!main_start_uploads(thread))
		{
			if (waiting)
			{
				deblog_info(_("Will enter wait state when pipeline flushed\n"));
				thread->wait_state = nsthread_WillWait;
				nsthread_send_command(thread, command_WaitPending, "");
			}
			else
			{
				deblog_info(_("Will shutdown when pipeline flushed\n"));
				thread->all_downloaded_msg = _("Finished");
				nsthread_send_command(thread, command_Shutdown, "");
			}
		}
		else
		{
			deblog_info(_("Will upload when pipeline flushed\n"));
		}
	}
	else
	{
		GroupData *group = thread->serv->group[thread->group_ix];
		BackFillGroup *bfgroup = group_get_backfill(group, FALSE);

		/* Never need to --group->nthreads because a thread won't leave it
		 * until it's AllDone */
		++group->nthreads;
		if (bfgroup && bfgroup->vec_used_members && !bfgroup->current)
		{
			deblog_info_thread(thread->thread_idx, _("%s has %ld backfills\n"),
					group->name, bfgroup->vec_used_members);
		}
		if (thread->group_ix == old_group_ix)
		{
			deblog_info_thread(thread->thread_idx,
				_("Group already selected\n"));
			main_show_group_status(thread, thread->group_ix,
					show_status_Download);
			main_get_next_article(thread, thread->group_ix);
		}
		else
		{
			nsthread_send_command(thread, command_EnterGroup, "%d %s",
					thread->group_ix, group->name);
		}
	}
}

static void main_wake_up_waiters(void)
{
	int ti;

	for (ti = 0; ti < maincf_get_max_threads(); ++ti)
	{
		NSThread *thread = nsthread_lookup(ti);

		if (!thread)
			continue;
		switch (thread->wait_state)
		{
			case nsthread_WillWait:
				thread->wait_state = nsthread_WakeUp;
				deblog_info_thread(ti,
				 _("Will look for serviceable groups when pipeline flushed\n"));
				break;
			case nsthread_Waiting:
				deblog_info_thread(ti,
						_("Waking up to download more articles\n"));
				main_enter_new_group(thread);
				break;
			default:
				/* Ignore other states */
				break;
		}
	}
}

static void main_group_complete(NSThread *thread, const char *data, BOOL prune,
		BOOL bogus)
{
	long int group_ix;
	GroupData *group;

	sscanf(data, "%ld", &group_ix);
	group = thread->serv->group[group_ix];
	if (bogus)
	{
		 group->status = group_Bogus;
	}
	else
	{
		groupcount_server_done(group->gc);
	}
	if (prune)
	{
		deblog_debug_thread(thread->thread_idx,
				_("Pruning unavailable articles from %s\n"),
				thread->serv->group[group_ix]->name);
		group_prune_unavailable(group);
		main_wake_up_waiters();
	}
	group_ix = server_get_next_group_stats(thread->serv);
	deblog_info_thread(thread->thread_idx,
			_("Now getting stats for group %ld\n"), group_ix);
	if (group_ix != -1)
	{
		main_show_group_status(thread, group_ix, show_status_GroupStats);
		thread->group_ix = group_ix;
		nsthread_send_command(thread, command_GetGroupStats, "%d %s",
				group_ix, thread->serv->group[group_ix]->name);
	}
	else
	{
		nscurses_show_status(thread->thread_idx, _("Downloading articles"));
		deblog_message_thread(thread->thread_idx, _("Downloading articles\n"));
		main_enter_new_group(thread);
	}
}

static void main_got_group_stats(NSThread *thread, const char *data)
{
	long int group_ix, first, last;
	GroupData *group;
	BackFillGroup *bfgroup;

	sscanf(data, "%ld %ld %ld", &group_ix, &first, &last);
	group = thread->serv->group[group_ix];
	if (group_got_stats(group, first, last, thread->thread_idx))
	{
		nsthread_send_command(thread, thread->serv->xover ? command_GetXover :
				command_GetArticles, "%ld %d-%d", group_ix,
				group->first_available, last);
		deblog_debug("Getting articles %d-%d for %s/%s\n",
				group->first_available, last,
				thread->serv->identifier, group->name);
	}
	else
	{
		deblog_debug("No articles to download from %s/%s\n",
				thread->serv->identifier, group->name);
		main_group_complete(thread, data, FALSE, FALSE);
	}

	bfgroup = group_get_backfill(group, FALSE);
	if (bfgroup)
	{
		deblog_debug("Removing backfills below artnum %ld\n", first);
		backfill_remove_old(bfgroup, first);
	}
}

/* Adds article read by in_fb to rnews batch file; uses filename for message
 * and to read size with stat, and to delete it */
static void main_hdl_rnews(rw_buf *in_fb, const char *filename)
{
	struct stat fdata;
	int fd;
	static rw_buf *out_fb = NULL;
	BOOL result = FALSE;

	if (!out_fb)
	{
		deblog_info(_("Opening rnews file %s\n"), RNEWS_FILE);
		result = ((fd = open(RNEWS_FILE, O_WRONLY | O_CREAT | O_APPEND, 0644))
				< 0);
		if (!result)
			result = ((out_fb = rw_create_buf(fd, FALSE, FALSE, -1)) == NULL);
		if (result)
			deblog_error(_("Couldn't open rnews file: %s\n"), strerror(errno));
	}
	/* Remember size of rnews file in case copy fails and it needs repairing */
	main_rnews_size = stat(RNEWS_FILE, &fdata) ? 0 : fdata.st_size;
	deblog_debug(_("Adding %s to rnews file\n"), filename);
	if (!result)
		result = (stat(filename, &fdata) != 0);
	if (!result)
	{
		result = (rw_writef(out_fb, "#! rnews %08d\n", fdata.st_size) <= 0);
	}
	if (!result)
		result = (rw_copy(in_fb, out_fb, NULL, NULL) != 0);
	if (result)
	{
		int e = errno;

		truncate(RNEWS_FILE, main_rnews_size);
		main_rnews_size = -1;
		deblog_error(_("Couldn't add article to rnews file: %s\n"),
				strerror(e));
	}
	main_rnews_size = -1;
	/* XREMOVE(filename); */
}

/* If INN has already seen a Message_ID with CHECK, we must use TAKETHIS to
 * post it, otherwise IHAVE */
static BOOL main_use_takethis = FALSE;

/* Posts article read by fbuf; returns TRUE if rejected */
static BOOL main_hdl_ihave(rw_buf *fbuf, const char *msgid, const char
		*filename)
{
	BOOL result;
	UNUSED(filename);

	rw_writef(local_buf, "%s %s\r\n",
			main_use_takethis ? "TAKETHIS" : "IHAVE", msgid);
	if (!main_use_takethis && !nntp_check_response(local_buf, "localhost", 
				_("Article rejected before sending"), "3",
				deblog_Level_Critical))
	{
		return TRUE;
	}
	if (rw_ftos(fbuf, local_buf, NULL, NULL))
	{
		deblog_error(_("Couldn't post article to local server: %s\n"),
				strerror(errno));
	}
	/* XREMOVE(filename); */
	/* The reason this is Level_Message instead of Warning is because we don't
	 * want it printed in quiet mode */
	result = nntp_check_response(local_buf, "localhost", 
			_("Article rejected after sending"), "2", deblog_Level_Message)
			== NULL;
	return result;
}

static void main_hdl_batch(const char *msgid, const char *filename)
{
	static FILE *fp = NULL;

	if (!fp)
	{
		deblog_info(_("Opening batch file %s\n"), BATCH_FILE);
		fp = fopen(BATCH_FILE, "a");
		if (!fp)
			deblog_error(_("Couldn't open batch file: %s\n"), strerror(errno));
	}
	if ((size_t) fprintf(fp, "%s\t%s\n", filename, msgid) != strlen(filename) +
			strlen(msgid) + 2)
	{
		deblog_error(_("Error writing batch file: %s\n"), strerror(errno));
	}
}

static void main_hdl_snstore(rw_buf *in_fb, const char *filename)
{
	static rw_buf *snstr_buf = NULL;

	if (!snstr_buf)
	{
		deblog_info(_("Opening pipe to snstore\n"));
		if ((snstr_buf = snstore_init(main_snstore)) == NULL)
			deblog_error(_("Couldn't open pipe to snstore\n"));
	}
	if (rw_ftos(in_fb, snstr_buf, NULL, NULL))
	{
		deblog_error(_("Couldn't send %s to snstore\n"), filename);
	}
	/* XREMOVE(filename); */
}

static void main_article_unavailable(NSThread *thread, const char *data)
{
	long int group_ix;
	long int artnum;

	sscanf(data, "%ld %ld", &group_ix, &artnum);
	/* As far as group/article status is concerned, we can treat this as if it
	 * were downloaded successfully */
	group_have_downloaded(thread->serv->group[group_ix], artnum);
	/*group_add_backfill(thread->serv->group[group_ix], artnum, FALSE);*/
	stats_dec_total_to_fetch();
}

static void main_hdl_parsed(const char *filename, const char *msgid)
{
	int fd;
	rw_buf *in_fb = NULL;
	BOOL result = FALSE;

	if (main_post_method != PM_batch)
	{
		result = ((fd = open(filename, O_RDONLY)) < 0);
		if (!result)
			result = ((in_fb = rw_create_buf(fd, FALSE, FALSE, -1)) == NULL);
		if (result)
			deblog_error(_("Couldn't open article file: %s\n"),
					strerror(errno));
	}
	result = FALSE;
	switch (main_post_method)
	{
		case PM_nntp:
			result = main_hdl_ihave(in_fb, msgid, filename);
			break;
		case PM_rnews:
			main_hdl_rnews(in_fb, filename);
			break;
		case PM_batch:
			main_hdl_batch(msgid, filename);
			break;
		case PM_snstore:
			main_hdl_snstore(in_fb, filename);
			break;
	}
	if (main_post_method != PM_batch)
	{
		rw_close_and_free(in_fb);
		XREMOVE(filename);
	}
}

static void main_have_downloaded(NSThread *thread, const char *data)
{
	long int group_ix;
	long int artnum;
	const char *msgid = NULL;
	char *bf_msgid = NULL;
	BOOL backfill;
	char *filename = XMALLOC(strlen(data));
	GroupData *group;
	BOOL bf_dup = FALSE;

	sscanf(data, "%ld %ld %d %s", &group_ix, &artnum, &backfill, filename);
	group = thread->serv->group[group_ix];
	if (backfill)
	{
		int fd;
		rw_buf *in_fb = NULL;
		BOOL result = FALSE;
		const char *line = NULL;

		result = ((fd = open(filename, O_RDONLY)) < 0);
		if (!result)
			result = ((in_fb = rw_create_buf(fd, FALSE, FALSE, -1)) == NULL);
		if (result)
			deblog_error(_("Couldn't open article file: %s\n"),
					strerror(errno));
		while (line = rw_read_line(in_fb),
				line && strncmp(line, "Message-I", 9));
		if (line)
		{
			bf_msgid = XSTRDUP(line + 12);
			msgid = bf_msgid;
		}
		else
		{
			deblog_critical(
					_("No Message-ID in downloaded backfill article\n"));
			msgid = "";
		}
		rw_close_and_free(in_fb);
		if (hashtree_lookup(the_hashtree, msgid) || history_check(msgid))
			bf_dup = TRUE;
	}
	else
	{
		msgid = group->artnums[artnum - group->first_available].msgid->msgid;
	}
	if (bf_dup)
	{
		XREMOVE(filename);
	}
	else
	{
		main_hdl_parsed(filename, msgid);
	}
	free(filename);
	/* Ignore result, behave as if fetch was successful */
	if (backfill)
	{
		backfill_remove_available(group->bfgroup, artnum);
		stats_inc_backfilled();
		free(bf_msgid);
	}
	else
	{
		group_have_downloaded(group, artnum);
		stats_inc_fetched();
	}
}

static void main_start_fetching(NSThread *thread)
{
	long int group_ix = server_get_next_group_stats(thread->serv);

	deblog_message_thread(thread->thread_idx,
			_("Getting list of articles to download\n"));
	deblog_debug_thread(thread->thread_idx,
			_("Ready to fetch, entering group %ld\n"), group_ix);
	if (group_ix == -1)
	{
		if (!thread->serv->ngroups)
		{
			deblog_message_thread(thread->thread_idx,
					_("No groups to fetch, shutting down thread\n"));
			nsthread_send_command(thread, command_Shutdown, "");
			return;
		}
		nscurses_show_status(thread->thread_idx, _("Downloading articles"));
		deblog_message_thread(thread->thread_idx, _("Downloading articles\n"));
		main_enter_new_group(thread);
	}
	else
	{
		thread->group_ix = group_ix;
		main_show_group_status(thread, group_ix, show_status_GroupStats);
		nsthread_send_command(thread, command_GetGroupStats,
				"%d %s", group_ix, thread->serv->group[group_ix]->name);
	}
}

static void main_upload_next(NSThread *thread)
{
	if (main_got_sigint)
		return;
	if (thread->serv->noutgoing > thread->serv->nuploaded)
	{
		upload_next(thread);
	}
	else
	{
		if (thread->serv->locked_upload)
		{
			deblog_info_thread(thread->thread_idx,
					_("Upload(s) locked, rerunning queue\n"));
			thread->serv->locked_upload = FALSE;
			thread->serv->nuploaded = 0;
			/* Avoid thrashing */
			if (clock() - thread->serv->locked_at < CLOCKS_PER_SEC)
				sleep(1);
			upload_next(thread);
		}
		deblog_message_thread(thread->thread_idx, _("Uploads completed\n"));
		if (thread->upload_first)
		{
			main_start_fetching(thread);
		}
		else
		{
			nscurses_show_status(thread->thread_idx, _("Finished"));
			deblog_info_thread(thread->thread_idx, _("Shutting down\n"));
			nsthread_send_command(thread, command_Shutdown, "");
		}
	}
}

static void main_postamble(NSThread *thread)
{
	if ((thread->upload_first  || !thread->serv->ngroups) &&
			thread->serv->noutgoing > thread->serv->nuploaded)
	{
		thread->serv->uploading = TRUE;
		nscurses_show_status(thread->thread_idx, _("Uploading"));
		deblog_message_thread(thread->thread_idx,
				_("Uploading articles\n"));
		upload_next(thread);
	}
	else
	{
		main_start_fetching(thread);
	}
}

static void main_found_xref_in_xover(NSThread *thread, const char *data)
{
	sscanf(data, "%ld %d %d", &thread->serv->xref_in_xover,
			&thread->serv->xref_includes_name,
			&thread->serv->xref_is_newsgroups);
}

static void main_start_waiting(NSThread *thread)
{
	switch (thread->wait_state)
	{
		case nsthread_NotWaiting:
		case nsthread_Waiting:
			deblog_info_thread(thread->thread_idx,
					_("Being asked to enter wait state when it shouldn't\n"));
		case nsthread_WakeUp:
			main_enter_new_group(thread);
			break;
		case nsthread_WillWait:
			thread->wait_state = nsthread_Waiting;
			nscurses_show_status(thread->thread_idx, _("Waiting"));
			break;
	}
}
	
static void main_shutdown_all(void)
{
	NSThread *thread = NULL;

	for (thread = kiddy_get_next_thread(NULL); thread;
			thread = kiddy_get_next_thread(thread))
	{
		nsthread_send_command(thread, command_Shutdown, "");
	}
}

static void parse_bad_upload(const char *cmd_in)
{
	char *pathname = XSTRDUP(cmd_in);
	char *server = strchr(pathname, ' ');
	char *reason = NULL;

	deblog_debug("Parsing BadUpload response: %s\n", cmd_in);
	if (server)
	{
		*(server++) = 0;
		reason = strchr(server, ' ');
	}
	if (reason)
		*(reason++) = 0;
	deblog_debug("Pathname: %s, Server: %s, Reason: %s\n",
			pathname, server, reason);
	uploads_bad(pathname, server, reason);
	free(pathname);
}

static void main_loop(void)
{
	const intercom_data *command;
	intercom_Command command_code = command_Null;
	NSThread *thread;
	struct timeval tv;
	long int bytes;
	long int group_ix;
	int level;
	
	for (;;)
	{
		command = kiddy_get_command();
		command_code = command->command;
		if (command->thread_idx >= maincf_get_max_threads())
		{
			deblog_info(_("Ignoring message from extraneous thread %d\n"),
					command->thread_idx);
			continue;
		}
		thread = nsthread_lookup(command->thread_idx);
		if (!thread)
		{
			char term = (command->data[strlen(command->data) - 1] == '\n')
				? 0 : '\n';

			deblog_info(_("Command %d received for NULL thread %d: %s%c"),
					command_code, command->thread_idx, command->data, term);
			continue;
		}
		/* Append args to any previous in case command was split */
		thread->command = XSTRDUPCAT(thread->command, command->data);
		/* If a SIGINT shutdown is in process, we ignore certain commands */
		if (main_got_sigint == sigint_Handled)
		{
			switch (command_code)
			{
				case command_Ready:
				case command_GroupArticleStats:
				case command_BogusGroup:
				case command_GroupComplete:
				case command_GetNextArticle:
				case command_WaitAck:
					command_code = command_Null;
					break;
				default:
					break;
			}
		}
		switch (command_code)
		{
			case command_Null:
				break;
			case command_Resolved:
				thread->serv->addr_s = XMALLOC(strlen(thread->command));
				sscanf(thread->command, "%s %d",
						thread->serv->addr_s, &thread->serv->proto);
				thread->serv->status = server_Resolved;
				if (!main_got_sigint)
				{
					nscurses_show_status(thread->thread_idx, _("Connecting"));
					deblog_message_thread(thread->thread_idx,
							_("Connecting to %s\n"), thread->serv->identifier);
					nsthread_send_command(thread, command_Connect,
							"%s %s %d %s %d %d %d", 
							thread->serv->identifier,
							thread->serv->addr_s,
							thread->serv->proto,
							thread->serv->path ? thread->serv->path : "!",
							thread->serv->pipeline,
							thread->serv->ignore_436,
							thread->serv->ssl);
					thread->serv->status = thread->serv->status ==
						server_Resolved ? server_Connecting : server_Connected;
					/* Try to start more threads in case we were blocked on
					 * missing info */
					main_start_new_threads();
				}
				break;
			case command_Log:
				level = atoi(thread->command);
				deblog_print_thread(command->thread_idx, level,
						strchr(thread->command, ' ') + 1);
				if (level == deblog_Level_Critical)
				{
					nscurses_show_status(thread->thread_idx, _("Error"));
				}
				break;
			case command_Connected:
				thread->serv->status = server_Connected;
				if (thread->serv->user && !main_got_sigint)
				{
					nscurses_show_status(thread->thread_idx,
							_("Authenticating user"));
					deblog_message_thread(thread->thread_idx, 
							_("Authenticating user\n"));
					nsthread_send_command(thread, command_LogOn, "%s %s",
							thread->serv->user, thread->serv->password);
					break;
				}
				/* else fall through */
			case command_Authenticated:
				if (thread->serv->mode_reader && !main_got_sigint)
				{
					nscurses_show_status(thread->thread_idx,
							_("Setting mode reader"));
					deblog_message_thread(thread->thread_idx,
							_("Setting mode to reader\n"));
					nsthread_send_command(thread, command_ModeReader, "");
					break;
				}
				/* else fall through */
			case command_Ready:
				if (!thread->serv->connx_msg)
				{
					deblog_brief(_("Connected to %s\n"),
							thread->serv->identifier);
					thread->serv->connx_msg = TRUE;
				}
				if (score_have_master() || thread->serv->scores)
				{
					nsthread_send_command(thread, command_FindXrefInXover, "");
				}
				else
				{
					main_postamble(thread);
				}
				break;
			case command_FoundXrefInXover:
				main_found_xref_in_xover(thread, thread->command);
				main_postamble(thread);
				break;
			case command_GroupArticleStats:
				main_got_group_stats(thread, thread->command);
				/* Try to start more threads in case we were blocked on missing
				 * info */
				main_start_new_threads();
				break;
			case command_GroupContains:
				main_got_message_id(thread, thread->command);
				break;
			case command_XoverData:
				main_got_xover(thread, thread->command);
				break;
			case command_BogusGroup:
				main_group_complete(thread, thread->command, FALSE, TRUE);
				break;
			case command_GroupComplete:
				main_group_complete(thread, thread->command, TRUE, FALSE);
				break;
			case command_GetNextArticle:
				main_get_next_article_str(thread, thread->command);
				break;
			case command_HaveDownloaded:
				main_have_downloaded(thread, thread->command);
				break;
			case command_ArticleUnavailable:
				main_article_unavailable(thread, thread->command);
				break;
			case command_Error:
				nscurses_show_status(thread->thread_idx, _("Error"));
				deblog_critical(thread->command);
				/* NB fall through to next clause */
			case command_QuietError:
				server_mark_error(thread->serv);
				/* NB fall through to next clause */
			case command_AllDone:
				nsthread_delete(thread);
				thread = NULL;
				--kiddy_nthreads;
				deblog_debug(_("%d threads remaining\n"), kiddy_nthreads);
				if (main_start_new_threads() && ! kiddy_nthreads)
				{
					return;
				}
				break;
			case command_AmountSent:
				sscanf(thread->command, "%ld %ld %ld", &bytes, &tv.tv_sec,
						&tv.tv_usec);
				stats_update_bytes_sent(bytes, &tv);
				break;
			case command_AmountRcvd:
				sscanf(thread->command, "%ld %ld %ld", &bytes, &tv.tv_sec,
						&tv.tv_usec);
				stats_update_bytes_rcvd(bytes, &tv);
				break;

			case command_Uploaded:
				stats_inc_uploaded();
				uploads_good(thread->command);
				main_upload_next(thread);
				break;
			case command_RemoteDuplicate:
				stats_dec_total_to_upload();
				uploads_good(thread->command);
				stats_inc_duplicates();
				main_upload_next(thread);
				break;
			case command_LockedDuplicate:
				if (!thread->serv->locked_upload)
				{
					thread->serv->locked_upload = TRUE;
					thread->serv->locked_at = clock();
				}
				/* Fall-through */
			case command_LocalDuplicate:
				main_upload_next(thread);
				break;
			case command_BadUpload:
				stats_inc_bad_uploads();
				parse_bad_upload(thread->command);
				main_upload_next(thread);
				break;
			case command_NoUploads:
				thread->serv->nuploaded = thread->serv->noutgoing;
				main_upload_next(thread);
				break;
			case command_EnteredGroup:
				sscanf(thread->command, "%ld", &group_ix);
				main_show_group_status(thread, group_ix, show_status_Download);
				break;
			case command_AllDownloaded:
				if (thread->all_downloaded_msg)
				{
					nscurses_show_status(thread->thread_idx,
							thread->all_downloaded_msg);
					deblog_message_thread(thread->thread_idx, "%s\n",
							thread->all_downloaded_msg);
					thread->all_downloaded_msg = NULL;
				}
				break;
			case command_WaitAck:
				main_start_waiting(thread);
				break;
			case command_SplitCommand:
				/* Do nothing, we've already saved initial portion of command */
				break;
			default:
				deblog_error(_("Parent doesn't understand command code %d\n"),
						command_code);
				break;
		}
		if (command_code != command_SplitCommand && thread)
		{
			free(thread->command);
			thread->command = NULL;
		}
		if (main_got_sigint == sigint_Received)
		{
			deblog_message(
					_("Received user interrupt, shutting down cleanly\n"));
			main_shutdown_all();
			main_got_sigint = sigint_Handled;
		}
	}
}

int main(int argc, char **argv)
{
	int arg;
	BOOL use_curses = FALSE;
	const char *servname = NULL;
	time_t now = time(NULL);
	BOOL use_snews = FALSE;

#if USE_PTHREADS
	main_thread = pthread_self();
#endif
	deblog_info("\n");
	deblog_info("\n");
	deblog_info(_("Starting run at %s\n"), ctime(&now));
	if (argc > 7)
		deblog_warning(_("Too many arguments\n"));
	for (arg = 1; arg < argc; ++arg)
	{
		if (!strcmp(argv[arg], "-v") || !strcmp(argv[arg], "-vv"))
		{
			deblog_enable_level(deblog_Level_Info);
			if (argv[arg][2] == 'v')
				deblog_enable_level(deblog_Level_Debug);
		}
		else if (!strcmp(argv[arg], "-s"))
		{
			deblog_separate_sinks();
		}
		else if (!strcmp(argv[arg], "-b"))
		{
			if (use_curses)
			{
				deblog_warning(_("Disabling -b option because using -f\n"));
			}
			else
			{
				deblog_be_brief = TRUE;
			}
		}
		else if (!strcmp(argv[arg], "-f"))
		{
#ifdef HAVE_CURSES_H
			use_curses = TRUE;
			if (deblog_be_brief)
			{
				deblog_be_brief = FALSE;
				deblog_warning(_("Disabling -b option because using -f\n"));
			}
			deblog_info(_("Full-screen option (-f) requested\n"));
#else
			deblog_warning(_("Full-screen option (-f) not supported\n"));
#endif
		}
		else if (argv[arg][0] == '-' && argv[arg][1] == 'o')
		{
			servname = argv[arg] + 2;
		}
		else if (!strcmp(argv[arg], "-sn"))
		{
			maincf_use_sn();
		}
		else if (!strcmp(argv[arg], "-snews"))
		{
			maincf_use_snews();
			use_snews = TRUE;
		}
		/* Since this is only supposed to be run from newsstar.pl, don't bother
		 * to check for -sn and -snews clash here, leave it up to the script */
		else if (!strncmp(argv[arg], "--snstore=", sizeof("--snstore=") - 1))
		{
			main_snstore = XSTRDUP(argv[arg] + sizeof("--snstore=") - 1);
		}
		else if (!strcmp(argv[arg], "-q"))
		{
			stats_quiet();
		}
		else if (!strcmp(argv[arg], "-r"))
		{
			histgdbm_use_reorganize(TRUE);
		}
		else
		{
			deblog_warning(_("Unknown option '%s'\n"), argv[arg]);
		}
	}
	deblog_message(_("%s version %s\n"), PACKAGE, VERSION);
	today_init();
	local_config_init();
	maincf_init();
	main_use_takethis = FALSE;
	if (!maincf_get_history_method() ||
			!strcmp(maincf_get_history_method(), "gdbm"))
	{
#ifdef HAVE_GDBM_H
		histgdbm_init();
#else
		if (maincf_get_history_method())
			deblog_error(_("gdbm history method not supported\n"));
#endif
	}
	else if (maincf_get_history_method())
	{
		if (!strcmp(maincf_get_history_method(), "check"))
		{
			histcheck_init();
			main_use_takethis = TRUE;
			if (nntp_set_mode(local_buf, "localhost", "STREAM"))
			{
				deblog_error(
					_("Unable to set local connection to streaming mode\n"));
			}
		}
		else if (!strcmp(maincf_get_history_method(), "snews"))
		{
#ifdef HAVE_GDBM_H
			histsnews_init();
#else
			deblog_error(_("snews history method not supported: no libgdbm\n"));
#endif
		}
		else if (strcmp(maincf_get_history_method(), "none"))
			deblog_error(_("history_method '%s' not known\n"));
	}
	kiddy_spawn_all();
	if (maincf_get_post_method())
	{
		if (!strcmp(maincf_get_post_method(), "nntp"))
		{
			main_post_method = PM_nntp;
			local_init();
		}
		else if (!strcmp(maincf_get_post_method(), "rnews"))
		{
			main_post_method = PM_rnews;
		}
		else if (!strcmp(maincf_get_post_method(), "batch"))
		{
			main_post_method = PM_batch;
		}
		else if (!strcmp(maincf_get_post_method(), "snstore"))
		{
			if (!main_snstore || !maincf_get_using_sn())
			{
				deblog_error(_("post_method is snstore, but -sn or --snstore"
							" option is missing\n"));
			}
			main_post_method = PM_snstore;
		}
		else
		{
			deblog_error(_("post_method '%s' not known\n"));
		}
		if (use_snews && main_post_method != PM_nntp)
		{
			deblog_error(_("Post method '%s' is incompatible with s-news\n"),
					maincf_get_post_method());
		}
	}
	else
	{
		main_post_method = use_snews ? PM_nntp : PM_rnews;
	}
	atexit(main_repair_rnews_atexit);
	signal(SIGHUP, main_signal_handler);
	signal(SIGINT, main_sigint_handler);
	signal(SIGPIPE, main_signal_handler);
	signal(SIGTERM, main_signal_handler);
	signal(SIGQUIT, main_drastic_handler);
#ifndef DONT_TRAP_NASTY_SIGNALS
	signal(SIGILL, main_drastic_handler);
	signal(SIGFPE, main_drastic_handler);
	signal(SIGSEGV, main_drastic_handler);
#endif
	signal(SIGABRT, main_drastic_handler);
	signal(SIGALRM, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
	signal(SIGUSR2, SIG_IGN);
	server_scan_rc_dir(servname);
	the_hashtree = hashtree_new();
	stats_start();
	if (use_curses)
	{
		nscurses_start(maincf_get_max_threads());
	}
	if (main_start_new_threads())
		return 0;
	main_loop();
	stats_update(TRUE);
	deblog_disable_stats_update();
	uploads_move_any_bad();
	deblog_message(_("News fetch complete\n"));
	if (use_curses && nscurses_active())
	{
		nscurses_stop();
	}
	stats_summary();

	main_normal_exit = TRUE;
	deblog_info(_("Clean exit\n"));
	return 0;
}
