/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"
#include "parserc.h"

#ifndef MAXTHREADS
#define MAXTHREADS 3
#endif

static BOOL maincf_using_sn = FALSE;
static BOOL maincf_using_snews = FALSE;

static struct {
    long int max_threads;
    long int threads_per_server;
    BOOL pipeline;
    char *path;
    char *history_method;
    char *post_method;
    long int gdbm_expiry;
    BOOL gdbm_fast;
    char *spool_dir;
    char *outgoing_dir;
    char *snews_history;
    BOOL regulate_updates;
    BOOL regex_ic;
    BOOL report_dups;
    BOOL local_ssl;
#if ENABLE_ICONV
    char *convert_to_charset;
    char *convert_from_charset;
    char *ole_charset;
    char *transfer_encoding;
#endif
} maincf_data;

/*
max_threads			 Absolute maximum number of threads to run
threads_per_server		 Maximum number of threads for any one server
*/

void maincf_init(void)
{
    maincf_data.max_threads = maincf_data.threads_per_server = MAXTHREADS;
    maincf_data.path = maincf_data.history_method = maincf_data.post_method =
	NULL;
    maincf_data.pipeline = TRUE;
    maincf_data.gdbm_expiry = 7;
    maincf_data.gdbm_fast = FALSE;
    maincf_data.report_dups = FALSE;
    maincf_data.regulate_updates = TRUE;
    maincf_data.regex_ic = TRUE;
    maincf_data.local_ssl = FALSE;
    maincf_data.snews_history = maincf_data.spool_dir =
	maincf_data.outgoing_dir = NULL;
#if ENABLE_ICONV
    maincf_data.convert_to_charset = maincf_data.convert_from_charset =
	maincf_data.ole_charset = maincf_data.transfer_encoding = NULL;
#endif
    deblog_info(_("Loading main config file\n"));
    
    parserc_register_int("max_threads", &maincf_data.max_threads);
    parserc_register_int("threads_per_server",
	    &maincf_data.threads_per_server);
    parserc_register_bool("pipeline", &maincf_data.pipeline);
    parserc_register_string("path", &maincf_data.path);
    parserc_register_string("history_method", &maincf_data.history_method);
    parserc_register_string("post_method", &maincf_data.post_method);
    parserc_register_int("gdbm_expiry", &maincf_data.gdbm_expiry);
    parserc_register_bool("gdbm_fast", &maincf_data.gdbm_fast);
    parserc_register_ignored("slow_hist");
    parserc_register_bool("regulate_updates",
	    &maincf_data.regulate_updates);
    parserc_register_bool("regex_ic", &maincf_data.regex_ic);
    parserc_register_string("spool_dir", &maincf_data.spool_dir);
    parserc_register_string("outgoing_dir", &maincf_data.outgoing_dir);
    parserc_register_string("snews_history", &maincf_data.snews_history);
    parserc_register_ignored("active_file");
    parserc_register_ignored("articles_dir");
    parserc_register_bool("report_dups", &maincf_data.report_dups);
    parserc_register_bool("local_ssl", &maincf_data.local_ssl);
#if ENABLE_ICONV
    parserc_register_string("convert_to_charset",
	    &maincf_data.convert_to_charset);
    parserc_register_string("convert_from_charset",
	    &maincf_data.convert_from_charset);
    parserc_register_string("ole_charset",
	    &maincf_data.ole_charset);
    parserc_register_string("transfer_encoding",
	    &maincf_data.transfer_encoding);
#endif

    parserc_read_named(MAIN_CONF, deblog_Level_Warning);

    if (maincf_data.max_threads < 1)
    {
	deblog_warning(_("Too few max threads (%d), changing to 1\n"),
		maincf_data.max_threads);
	maincf_data.max_threads = 1;
    }
    if (maincf_data.threads_per_server < 1)
    {
	deblog_warning(_("Too few threads per server (%d), changing to 1\n"),
		maincf_data.threads_per_server);
	maincf_data.threads_per_server = 1;
    }
    if (maincf_data.threads_per_server > maincf_data.max_threads)
    {
	deblog_warning(_("Threads per server (%d) greater than max threads (%d)"
		    ", truncating\n"), maincf_data.threads_per_server,
		    maincf_data.max_threads);
	maincf_data.threads_per_server = maincf_data.max_threads;
    }

    deblog_debug(_("max_threads = %d\n"), maincf_data.max_threads);
    deblog_debug(_("threads_per_server = %d\n"),
	    maincf_data.threads_per_server);
    if (maincf_using_sn)
    {
	if (!maincf_data.spool_dir)
	{
	    maincf_data.spool_dir = XSTRDUP("/var/spool/sn");
	}
	if (!maincf_data.outgoing_dir)
	{
	    maincf_data.outgoing_dir = XSTRDUPCAT(maincf_data.spool_dir,
		    "/.outgoing");
	}
    }
    else
    {
	if (!maincf_data.spool_dir)
	{
	    maincf_data.spool_dir = XSTRDUP("/var/spool/news");
	}
	if (!maincf_data.outgoing_dir)
	{
	    maincf_data.outgoing_dir = maincf_using_snews ?
		XSTRDUP("/var/lib/news/suck") :
		XSTRDUPCAT(maincf_data.spool_dir, "/out.going");
	}
    }
    if (!maincf_data.snews_history)
    {
	maincf_data.snews_history = XSTRDUP("/var/lib/news/history");
    }
#if ENABLE_ICONV
    if (!maincf_data.ole_charset)
	maincf_data.transfer_encoding = XSTRDUP("8bit");
#endif
}

long int maincf_get_threads_per_server(void)
{
    return maincf_data.threads_per_server;
}

long int maincf_get_max_threads(void)
{
    return maincf_data.max_threads;
}

BOOL maincf_get_pipeline(void)
{
    return maincf_data.pipeline;
}

const char *maincf_get_path(void)
{
    return maincf_data.path;
}

const char *maincf_get_history_method(void)
{
    return maincf_data.history_method;
}

const char *maincf_get_post_method(void)
{
    return maincf_data.post_method;
}

void maincf_reset_max_threads(long int mt)
{
    maincf_data.max_threads = mt;
}

long int maincf_get_gdbm_expiry(void)
{
    return maincf_data.gdbm_expiry;
}

BOOL maincf_get_gdbm_fast(void)
{
    return maincf_data.gdbm_fast;
}

const char *maincf_get_spool_dir(void)
{
    return maincf_data.spool_dir;
}

const char *maincf_get_outgoing_dir(void)
{
    return maincf_data.outgoing_dir;
}

BOOL maincf_get_using_sn(void)
{
    return maincf_using_sn;
}

void maincf_use_sn(void)
{
    maincf_using_sn = TRUE;
}

void maincf_use_snews(void)
{
    maincf_using_snews = TRUE;
}

const char *maincf_get_snews_history(void)
{
    return maincf_data.snews_history;
}

BOOL maincf_get_regulate_updates(void)
{
    return maincf_data.regulate_updates;
}

BOOL maincf_get_regex_ic(void)
{
    return maincf_data.regex_ic;
}

BOOL maincf_get_report_dups(void)
{
    return maincf_data.report_dups;
}

BOOL maincf_get_local_ssl(void)
{
    return maincf_data.local_ssl;
}

#if ENABLE_ICONV
const char *maincf_get_convert_to_charset(void)
{
    return maincf_data.convert_to_charset;
}

const char *maincf_get_convert_from_charset(void)
{
    return maincf_data.convert_from_charset;
}

const char *maincf_get_ole_charset(void)
{
    return maincf_data.ole_charset;
}

const char *maincf_get_transfer_encoding(void)
{
    return maincf_data.transfer_encoding;
}
#endif

