/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "nscurses.h"

#ifdef HAVE_CURSES_H

#include <fcntl.h>
#include <termios.h>

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#else
# ifdef HAVE_IOCTL_H
# include <ioctl.h>
# endif
#endif

#include <signal.h>

#ifdef HAVE_NCURSES_H
#include <ncurses.h>
#else
#include <curses.h>
#endif

#include "parserc.h"

/* How many blank lines separate threads from stats */
#define NSCURSES_SEPARATOR 1

#define NSCURSES_ACTIVE_FILE RC_DIR "/curses_on"

typedef struct {
    chtype attr;
    int fg_colour;
    BOOL set;
} nscurses_fg_and_attr;

#define NSCURSES_NUM_MISC_ATTR 10
#define NSCURSES_NUM_DEBLOG_ATTR 6
static struct {
    chtype normal;
    int fg_colour, bg_colour;
    union {
	struct {
	    nscurses_fg_and_attr labels, server, status, stats,
		    downloaded, download_total, download_rate,
		    uploaded, upload_total, upload_rate;
	} named;
	nscurses_fg_and_attr numbered[NSCURSES_NUM_MISC_ATTR];
    } misc;
    union {
	struct {
	    nscurses_fg_and_attr error, critical, warning, message, info, debug;
	} named;
	nscurses_fg_and_attr numbered[NSCURSES_NUM_DEBLOG_ATTR];
    } deblog;
} nscurses_attr;

static long int nscurses_pause = 100;
static int nscurses_columns, nscurses_lines;
static unsigned int nscurses_stats_column, nscurses_rate_column,
	nscurses_server_column, nscurses_status_column;
static int nscurses_lines_per_thread;
static WINDOW **nscurses_window = NULL;
static char **nscurses_servers = NULL;
static char **nscurses_statuses = NULL;

#define NSCURSES_DOWNLOAD_LABEL _("Downloaded")
#define NSCURSES_UPLOAD_LABEL _("Uploaded")
#define NSCURSES_AT_LABEL _("at")
/* #define NSCURSES_THREAD_LABEL _("Thread %d") */
#define NSCURSES_THREAD_LABEL _("%d")
#define NSCURSES_MAIN_LABEL _("Main")

#define ATTRNORM() attrset(nscurses_attr.normal)
#define WATTRNORM(w) wattrset((w), nscurses_attr.normal)
#define ATTRIF(a) attrset((a).set ? (a).attr : nscurses_attr.normal)
#define WATTRIF(w, a) wattrset((w), (a).set ? (a).attr : nscurses_attr.normal)

static BOOL nscurses_is_active = FALSE;

#ifdef HAVE_CURSES_RESIZE
static BOOL nscurses_winch = FALSE;
#endif

/* Whether we're using curses interface */
BOOL nscurses_active(void)
{
    return nscurses_is_active;
}

static void nscurses_reset_attr(void)
{
    int a;

    nscurses_attr.normal = A_NORMAL;
    nscurses_attr.fg_colour = -1;
    nscurses_attr.bg_colour = -1;
    for (a = 0; a < NSCURSES_NUM_MISC_ATTR; ++a)
    {
	nscurses_attr.misc.numbered[a].attr = A_NORMAL;
	nscurses_attr.misc.numbered[a].set = FALSE;
	nscurses_attr.misc.numbered[a].fg_colour = -1;
    }
    for (a = 0; a < NSCURSES_NUM_DEBLOG_ATTR; ++a)
    {
	nscurses_attr.deblog.numbered[a].attr = A_NORMAL;
	nscurses_attr.deblog.numbered[a].set = FALSE;
	nscurses_attr.deblog.numbered[a].fg_colour = -1;
    }
}

/* Returns curses colour number, -1 for default, -2 for invalid */
static int nscurses_decode_colour(const char *name)
{
#ifndef HAVE_CURSES_COLOR
    static BOOL nscurses_warned_no_colour = FALSE;

    UNUSED(name);

    if (!nscurses_warned_no_colour)
    {
	nscurses_warned_no_colour = TRUE;
	deblog_warning(_("Terminal does not support colour\n"));
    }
    return -1;
#else
    if (!strcmp(name, "black"))
	return COLOR_BLACK;
    else if (!strcmp(name, "white"))
	return COLOR_WHITE;
    else if (!strcmp(name, "blue"))
	return COLOR_BLUE;
    else if (!strcmp(name, "red"))
	return COLOR_RED;
    else if (!strcmp(name, "green"))
	return COLOR_GREEN;
    else if (!strcmp(name, "yellow"))
	return COLOR_YELLOW;
    else if (!strcmp(name, "brown"))
	return COLOR_YELLOW;
    else if (!strcmp(name, "cyan"))
	return COLOR_CYAN;
    else if (!strcmp(name, "magenta"))
	return COLOR_MAGENTA;
    else if (!strcmp(name, "default"))
	return -1;
    return -2;
#endif
}

static const char *nscurses_fgbg_handler(const char *keyword, const char
	*value, void *data_handle)
{
    int colour;

    UNUSED(keyword);
    colour = nscurses_decode_colour(value);
    if (colour == -2)
	return _("Not a valid colour name\n");
    *((int *) data_handle) = colour;
    return NULL;
}

static const char *nscurses_attr_handler(const char *keyword, const char
	*value, void *data_handle)
{
    char *valtok;
    char *tokin;
    char *token;
    BOOL colour_already = FALSE;
    nscurses_fg_and_attr *fgattr = data_handle;
    const char *result = NULL;

    UNUSED(keyword);
    valtok = XSTRDUP(value);
    tokin = valtok;
    while ((token = strtok(tokin, ",| \t")) != NULL)
    {
	int colour;
	
	tokin = NULL;
	if ((colour = nscurses_decode_colour(token)) != -2)
	{
	    if (colour_already)
	    {
		if (!result)
		    result = _("Multiple colours on one line");
	    }
	    else
	    {
		colour_already = TRUE;
		fgattr->fg_colour = colour;
		fgattr->set = TRUE;
	    }
	}
	else if (!strcmp(token, "bold"))
	{
	    fgattr->attr |= A_BOLD;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "underline"))
	{
	    fgattr->attr |= A_UNDERLINE;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "standout"))
	{
	    fgattr->attr |= A_STANDOUT;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "blink"))
	{
	    fgattr->attr |= A_BLINK;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "reverse"))
	{
	    fgattr->attr |= A_REVERSE;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "dim"))
	{
	    fgattr->attr |= A_DIM;
	    fgattr->set = TRUE;
	}
	else if (!strcmp(token, "normal"))
	{
	    fgattr->set = TRUE;
	}
	else
	{
	    result = _("Unknown attribute/colour");
	}
    }
    free(valtok);
    return result;
}

static void nscurses_read_config(void)
{
    parserc_register_handler("foreground", nscurses_fgbg_handler,
	    &nscurses_attr.fg_colour);
    parserc_register_handler("background", nscurses_fgbg_handler,
	    &nscurses_attr.bg_colour);
    parserc_register_handler("labels", nscurses_attr_handler,
	    &nscurses_attr.misc.named.labels);
    parserc_register_handler("server", nscurses_attr_handler,
	    &nscurses_attr.misc.named.server);
    parserc_register_handler("status", nscurses_attr_handler,
	    &nscurses_attr.misc.named.status);
    parserc_register_handler("stats", nscurses_attr_handler,
	    &nscurses_attr.misc.named.stats);
    parserc_register_handler("downloaded", nscurses_attr_handler,
	    &nscurses_attr.misc.named.downloaded);
    parserc_register_handler("download_total", nscurses_attr_handler,
	    &nscurses_attr.misc.named.download_total);
    parserc_register_handler("download_rate", nscurses_attr_handler,
	    &nscurses_attr.misc.named.download_rate);
    parserc_register_handler("uploaded", nscurses_attr_handler,
	    &nscurses_attr.misc.named.uploaded);
    parserc_register_handler("upload_total", nscurses_attr_handler,
	    &nscurses_attr.misc.named.upload_total);
    parserc_register_handler("upload_rate", nscurses_attr_handler,
	    &nscurses_attr.misc.named.upload_rate);
    parserc_register_handler("error", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.error);
    parserc_register_handler("critical", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.critical);
    parserc_register_handler("warning", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.warning);
    parserc_register_handler("message", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.message);
    parserc_register_handler("info", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.info);
    parserc_register_handler("debug", nscurses_attr_handler,
	    &nscurses_attr.deblog.named.debug);
    parserc_register_int("pause", &nscurses_pause);
    parserc_read_named(CONF_DIR "/curses.cf", deblog_Level_Warning);
}

#ifdef HAVE_CURSES_COLOR
static void nscurses_init_colours(void)
{
    int a;
    int default_pair_number = 0;
    int last_pair = 0;

    start_color();
#ifdef HAVE_NCURSES_H
    if (use_default_colors() != OK)
    {
#endif
	/* Obsolete curses means more likely to be using white on black */
	if (nscurses_attr.fg_colour == -1)
	    nscurses_attr.fg_colour = COLOR_WHITE;
	if (nscurses_attr.bg_colour == -1)
	    nscurses_attr.bg_colour = COLOR_BLACK;
#ifdef HAVE_NCURSES_H
    }
    if (nscurses_attr.fg_colour == -1 &&
	    nscurses_attr.bg_colour == -1)
    {
	use_default_colors();
    }
    else 
#endif
    if (nscurses_attr.fg_colour != COLOR_WHITE ||
	    nscurses_attr.bg_colour != COLOR_BLACK)
    {
	init_pair(last_pair = default_pair_number = 1,
		nscurses_attr.fg_colour,
		nscurses_attr.bg_colour);
    }
    nscurses_attr.normal = COLOR_PAIR(default_pair_number);
    for (a = 0; a < NSCURSES_NUM_MISC_ATTR; ++a)
    {
	if (nscurses_attr.misc.numbered[a].fg_colour != -1) 
	{
	    init_pair(++last_pair,
		    nscurses_attr.misc.numbered[a].fg_colour,
		    nscurses_attr.bg_colour);
	    nscurses_attr.misc.numbered[a].attr |=
		COLOR_PAIR(last_pair);
	}
	else if (default_pair_number)
	{
	    nscurses_attr.misc.numbered[a].attr |=
		COLOR_PAIR(default_pair_number);
	}
    } /* for (a = 0; a < NSCURSES_NUM_MISC_ATTR; ++a) */
    for (a = 0; a < NSCURSES_NUM_DEBLOG_ATTR; ++a)
    {
	if (nscurses_attr.deblog.numbered[a].fg_colour != -1) 
	{
	    init_pair(++last_pair,
		    nscurses_attr.deblog.numbered[a].fg_colour,
		    nscurses_attr.bg_colour);
	    nscurses_attr.deblog.numbered[a].attr |=
		COLOR_PAIR(last_pair);
	}
	else if (default_pair_number)
	{
	    nscurses_attr.deblog.numbered[a].attr |=
		COLOR_PAIR(default_pair_number);
	}
    } /* for (a = 0; a < NSCURSES_NUM_DEBLOG_ATTR; ++a) */
}
#endif /* HAVE_CURSES_COLOR */

/* Fills in the options that use stats if they don't have their own value */
static void nscurses_normalise_attr(void)
{
    int a;

    for (a = 3; a < NSCURSES_NUM_MISC_ATTR; ++a)
    {
	if (!nscurses_attr.misc.numbered[a].set)
	    nscurses_attr.misc.numbered[a] =
		nscurses_attr.misc.named.stats;
    }
}

static void nscurses_delete_windows(int nthreads)
{
    int t;
    
    if (!nscurses_window)
	return;
    for (t = 0; t <= nthreads; ++t)
    {
	if (nscurses_window[t])
	{
	    delwin(nscurses_window[t]);
	}
    }
}

/* A function to create a new window or resize an existing one; returns TRUE
 * for failure. Note errors have to be reported with printf in case of
 * recursive deblog */
typedef BOOL (*nscurses_windower)(WINDOW **win, int nlines, int ncols,
	int top, int left);

static BOOL nscurses_create_window(WINDOW **win, int nlines, int ncols,
	int top, int left)
{
    *win = newwin(nlines, ncols, top, left);
    if (!*win)
    {
	nscurses_stop();
	printf(_("Critical: Can't create subwindows\n"));
	fflush(stdout);
	return TRUE;
    }
    scrollok(*win, TRUE);
    return FALSE;
}

#if 0 /* Not used currently, may try it later */
#ifdef HAVE_CURSES_RESIZE
static BOOL nscurses_resize_window(WINDOW **win, int nlines, int ncols,
	int top, int left)
{
    if (wresize(*win, nlines, ncols) == ERR || mvwin(*win, top, left) == ERR)
    {
	nscurses_stop();
	printf(_("Critical: Can't resize subwindows\n"));
	fflush(stdout);
	return TRUE;
    }
    return FALSE;
}
#endif /* HAVE_CURSES_RESIZE */
#endif /* Comment block */

static void nscurses_print_server(int thread_idx)
{
    ATTRIF(nscurses_attr.misc.named.server);
    mvprintw(nscurses_lines_per_thread * (thread_idx + 1),
	    nscurses_server_column, "%-*s", nscurses_status_column -
	    nscurses_server_column - 1, nscurses_servers[thread_idx] ?
	    nscurses_servers[thread_idx] : ""); 
}

static void nscurses_print_status(int thread_idx)
{
    ATTRIF(nscurses_attr.misc.named.status);
    mvprintw(nscurses_lines_per_thread * (thread_idx + 1),
	    nscurses_status_column, "%*s", nscurses_columns -
	    nscurses_status_column - 1, nscurses_statuses[thread_idx] ?
	    nscurses_statuses[thread_idx] : ""); 
}

/* Returns TRUE if fails */
static BOOL nscurses_init_or_resize_windows(int nthreads, nscurses_windower
	windower)
{
    int t;
    
#ifdef HAVE_CURSES_RESIZE
    nscurses_winch = FALSE;
#endif
    nscurses_stats_column = XMAX(strlen(NSCURSES_DOWNLOAD_LABEL),
	    strlen(NSCURSES_UPLOAD_LABEL)) + 2;
    nscurses_rate_column = nscurses_columns - 10;
    /* nscurses_server_column = XMAX(strlen(NSCURSES_THREAD_LABEL) + 4,
	    strlen(NSCURSES_MAIN_LABEL)); */
    nscurses_server_column = strlen(NSCURSES_THREAD_LABEL) + 4;
    nscurses_status_column = (nscurses_columns + nscurses_server_column) / 2;
    if (nscurses_rate_column > nscurses_stats_column + 23 +
	    strlen(NSCURSES_AT_LABEL))
    {
	nscurses_rate_column = nscurses_stats_column + 23 +
	    strlen(NSCURSES_AT_LABEL);
    }
    nscurses_lines_per_thread = (nscurses_lines - 2 - NSCURSES_SEPARATOR) /
	(nthreads + 1);
    if (nscurses_lines_per_thread < 4 || nscurses_rate_column -
	    strlen(NSCURSES_AT_LABEL) - nscurses_stats_column < 12)
    {
	nscurses_stop();
	printf(
	    _("Critical: Terminal not large enough for full-screen display\n"));
	fflush(stdout);
	return TRUE;
    }
    for (t = 0; t <= nthreads; ++t)
    {
	if ((*windower)(&nscurses_window[t], nscurses_lines_per_thread - 1,
		nscurses_columns, nscurses_lines_per_thread * t + 1, 0))
	{
	    fprintf(stderr, _("Info: Failed at window %d/%d\n"), t, nthreads);
	    fprintf(stderr, _("Info: H:%d x W:%d, Y:%d X:%d\n"),
		    nscurses_lines_per_thread - 1, nscurses_columns,
		    nscurses_lines_per_thread * t  + 1, 0);
	    return TRUE;
	}
	scrollok(nscurses_window[t], TRUE);
	if (t)
	{
	    ATTRIF(nscurses_attr.misc.named.labels);
	    mvprintw(nscurses_lines_per_thread * t , 0, NSCURSES_THREAD_LABEL,
		    t);
	    nscurses_print_server(t - 1);
	    nscurses_print_status(t - 1);
	}
    }
    ATTRIF(nscurses_attr.misc.named.labels);
    mvprintw(0, 0, NSCURSES_MAIN_LABEL);
    mvprintw(nscurses_lines - 2, 0, NSCURSES_DOWNLOAD_LABEL); 
    mvprintw(nscurses_lines - 1, 0, NSCURSES_UPLOAD_LABEL); 
    mvprintw(nscurses_lines - 2, nscurses_rate_column -
	    strlen(NSCURSES_AT_LABEL) - 2, NSCURSES_AT_LABEL); 
    mvprintw(nscurses_lines - 1, nscurses_rate_column -
	    strlen(NSCURSES_AT_LABEL) - 2, NSCURSES_AT_LABEL); 
    refresh();
    return FALSE;
}

static BOOL nscurses_init_windows(int nthreads)
{
    return nscurses_init_or_resize_windows(nthreads, nscurses_create_window);
}

#ifdef HAVE_CURSES_RESIZE
static void nscurses_winch_handler(int sig)
{
    UNUSED(sig);
    nscurses_winch = TRUE;
}
#endif /* HAVE_CURSES_RESIZE */

void nscurses_start(int nthreads)
{
    int t;
    FILE *fp;

    nscurses_window = XMALLOC(sizeof(WINDOW *) * (nthreads + 1));
    nscurses_servers = XMALLOC(sizeof(char const *) * nthreads);
    nscurses_statuses = XMALLOC(sizeof(char const *) * nthreads);
    for (t = 0; t < nthreads; ++t)
    {
	nscurses_window[t] = NULL;
	nscurses_servers[t] = nscurses_statuses[t] = NULL;
    }
    nscurses_window[nthreads] = NULL;
    nscurses_reset_attr();
    nscurses_read_config();
    /* Give user brief chance to read prior messages */
    deblog_info(_("Starting full-screen display\n"));
    usleep(nscurses_pause * 10000);
    if (!initscr())
    {
	/* Most versions of curses will never get here */
	deblog_error(_("Could not initialise terminal\n"));
    }
    fp = fopen(NSCURSES_ACTIVE_FILE, "w");
    fclose(fp);
    getmaxyx(stdscr, nscurses_lines, nscurses_columns);
#ifdef HAVE_CURSES_RESIZE
    signal(SIGWINCH, nscurses_winch_handler);
#endif
    nscurses_is_active = TRUE;
#ifdef HAVE_CURSES_COLOR
    if (has_colors())
	nscurses_init_colours();
#endif
    nscurses_normalise_attr();
    if (nscurses_init_windows(nthreads))
    {
	nscurses_emergency_stop();
    }
    else
    {
	atexit(nscurses_emergency_stop);
    }
}

void nscurses_stop(void)
{
    if (nscurses_is_active)
    {
	/* Give user chance to read last messages */
	if (nscurses_window && nscurses_window[0])
	{
	    /*
	    wattron(nscurses_window[0], A_STANDOUT);
	    waddstr(nscurses_window[0], _("Leaving full-screen display\n"));
	    wrefresh(nscurses_window[0]);
	    */
	    /*deblog_info(_("Leaving full-screen display\n"));*/
	}
	usleep(nscurses_pause * 10000);
	nscurses_delete_windows(maincf_get_max_threads());
	nscurses_emergency_stop();
    }
}

void nscurses_emergency_stop(void)
{
    if (nscurses_is_active)
    {
	endwin();
	nscurses_is_active = FALSE;
	unlink(NSCURSES_ACTIVE_FILE);
    }
    signal(SIGWINCH, SIG_IGN);
}


/* Reads terminal size, because curses seems to have no way of doing it
 * (cribbed from mutt source) */
static void nscurses_get_term_size(int *y, int *x)
{
  char *cp;
  int fd;
  struct winsize w;

  *y = -1;
  *x = -1;
  if ((fd = open ("/dev/tty", O_RDONLY)) != -1)
  {
    if (ioctl (fd, TIOCGWINSZ, &w) != -1)
    {
      *y = w.ws_row;
      *x = w.ws_col;
    }
    close (fd);
  }
  if (*y <= 0)
  {
    if ((cp = getenv ("LINES")) != NULL)
    {
      *y = atoi (cp);
    }
    else
      *y = 24;
  }
  if (*x <= 0)
  {
    if ((cp = getenv ("COLUMNS")) != NULL)
      *x = atoi (cp);
    else
      *x = 80;
  }
}

#ifdef HAVE_CURSES_RESIZE
/* Returns TRUE if there was a resize */
static BOOL nscurses_check_size(void)
{
    int ncols, nlines;

    if (!nscurses_winch)
	return FALSE;
    nscurses_get_term_size(&nlines, &ncols);
    if (nlines == nscurses_lines && ncols == nscurses_columns)
    {
	nscurses_winch = FALSE;
	return FALSE;
    }
    resizeterm(nlines, ncols);
    nscurses_lines = nlines;
    nscurses_columns = ncols;
    nscurses_delete_windows(maincf_get_max_threads());
    clear();
    nscurses_init_windows(maincf_get_max_threads());
    return TRUE;
}
#else
#define nscurses_check_size() FALSE
#endif /* HAVE_CURSES_RESIZE */

/* Output a message for deblog (thread_idx == -1 for parent) */
void nscurses_deblog(int thread_idx, deblog_level_t level, const char *message)
{
    WINDOW *win = nscurses_window[thread_idx + 1];

    nscurses_check_size();
    WATTRIF(win, nscurses_attr.deblog.numbered[level]);
    switch (level)
    {
	case deblog_Level_Error:
	    waddstr(win, _("Error: "));
	    break;
	case deblog_Level_Critical:
	    waddstr(win, _("Critical: "));
	    break;
	case deblog_Level_Warning:
	    waddstr(win, _("Warning: "));
	    break;
	case deblog_Level_Info:
	    waddstr(win, _("Info: "));
	    break;
	case deblog_Level_Debug:
	    waddstr(win, _("Debug: "));
	    break;
	default:
	    break;
    }
    waddstr(win, message);
    wrefresh(win);
}

/* Output stats (rates are xntokilomeg strings) */
void nscurses_stats(long int downloaded, long int total_downloads,
	long int total_seen,
	const char *download_rate,
	long int uploaded, long int total_uploads, const char *upload_rate)
{
    UNUSED(total_seen);

    nscurses_check_size();
    ATTRIF(nscurses_attr.misc.named.stats);
    mvprintw(nscurses_lines - 2, nscurses_stats_column, "%*s",
	    nscurses_rate_column - 2 - strlen(NSCURSES_AT_LABEL) -
	    nscurses_stats_column, " ");
    mvprintw(nscurses_lines - 1, nscurses_stats_column, "%*s",
	    nscurses_rate_column - 2 - strlen(NSCURSES_AT_LABEL) -
	    nscurses_stats_column, " ");
    mvprintw(nscurses_lines - 2, nscurses_rate_column, "%*s",
	    nscurses_columns - 1 - nscurses_rate_column, " ");
    mvprintw(nscurses_lines - 1, nscurses_rate_column, "%*s",
	    nscurses_columns - 1 - nscurses_rate_column, " ");
    ATTRIF(nscurses_attr.misc.named.downloaded);
    mvprintw(nscurses_lines - 2, nscurses_stats_column, "%ld", downloaded);
    ATTRIF(nscurses_attr.misc.named.stats);
    addch('/');
    ATTRIF(nscurses_attr.misc.named.download_total);
    printw("%ld", total_downloads);
    /*
    ATTRIF(nscurses_attr.misc.named.stats);
    addch('/');
    ATTRIF(nscurses_attr.misc.named.download_total);
    printw("%ld", total_seen);
    */
    ATTRIF(nscurses_attr.misc.named.uploaded);
    mvprintw(nscurses_lines - 1, nscurses_stats_column, "%ld", uploaded);
    ATTRIF(nscurses_attr.misc.named.stats);
    addch('/');
    ATTRIF(nscurses_attr.misc.named.upload_total);
    printw("%ld", total_uploads);
    ATTRIF(nscurses_attr.misc.named.download_rate);
    mvprintw(nscurses_lines - 2, nscurses_rate_column, "%s", download_rate);
    ATTRIF(nscurses_attr.misc.named.stats);
    addstr(_(" Bps"));
    ATTRIF(nscurses_attr.misc.named.upload_rate);
    mvprintw(nscurses_lines - 1, nscurses_rate_column, "%s", upload_rate);
    ATTRIF(nscurses_attr.misc.named.stats);
    addstr(_(" Bps"));
    refresh();
}

/* Trims a string if necessary to fit between 2 columns, inserting ... */
static char *nscurses_trim_message(const char *message, int width)
{
    char *trimmed = NULL;
    int cutoff;

    if ((int) strlen(message) <= width)
    {
	return XSTRDUP(message);
    }
    cutoff = width / 2 - 2;
    trimmed = XMALLOC(width);
    sprintf(trimmed, "%.*s...%s", cutoff, message, message + strlen(message)
	    - cutoff);
    deblog_debug(_("Trimmed status message to '%s'\n"), trimmed);
    return trimmed;
}

void nscurses_server(int thread_idx, const char *name)
{
    free(nscurses_servers[thread_idx]);
    nscurses_servers[thread_idx] = NULL;
    nscurses_servers[thread_idx] = nscurses_trim_message(name,
	    nscurses_status_column - nscurses_server_column - 1);
    if (nscurses_check_size())
    {
	/* Server will already have been reprinted as side-effect */
	return;
    }
    nscurses_print_server(thread_idx);
    refresh();
}

void nscurses_show_status(int thread_idx, const char *message)
{
    if (!nscurses_is_active)
	return;
    free(nscurses_statuses[thread_idx]);
    nscurses_statuses[thread_idx] = NULL;
    nscurses_statuses[thread_idx] = nscurses_trim_message(message,
	    nscurses_columns - nscurses_status_column - 1);
    if (nscurses_check_size())
    {
	/* Server will already have been reprinted as side-effect */
	return;
    }
    nscurses_print_status(thread_idx);
    refresh();
}

#endif /* HAVE_CURSES_H */
