/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "group.h"
#include "intercom.h"
#include "kiddy.h"
#include "nscurses.h"
#include "nsthread.h"
#include "server.h"

static NSThread **nsthread_table = NULL;

void nsthread_send_command(NSThread *thread, intercom_Command code, const char
		*data, ...)
{
	va_list ap;

	va_start(ap, data);
	intercom_send_vcommand(thread->kiddy->queue, code, -1,
			data, ap);
	va_end(ap);
}

NSThread *nsthread_new(ServerData *serv, BOOL upload_first)
{
	NSThread *thrd;
	
	thrd = XMALLOC(sizeof(NSThread));
	thrd->kiddy = kiddy_get_available();
	if (!thrd->kiddy)
	{
		free(thrd);
		return NULL;
	}
	if (!nsthread_table)
	{
		int i;

		nsthread_table = XMALLOC(maincf_get_max_threads() * sizeof(NSThread *));
		for (i = 0; i < maincf_get_max_threads(); ++i)
			nsthread_table[i] = NULL;
	}
	thrd->kiddy->thread = thrd;
	thrd->serv = serv;
	thrd->thread_idx = thrd->kiddy->thread_idx;
	thrd->upload_first = upload_first;
	thrd->group_ix = -1;
	thrd->all_downloaded_msg = NULL;
	thrd->wait_state = nsthread_NotWaiting;
	thrd->command = NULL;
	if (nscurses_active())
	{
		nscurses_server(thrd->thread_idx, serv->identifier);
	}
	deblog_info_thread(thrd->thread_idx, _("Servicing %s, %s first\n"),
			serv->identifier, upload_first ? _("uploading") : _("downloading"));
	nsthread_table[thrd->thread_idx] = thrd;
	if (serv->status == server_Unresolved)
	{
		nscurses_show_status(thrd->thread_idx, _("Looking up address"));
		deblog_message_thread(thrd->thread_idx, _("Looking up address of %s\n"),
				serv->address);
		serv->status = server_Resolving;
		nsthread_send_command(thrd, command_Resolve, "%s %d",
				serv->address, serv->port);
	}
	else
	{
		nscurses_show_status(thrd->thread_idx, _("Connecting"));
		deblog_message_thread(thrd->thread_idx, _("Connecting to %s\n"),
				serv->identifier);
		nsthread_send_command(thrd, command_Connect,
				"%s %s %d %s %d %d %d", 
				serv->identifier, serv->addr_s, serv->proto,
				thrd->serv->path ? thrd->serv->path : "!",
				thrd->serv->pipeline, thrd->serv->ignore_436, thrd->serv->ssl);
		serv->status = serv->status == server_Resolved ? server_Connecting :
			server_Connected;
	}
	return thrd;
}

NSThread *nsthread_lookup(int thread_idx)
{
	return nsthread_table[thread_idx];
}

void nsthread_delete(NSThread *thread)
{
	deblog_debug("Deleting thread %p\n", thread);
	if (!thread)
	{
		deblog_debug(_("nsthread_delete: NULL thread\n"));
		return;
	}
	deblog_debug(_("Deleting thread %d\n"), thread->thread_idx);
	kiddy_detach_thread(thread->kiddy);
	deblog_debug(_("Detached thread\n"));
	--thread->serv->nthreads;
	deblog_debug(_("Decremented %s's thread count\n"),
			thread->serv->identifier);
	free(thread->command);
	nsthread_table[thread->thread_idx] = NULL;
	free(thread);
	deblog_debug(_("Freed memory\n"));
}

