/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Parses a config file */


#include "common.h"

#include <sys/stat.h>
#include <fcntl.h>

#include "parserc.h"
#include "rw.h"

typedef struct parserc_node {
    struct parserc_node *next;
    const char *keyword;
    parserc_handler handler;
    void *data_handle;
} parserc_node;

static parserc_node *parserc_handler_list = NULL;

void parserc_register_handler(const char *keyword, parserc_handler handler,
        void *data_handle)
{
    parserc_node *node;

    node = XMALLOC(sizeof(parserc_node));
    node->next = parserc_handler_list;
    parserc_handler_list = node;
    node->keyword = keyword;
    node->handler = handler;
    node->data_handle = data_handle;
}

static const char *parserc_string_handler(const char *keyword, const char
	*value, void *dh)
{
    char **valout = dh;

    UNUSED(keyword);
    
    *valout = XSTRDUP(value);
    return NULL;
}

static const char *parserc_int_handler(const char *keyword, const char *value,
	void *dh)
{
    int len;
    int i;
    long int *valout = dh;

    UNUSED(keyword);
    
    len = strlen(value);
    if (!len)
	return _("Bad number");
    for (i = 0; i < len; ++i)
    {
        if (!isdigit(value[i]) && !(i == 0 && value[i] == '-'))
        {
            return _("Bad number");
        }
    }
    *valout = strtol(value, NULL, 10);
    return NULL;
}

static const char *parserc_bool_handler(const char *keyword, const char *value,
	void *dh)
{
    BOOL *valout = dh;

    UNUSED(keyword);
    
    switch (tolower(value[0]))
    {
        case 't': case 'y': case '1':
            *valout = TRUE;
            break;
        case 'f': case 'n': case '0':
            *valout = FALSE;
            break;
	case 'o':
	    if (tolower(value[1]) == 'n')
	    {
		*valout = TRUE;
		break;
	    }
	    else if (tolower(value[1]) == 'f' && tolower(value[2]) == 'f')
	    {
		*valout = FALSE;
		break;
	    }
	    /* Otherwise fall through to error */
        default:
            return _("Bad boolean");
    }
    return NULL;
}

static const char *parserc_ignore_handler(const char *keyword, const char
	*value, void *dh)
{
    UNUSED(keyword);
    UNUSED(value);
    UNUSED(dh);
    
    return NULL;
}

void parserc_register_int(const char *keyword, long int *value)
{
    parserc_register_handler(keyword, parserc_int_handler, value);
}

void parserc_register_bool(const char *keyword, BOOL *value)
{
    parserc_register_handler(keyword, parserc_bool_handler, value);
}

void parserc_register_string(const char *keyword, char **value)
{
    parserc_register_handler(keyword, parserc_string_handler, value);
}

void parserc_register_ignored(const char *keyword)
{
    parserc_register_handler(keyword, parserc_ignore_handler, NULL);
}

/* Default master handler, tries to find a sub-handler with matching keyword */
static const char *parserc_default_master(const char *keyword, const char
	*value, void *data_handle)
{
    parserc_node *node;
    BOOL matched = FALSE;
    
    UNUSED(data_handle);

    for (node = parserc_handler_list; node; node = node->next) 
    { 
        if (!strcmp(keyword, node->keyword))
        {
            const char *errmess;

            matched = TRUE;
            errmess = node->handler(keyword, value, node->data_handle);
            if (errmess) 
		return errmess;
	    break;
        }
    }
    return matched ? NULL : _("Unknown keyword");
}

static parserc_handler parserc_master_handler = parserc_default_master;
static void *parserc_master_data_handle = NULL;

void parserc_register_master_handler(parserc_handler handler, void *data_handle)
{
    parserc_master_handler = handler;
    parserc_master_data_handle = data_handle;
    deblog_debug(_("parserc_master_handler changed\n"));
}

static void parserc_deregister(void)
{
    parserc_node *node = parserc_handler_list;

    while (node)
    {
        parserc_node *next = node->next;

        free(node);
        node = next;
    }
    parserc_handler_list = NULL;
    parserc_master_handler = parserc_default_master;
    deblog_debug(_("parserc handlers cleared\n"));
}

/* Returns message if error */
static const char *parserc_parse_line(const char *line)
{
    unsigned int n;
    char *keyword;
    const char *value;
    char *tempval;
    BOOL quote = FALSE;
    const char *end;

    while (*line && isspace(*line))
        ++line;
    if (!*line || *line == '#') 
    { 
        return NULL;
    }
    /* Isolate keyword */
    for (n = 0; line[n] && !isspace(line[n]); ++n); 
    keyword = XMALLOC(n + 1);
    strncpy(keyword, line, n); 
    keyword[n] = 0; 
    /* Isolate value */
    end = line + strlen(line);
    for (value = line; value != end && !isspace(*value); ++value);
    for (; value != end && isspace(*value); ++value);
    /*
    if (value == end)
    {
	free(keyword);
	return _("Missing value");
    }
    */
    if (*value == '\"')
    {
        quote = 1;
        ++value;
        if (value == end || *(end - 1) != '\"')
	    return _("Missing end quote");
    }
    tempval = XMALLOC(strlen(value) + 1);
    strcpy(tempval, value);
    if (quote)
    {
        tempval[strlen(value) - 1] = 0;
        /* No sense in worrying over wasted byte */
    }
    end = (*parserc_master_handler)(keyword, tempval,
	    parserc_master_data_handle);
    free(keyword);
    free(tempval);
    return end;
}

void parserc_read(int filedes)
{
    int reading = TRUE;
    int ln = 1;
    rw_buf *buf = rw_create_buf(filedes, FALSE, FALSE, -1);
    
    deblog_debug(_("Entering parserc_read\n"));
    if (!buf)
    {
        deblog_error(_("Out of memory\n"));
        exit(1);
    }
    while (reading)
    {
        const char *line = rw_read_line(buf);

        if (line)
        {
            const char *errmess = parserc_parse_line(line);
	    if (errmess)
	    {
		deblog_critical(_("%s in line %d: `%s'\n"), errmess, ln, line);
	    }
        }
        else
        {
            reading = FALSE;
        }
        ++ln;
    }
    if (!rw_buf_eof(buf))
    {
	deblog_critical(_("%s reading config file\n"), strerror(errno));
    }
    rw_close_and_free(buf);
    parserc_deregister();
    deblog_debug(_("Exiting parserc_read\n"));
}

void parserc_read_named(const char *filename, deblog_level_t level)
{
    int fd = open(filename, O_RDONLY);

    deblog_info(_("Reading config file `%s'\n"), filename);
    if (fd == -1)
    {
	deblog_printf(level, _("%s opening `%s'\n"), strerror(errno), filename);
        parserc_deregister();
        return;
    }
    parserc_read(fd);
    close(fd);
}
