#ifndef RW_H
#define RW_H

/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Routines for reading and writing files etc */


#ifndef COMMON_H
#include "common.h"
#endif

#ifdef SUPPORT_SSL
#ifdef HAVE_OPENSSL_SSL_H
#include "openssl/ssl.h"
#else
#include "ssl.h"
#endif
#endif

#if USE_PTHREADS
#include <pthread.h>
#endif

/* Buffer for line at a time reading */
typedef struct rw_buf rw_buf;

typedef enum {
    rw_Normal,
    rw_Eof,
    rw_Err
} rw_buf_status;

struct rw_buf {
    int fd;
    char *buf;
    size_t buflen;
    size_t surplus_start, surplus_size;
    rw_buf_status status;
    BOOL crlf;
    BOOL threadsafe;
    int thread_idx;
    struct timeval tv_read, tv_write;
    long int rcvd, sent;
    BOOL scrutinise;
#ifdef SUPPORT_SSL
    SSL *ssl;
#endif
#if USE_PTHREADS
    pthread_mutex_t mutex;
#endif
};

/* Create a buffer attached to a file descriptor; crlf is TRUE if line
 * terminators should be CRLF rather than LF. If threadsafe is TRUE, incomplete
 * writes are treated as a fatal error.
 * If thread_idx != -1 the read/write routines should send stats to the parent
 * process, using thread_idx to identify themselves */
#ifdef SUPPORT_SSL
rw_buf *rw_create_buf_ssl(int fd, SSL *ssl, BOOL crlf,
	BOOL threadsafe, int thread_idx);
#define rw_create_buf(f, c, t, i) rw_create_buf_ssl((f), NULL, (c), (t), (i))
#else
rw_buf *rw_create_buf(int fd, BOOL crlf, BOOL threadsafe, int thread_idx);
#endif

/* Free buffer (doesn't close attached file) */
void rw_free_buf(rw_buf *);

/* Close attached file, return value as for close(). This will also call
 * SSL_shutdown if supported and ssl is non-NULL */
int rw_close(rw_buf *);

/* Do both of above */
int rw_close_and_free(rw_buf *);

/* When reading lines, newlines are stripped, and added when sending */

/* Read a line from a buffered file.
 * String overwritten by any subsequent calls for same rw_buf
 * Returns NULL if error or EOF, use rw_buf_eof() to detect which.
 */
const char *rw_read_line(rw_buf *);

/* Whether EOF has been reached */
BOOL rw_buf_eof(const rw_buf *);

/* Write a line to a file. Returns number of bytes written excluding
 * terminator. Should always be length of string but value provided to support
 * non-blocking in possible future versions.  Returns -1 for error */
int rw_write_line(rw_buf *, const char *);

/* As above but without adding line terminator */
int rw_write(rw_buf *, const char *);

/* As rw_write, but formatted */
int rw_writef(rw_buf *, const char *, ...);
int rw_vwritef(rw_buf *, const char *, va_list);

/* Prototype for a user-supplied modifier function to be used with the
 * following functions. It may replace the passed string with a new one. It can
 * return NULL for a line to be deleted altogether. Care is needed to avoid
 * memory leaks.
 *
 * The modifier function is passed whatever handle is given to rw_ftos or
 * rw_stof.
 *
 * A NULL function pointer can be passed to the functions if no modifier is
 * needed.
 */
typedef const char *(*rw_modifier)(const char *, void *handle);

/* Copies from a file to a socket, using . as EOF indicator with escaping.
   Returns 0 for success, >0 for incomplete, <0 for error */
int rw_ftos(rw_buf *src, rw_buf *dest, rw_modifier, void *handle);

/* Complement of above, removing indicator and escaping */
int rw_stof(rw_buf *src, rw_buf *dest, rw_modifier, void *handle);

/* Simple copy with no special behaviour for . */
int rw_copy(rw_buf *src, rw_buf *dest, rw_modifier, void *handle);

#endif /* RW_H */
