/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Score/killfile */


#include "common.h"

#include <sys/stat.h>
#include <fcntl.h>

#include "parserc.h"
#include "score.h"

/* Opens score file for reading, returning fd (-1 if doesn't exist) */
static int score_open_file(const char *server_name)
{
    int fd;
    char *filename = XSTRDUPCAT(CONF_DIR "/score.", server_name);

    deblog_info(_("Opening scorefile %s\n"), filename);
    fd = open(filename, O_RDONLY);
    if (fd == -1)
    {
	deblog_info(_("Scorefile %s doesn't exist\n"), filename);
    }
    free(filename);
    return fd;
}

#ifdef HAVE_REGEX_H

#include <regex.h>

struct score_t {
    regex_t regex;
    long int score;	/* 0 marks end of array */
};

/* If score_master_present is TRUE but score_master_array is NULL, this means
 * we haven't checked yet */
static score_t *score_master_array = NULL;
static BOOL score_master_present = TRUE;

inline static BOOL score_need_to_load_master(void)
{
    return score_master_present && !score_master_array;
}

/* Scores array has to hold one more entry than loaded with a score of 0 to
 * represent end of array */
typedef struct {
    score_t *scores;
    size_t scores_size;
    size_t nscores;
} score_parserc_handle;

/* keyword is actually score represented as a string; value is regex; handle is
 * score_parserc_handle */
static const char *score_parserc_handler(const char *keyword, const char
	*value, void *handle)
{
    score_parserc_handle *shandle = (score_parserc_handle * ) handle;
    score_t *score;
    regex_t reg;
    int errcode;
    long int scoreval = atol(keyword);

    if (!scoreval)
	return _("0 score disallowed");
    if ((errcode = regcomp(&reg, value, REG_EXTENDED | REG_NEWLINE | REG_NOSUB
		    | (maincf_get_regex_ic() ? REG_ICASE : 0))) != 0)
    {
	static char score_regexerr[256];

	regerror(errcode, &reg, score_regexerr, 256);
	return score_regexerr;
    }
    if (shandle->nscores + 2 > shandle->scores_size)
    {
	shandle->scores = XREALLOC(shandle->scores, (shandle->scores_size *= 2)
		* sizeof(score_t));
    }
    deblog_debug(_("  Rule %d scores %ld for %s\n"), shandle->nscores,
	    scoreval, value);
    score = shandle->scores + shandle->nscores;
    score->score = scoreval;
    score->regex = reg;
    score = shandle->scores + ++shandle->nscores;
    score->score = 0;
    deblog_debug("Added\n");
    return NULL;
}

/* NULL server_name means load master.score if not already loaded */
score_t *score_load(const char *server_name)
{
    score_parserc_handle handle;
    int fd = -1;
    
    if (!server_name)
    {
	if (score_need_to_load_master())
	{
	    deblog_info(_("Opening scorefile " CONF_DIR "/master.score\n"));
	    fd = open(CONF_DIR "/master.score", O_RDONLY);
	    if (fd == -1)
	    {
		score_master_present = FALSE;
		deblog_info(_("master.score doesn't exist\n"));
	    }
	}
	else
	{
	    return NULL;
	}
    }
    else
    {
	if (score_need_to_load_master())
	{
	    score_load(NULL);
	}
	fd = score_open_file(server_name);
    }

    if (fd == -1)
	return NULL;
    handle.scores = XMALLOC(sizeof(score_t));
    handle.scores_size = 1;
    handle.nscores = 0;
    handle.scores[0].score = 0;
    parserc_register_master_handler(score_parserc_handler, &handle);
    if (server_name)
    {
	deblog_info(_("Reading scorefile for %s\n"), server_name);
    }
    else
    {
	deblog_info(_("Reading master.score\n"));
    }
    parserc_read(fd);
    if (!handle.nscores)
    {
	free(handle.scores);
	handle.scores = NULL;
    }
    if (!server_name)
    {
	score_master_array = handle.scores;
	score_master_present = score_master_array != NULL;
    }
    return handle.scores;
}

/* NULL scores arg means use master */
long int score_calculate(const score_t *scores, const char *headers)
{
    size_t n;
    long int score = 0;

    if (scores)
    {
	if (score_master_present)
	{
	    deblog_debug(_("Checking headers against master.score\n"));
	    score = score_calculate(NULL, headers);
	}
    }
    else
    {
	scores = score_master_array;
    }

    for (n = 0; scores[n].score; ++n)
    {
	if (!regexec(&scores[n].regex, headers, 0, NULL, 0))
	{
	    score += scores[n].score;
	    deblog_debug(_("  Expression %d matches, new score %ld\n"), n,
		    score);
	}
    }
    return score;
}

BOOL score_have_master(void)
{
    if (score_need_to_load_master())
	score_load(NULL);
    return score_master_present;
}

#else /* HAVE_REGEX_H */

struct score_t {
    long int score;
};

score_t *score_load(const char *server_name)
{
    int fd = score_open_file(server_name);

    if (fd != -1)
    {
	close(fd);
	deblog_warning(_("%s has a scorefile, but scoring isn't supported\n"),
		server_name);
    }
    return NULL;
}

#endif /* HAVE_REGEX_H */

