/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include <limits.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include <netdb.h>
#include <arpa/inet.h>

#include "backfill.h"
#include "group.h"
#include "main.h"
#include "parserc.h"
#include "server.h"
#include "stats.h"

#ifndef MAX_SERVERS
#define MAX_SERVERS 20
#endif

static ServerData *server_table[MAX_SERVERS];

typedef struct {
		const char *feed_name;
		const ServerData *serv;
} FeedData;

static FeedData server_feeds[MAX_SERVERS];
static int server_nfeeds = 0;

static char *server_newsrc_filename(ServerData *serv)
{
	char *result = XSTRDUPCAT(RC_DIR "/newsrc.", serv->identifier);
	return result;
}

static void server_clear_groups(ServerData *serv)
{
	long int n;
	
	for (n = 0; n < serv->ngroups; ++n)
		group_delete(serv->group[n]);
	free(serv->group);
	serv->group = NULL;
	serv->ngroups = 0;
	for (n = 0; n < serv->noutgoing; ++n)
		free(serv->outgoing[n]);
	free(serv->outgoing);
	serv->outgoing = NULL;
}

void server_mark_error(ServerData *serv)
{
	/*server_clear_groups(serv);*/
	serv->status = server_Error;
}

void server_delete(ServerData *serv)
{
	deblog_debug(_("Deleting server data for %s\n"), serv->identifier);
	server_clear_groups(serv);
	server_table[serv->sidx] = NULL;
	free(serv->identifier);
	free(serv->addr_s);
	free(serv->address);
	free(serv);
}

void server_add_group(ServerData *serv, GroupData *group)
{
	if (serv->ngroups >= serv->group_size)
	{
		serv->group = XREALLOC(serv->group, sizeof(GroupData *) *
				(serv->group_size = serv->group_size ? serv->group_size * 2 :
				 4));
	}
	deblog_debug(_("Group %s added to server %s at index %ld\n"),
			group->name, serv->identifier, serv->ngroups);
	serv->group[serv->ngroups++] = group;
}

/* NB filename is used in situ, not duplicated */
static void server_add_outgoing(ServerData *serv, char *filename,
		BOOL alter_stats)
{
	if (serv->noutgoing >= serv->outgoing_size)
	{
		serv->outgoing = XREALLOC(serv->outgoing, sizeof(char *) *
				(serv->outgoing_size = serv->outgoing_size ?
				 serv->outgoing_size * 2 : 4));
	}
	deblog_debug(_("Outgoing %s added to server %s at index %ld\n"),
			filename, serv->identifier, serv->noutgoing);
	serv->outgoing[serv->noutgoing++] = filename;
	if (alter_stats)
		stats_inc_total_to_upload();
}

static void server_scan_outgoing_dir(ServerData *serv, const char *feed_name)
{
	DIR *dp;
	struct dirent *entry;
	char *parent_dir = XSTRDUPCAT(maincf_get_outgoing_dir(),
			maincf_get_using_sn() ? "/.newsstar/" : "/newsstar/");
	char *dirname = XSTRDUPCAT(parent_dir, feed_name);
	int n;
	BOOL alter_stats = TRUE;
	
	free(parent_dir);
	
	if ((dp = opendir(dirname)) == NULL)
	{
		deblog_info(_("No outgoing directory for feed %s\n"), feed_name);
		return;
	}

	/* If another server is already using this feed we don't alter total of
	 * articles to upload with this one. A simpler check would be to see if
	 * feed_name == identifier but that could be fooled by a feed which is only
	 * accessed by extra_feeds and no native feeds */
	for (n = 0; n < server_nfeeds; ++n)
	{
		if (!strcmp(server_feeds[n].feed_name, feed_name))
		{
			if (server_feeds[n].serv != serv)
				alter_stats = FALSE;
		}
		break;
	}
	if (n == server_nfeeds)
	{
		server_feeds[n].serv = serv;
		server_feeds[n].feed_name = feed_name;
		++server_nfeeds;
	}

	while ((entry = readdir(dp)) != NULL)
	{
		char *filename;
		
		if (entry->d_name[0] == '.')
			continue;
		filename = XMALLOC(strlen(dirname) + strlen(entry->d_name) + 2);
		sprintf(filename, "%s/%s", dirname, entry->d_name);
		server_add_outgoing(serv, filename, alter_stats);
		/* Don't free filename, see server_add_outgoing() */
	}
	closedir(dp);
	free(dirname);
}

static const char *server_parserc_handler(const char *group_name, const char
		*valstr, void *dh)
{
	long int last_downloaded = 0;
	ServerData *serv = dh;
	long int n;

	for (n = 0; n < serv->ngroups; ++n)
	{
		if (!strcmp(serv->group[n]->name, group_name))
		{
			deblog_critical(_("Duplicate entry for group %s on server %s\n"),
					group_name, serv->identifier);
			return NULL;
		}
	}
	if (valstr && valstr[0])
		last_downloaded = strtol(valstr, NULL, 10);
	server_add_group(serv, group_new(group_name, last_downloaded, serv));
	return NULL;
}

void server_read_newsrc(ServerData *serv)
{
	char *filename;

	deblog_info(_("Reading newsrc file for %s\n"), serv->identifier);
	filename = server_newsrc_filename(serv);
	parserc_register_master_handler(server_parserc_handler, serv);
	parserc_read_named(filename, deblog_Level_Error);
	free(filename);
	serv->status = server_Unresolved;
}

void server_write_newsrc(ServerData *serv)
{
	char *filename;
	FILE *fp;
	long int n;

	if (serv->backfill_expiry)
		backfill_save(serv->bfserv, serv->backfill_expiry);
	if (!serv->ngroups || !serv->group)
	{
		if (main_normal_exit)
		{
			deblog_info(_("No groups to write to newsrc.%s\n"),
					serv->identifier);
		}
		return;
	}
	if (main_normal_exit)
		deblog_info(_("Writing newsrc.%s\n"), serv->identifier);
	filename = server_newsrc_filename(serv);
	fp = fopen(filename, "w");
	if (!fp)
	{
		fprintf(stderr, _("<M> Critical: Can't open %s for writing\n"),
				filename);
		free(filename);
		return;
	}
	for (n = 0; n < serv->ngroups; ++n)
		group_write_newsrc(serv->group[n], fp);
	fclose(fp);
	free(filename);
}

void server_read_cf(ServerData *serv)
{
	char *filename;
	char *path = NULL;
	long int port = -1;

	deblog_info(_("Reading cf file for %s\n"), serv->identifier);
	filename = XSTRDUPCAT(CONF_DIR "/cf.", serv->identifier);
	parserc_register_string("server", &serv->address);
	parserc_register_string("path", &path);
	parserc_register_int("port", &port);
	parserc_register_string("user", &serv->user);
	parserc_register_string("password", &serv->password);
	parserc_register_bool("feed", &serv->feed);
	parserc_register_string("extra_feed", &serv->extra_feed);
	parserc_register_bool("mode_reader", &serv->mode_reader);
	parserc_register_bool("enter_group", &serv->enter_group);
	parserc_register_bool("pipeline", &serv->pipeline);
	parserc_register_bool("xover", &serv->xover);
	parserc_register_bool("ignore_436", &serv->ignore_436);
	parserc_register_bool("ssl", &serv->ssl);
	parserc_register_int("max_threads", &serv->max_threads);
	parserc_register_int("server_weight", &serv->server_weight);
	parserc_register_int("min_worth_getting", &serv->min_worth_getting);
	parserc_register_int("default_score", &serv->default_score);
	parserc_register_int("kill_score", &serv->kill_score);
	parserc_register_int("backfill_expiry", &serv->backfill_expiry);
	parserc_read_named(filename, deblog_Level_Info);
	serv->port = port;
	if (serv->kill_score > 0)
	{
		deblog_warning(_("Positive kill_score in %s, negating\n"),
				filename);
		serv->kill_score = -serv->kill_score;
	}
	if (serv->min_worth_getting < 0)
	{
		deblog_warning(_("Negative min_worth_getting in %s, using 0\n"),
				filename);
		serv->min_worth_getting = 0;
	}
	free(filename);
	if (path)
	{
		if (serv->path)
			free(serv->path);
		serv->path = path;
	}
#ifndef SUPPORT_SSL
	if (serv->ssl)
	{
		deblog_error(_("Server %s has ssl option set, "
					"but support is not compiled in\n"), serv->identifier);
	}
#endif
}

static void server_exit_handler(void)
{
	static BOOL recursive_exit = FALSE;

	if (!recursive_exit)
	{
		int n;

		recursive_exit = TRUE;
		for (n = 0; n < MAX_SERVERS; ++n)
		{
			if (!server_table[n])
				continue;
			server_write_newsrc(server_table[n]);
		}
	}
}

static ServerData *server_new(const char *identifier)
{
	static BOOL in_use_initialised = FALSE;
	ServerData *serv;
	int i;

	deblog_info(_("Generating data for server %s\n"), identifier);
	
	if (!in_use_initialised)
	{
		for (i = 0; i < MAX_SERVERS; ++i)
			server_table[i] = NULL;
		in_use_initialised = TRUE;
	}
	for (i = 0; i < MAX_SERVERS && server_table[i]; ++i);
	if (i >= MAX_SERVERS)
		deblog_error(_("Trying to create too many servers\n"));
	serv = XMALLOC(sizeof(ServerData));
	server_table[i] = serv;
	serv->sidx = i;
	serv->status = server_Unresolved;
	serv->identifier = XSTRDUP(identifier);
	serv->addr_s = NULL;
	serv->address = serv->user = serv->password = NULL;
	serv->user = serv->password = NULL;
	if (maincf_get_path())
		serv->path = XSTRDUP(maincf_get_path());
	else
		serv->path = NULL;
	serv->mode_reader = TRUE;
	serv->enter_group = FALSE;
	serv->xover = -1;
	serv->pipeline = maincf_get_pipeline();
	serv->max_threads = maincf_get_threads_per_server();
	serv->nthreads = 0;
	serv->next_group_stats = 0;
	serv->ngroups = serv->group_size = 0;
	serv->group = NULL;
	serv->server_weight = 0;
	serv->feed = TRUE;
	serv->extra_feed = NULL;
	serv->outgoing_size = serv->noutgoing = serv->nuploaded = 0;
	serv->outgoing = NULL;
	serv->total_to_download = 0;
	serv->uploading = FALSE;
	serv->min_worth_getting = 0;
	serv->default_score = 100;
	serv->kill_score = 0;
	serv->scores = NULL;
	serv->backfill_expiry = 0;
	serv->bfserv = NULL;
	serv->xref_in_xover = -1;
	serv->xref_includes_name = FALSE;
	serv->xref_is_newsgroups = FALSE;
	serv->ignore_436 = FALSE;
	serv->ssl = FALSE;
	serv->connx_msg = FALSE;
	serv->locked_upload = FALSE;

	deblog_info(_("Server %s has index %d\n"), identifier, i);
	server_read_newsrc(serv);
	server_read_cf(serv);
	serv->scores = score_load(serv->identifier);
	deblog_debug("Loaded scores\n");
	if (score_have_master() || serv->scores)
	{
		if (!serv->xover)
		{
			deblog_warning(
					_("Scorefile in use for %s, forcing xover option\n"),
					identifier);
		}
		deblog_debug(_("Scores exist, forcing/defaulting xover option on\n"));
		serv->xover = TRUE;
	}
	else if (serv->xover == -1)
	{
		deblog_debug(_("No scores, xover option defaulting to off\n"));
		serv->xover = FALSE;
	}
	if (!serv->address)
		serv->address = XSTRDUP(serv->identifier);
	if (serv->feed)
		server_scan_outgoing_dir(serv, serv->identifier);
	if (serv->extra_feed)
		server_scan_outgoing_dir(serv, serv->extra_feed);
	if (!serv->noutgoing)
		serv->feed = FALSE;
	if (serv->backfill_expiry)
		serv->bfserv = backfill_load(serv->identifier);
	deblog_debug(_("Server %s pipelining %s, ignore_436 %s\n"),
			identifier, serv->pipeline ?  _("enabled") : _("disabled"),
			identifier, serv->ignore_436 ?  _("enabled") : _("disabled"));

	return serv;
}

void server_scan_rc_dir(const char *servname)
{
	long int max_threads = 0;

	if (servname)
	{
		ServerData *serv = server_new(servname);
		max_threads += serv->max_threads;
	}
	else
	{
		DIR *dp;
		struct dirent *entry;
		BOOL foundrcs = FALSE;		/* for determining if at least one file was
								   found */
 
		if ((dp = opendir(RC_DIR)) != NULL)
		{
			while ((entry = readdir(dp)) != NULL)
			{
				deblog_debug(_("file found `%s'\n"), entry->d_name);
				if (!strncmp(entry->d_name, "newsrc.", sizeof("newsrc.")
							- 1))
				{
					deblog_debug(_("newsrc file found '%s'\n"),
							entry->d_name);
					if (!strcmp(entry->d_name, "newsrc.sample"))
					{
						deblog_info( _("Skipping newsrc.sample\n"));
					}
					else
					{
						ServerData *serv = server_new(entry->d_name +
								sizeof("newsrc.") - 1);
						max_threads += serv->max_threads;
						foundrcs = TRUE;
					}
				}
			}
			if (!foundrcs)
			{
				deblog_error(_("No newsrc files\n"));
			}
			closedir(dp);
		}
		else
		{
			deblog_error(_("Could not open directory: %s\n"), RC_DIR);
		}
	}
	atexit(server_exit_handler);
	if (max_threads < maincf_get_max_threads())
		maincf_reset_max_threads(max_threads);
}

long int server_get_next_group_stats(ServerData *serv)
{
	if (serv->next_group_stats >= serv->ngroups)
		return -1;
	serv->group[serv->next_group_stats]->status = group_Allocated;
	return serv->next_group_stats++;
}

long int server_get_best_group_articles(ServerData *serv, long int
		favoured_group, BOOL *waiting)
{
	long int best = 0;
	long int best_ix = -1;
	long int best_artnum = 0;
	int best_nthreads = INT_MAX;
	long int group_ix = favoured_group;

	deblog_debug(_("Choosing best group for fetching, current %ld %s\n"),
			group_ix, group_ix == -1 ? _("<none>") :
			serv->group[group_ix]->name);
	if (waiting)
		*waiting = FALSE;
	/* Any available in favoured_group? */
	if (group_ix != -1 && serv->group[group_ix]->status == group_Downloading &&
			serv->group[group_ix]->last_available >
			serv->group[group_ix]->last_downloaded)
	{
		deblog_debug(_("Current group has articles apparently available\n"));
		switch (best_artnum = group_get_next_article(serv->group[group_ix],
					NULL))
		{
			case -1:		/* Don't know what's available yet */
				best = LONG_MAX; /* This will be best if nothing else has info
									available either */
				best_ix = group_ix;
				if (waiting)
					*waiting = TRUE;
				deblog_debug(
						_("Not sure of availability yet, looking further\n"));
				break;
			case 0:		/* Nothing available */
				deblog_debug(_("Not really available, ruling it out\n"));
				best_ix = -1;
				break;
			default:
				deblog_debug(_("Confirmed available, go with this\n"));
				return group_ix;
		}
	}
	/* Check all other groups */
	for (group_ix = 0; group_ix < serv->ngroups; ++ group_ix)
	{
		const GroupData *group = serv->group[group_ix];

		if (group_ix == favoured_group)
			continue;
		deblog_debug(_("Group %ld %s status %d\n"), group_ix, group->name,
				group->status);
		if (group->status == group_Downloading)
		{
			long int n_available = 
				group->last_available - group->last_downloaded;

			deblog_debug(
				_("%ld %s has %ld available and %d threads %ld\n"),
					group_ix, serv->group[group_ix]->name,
					n_available, group->nthreads);
			if (best_artnum == -1 || group->nthreads < best_nthreads ||
				(group->nthreads == best_nthreads && n_available > best))
			{
				long int first_artnum;

				deblog_debug(
						_("Beats previous best %ld, confirming availability\n"),
						best);
				first_artnum = group_get_next_article(serv->group[group_ix],
						NULL);
				switch (first_artnum)
				{
					case 0:
						deblog_debug(
								_("Not really available, ruling it out\n"));
						break;
					case -1:
						deblog_debug(
								_("Not sure of availability yet\n"));
						if (waiting)
							*waiting = TRUE;
						break;
					default:
						deblog_debug(_("Confirmed availability\n"));
						break;
				}
				if (first_artnum > 0 ||
						(first_artnum == -1 && best_artnum == -1))
				{
					deblog_debug(_("Best so far\n"));
					best = n_available;
					best_ix = group_ix;
					best_artnum = first_artnum;
					best_nthreads = group->nthreads;
				}
			}
		}
	}
	deblog_debug(_("Best is %ld %s\n"), best_ix,
			best_ix == -1 ? _("<none>") : serv->group[best_ix]->name);
	return best_ix;
}

server_GFT_Result server_get_for_threading(ServerData **serv_out, BOOL
		*upload_first)
{
	ServerData *best_server = NULL;
	long int best_score = 0;
	int n;
	BOOL all_done = TRUE;

	/* Find most eligible server */
	deblog_info(
			_("gft: Finding most eligible server to start a thread for\n"));
	for (n = 0; n < MAX_SERVERS; ++n)
	{
		long int score = 0;
		ServerData *serv;
		BOOL all_unknown;
		long int group_ix;
		BOOL check_uploads = FALSE;
		BOOL all_groups_done;

		serv = server_table[n];
		if (!serv)
			continue;
		deblog_info(_("gft: %s has %d/%d threads allocated\n"),
				serv->identifier, serv->nthreads, serv->max_threads);
		deblog_info(_("gft: %s has had %d/%d groups looked at\n"),
				serv->identifier, serv->next_group_stats, serv->ngroups);
		if (serv->nthreads >= serv->max_threads)
		{
			/* This is basically equivalent to AllDone */
			score = -2;
			continue;
		}
		if (!serv->ngroups && !serv->noutgoing)
		{
			deblog_info(_("No groups or uploads for %s\n"), serv->identifier);
			score = -2;
			continue;
		}
		deblog_info(_("gft: %s has status %d\n"), serv->identifier,
				serv->status);
		switch (serv->status)
		{
			case server_Error:
				deblog_info(_("gft: %s has an error condition\n"),
						serv->identifier);
				score = -2;
				break;
			case server_AllDone:
				deblog_info(_("gft: %s all done\n"),
						serv->identifier);
				score = -2;
				break;
			case server_Connecting:
			case server_Resolving:
				deblog_info(
						_("gft: Still resolving/awaiting connection for %s\n"),
						serv->identifier);
				score = -2;
				all_done = FALSE;
				break;
			case server_Unresolved:
				/* Great, we really want to connect to this one */
				deblog_info(_("gft: %s has unknown address\n"),
						serv->identifier);
				score += 20000;
				all_done = FALSE;
				check_uploads = TRUE;
				break;
			case server_Resolved:
				/* Great, we really want to connect to this one */
				deblog_info(
			_("gft: %s has known address but hasn't been connected to yet\n"),
						serv->identifier);
				score += 10000;
				all_done = FALSE;
				check_uploads = TRUE;
				break;
			case server_Connected:
				score -= serv->min_worth_getting;
				/* Might have some info available about number of articles
				 * available we can use to influence score */
				deblog_info(_("gft: %s has already been connected to\n"),
						serv->identifier);
				/* all_unknown and all_groups_done are set by following loop */
				all_unknown = TRUE;
				all_groups_done = TRUE;
				for (group_ix = 0; group_ix < serv->next_group_stats;
						++group_ix)
				{
					GroupData *group = serv->group[group_ix];

					switch (group->status)
					{
						long int real_available;

						case group_Downloading:
							/* Check whether these articles really are
							 * available, they might have already been fetched
							 */
							all_unknown = FALSE;
							real_available = group_get_next_article(group,
									NULL);
							if (real_available != 0)
							{
								if (real_available == -1)
									real_available = group->last_requested;
								score += group->last_available - real_available;
								all_groups_done = FALSE;
								all_done = FALSE;
								deblog_info(
									_("gft: %s on %s has %d %s available\n"),
									group->name, serv->identifier,
									group->last_available -
									group->last_requested,
									group->last_available -
									group->last_requested == 1 ?
										_("article") : _("articles"));
								break;
							}
							/* Note if none are really available, we fall
							 * through to next case */
						case group_AllDone:
							all_unknown = FALSE;
							deblog_info(
								_("gft: %s on %s has no articles available\n"),
								group->name, serv->identifier);
							break;
						default:
							all_groups_done = FALSE;
							all_done = FALSE;
							deblog_info(
								_("gft: No stats about %s on %s available\n"),
								group->name, serv->identifier);
							break;
					}
				} /* for (group_ix...) */
				if (all_unknown)
				{
					deblog_info(_("gft: %s has no stats available\n"),
							serv->identifier);
					/* if (serv->min_worth_getting)
						score = -2; */
				}
				if (all_groups_done)
				{
					deblog_info(_("gft: all groups in %s downloaded\n"),
							serv->identifier);
					score = -2;
				}
				check_uploads = TRUE;
				break;
			default:
				deblog_info(_("gft: %s's status %d unrecognised!\n"),
						serv->identifier, serv->status);
				break;

		} /* switch (serv->status) */
		if (check_uploads && serv->noutgoing > serv->nuploaded &&
				!serv->uploading)
		{
			deblog_info(_("%s has articles to upload, score +1000\n"),
					serv->identifier);
			score += 1000;
			if (*upload_first)
			{
				deblog_info(_("Uploads preferred, score +10000\n"));
				score += 10000;
			}
		}
		if (score >= 0)
		{
			score += serv->server_weight;
			/* Take number of threads into account */
			score += 1000 * (serv->max_threads - serv->nthreads);
			if (!serv->nthreads)
			{
				/* Give it a boost */
				score += 1000;
			}
		}
		deblog_info(_("gft: %s has scored %ld (best %ld)\n"),
				serv->identifier, score, best_score);
		if (score > best_score)
		{
			best_server = serv;
			best_score = score;
		}
	} /* for (n = 0; n < MAX_SERVERS; ++n) */
	if (best_server)
	{
		deblog_info(_("gft: %s has best score with %ld\n"),
				best_server->identifier, best_score);
		if (best_server->noutgoing <= best_server->nuploaded ||
				best_server->uploading)
		{
			*upload_first = FALSE;
		}
		deblog_info(_("Thread will %s first\n"), *upload_first ? _("upload") :
				_("download"));
		*serv_out = best_server;
		return server_GFT_OK;
	}
	deblog_info(all_done ? _("gft: All done\n") : _("All busy\n"));
	return all_done ? server_GFT_AllDone : server_GFT_AllBusy;
}

