#ifndef SERVER_H
#define SERVER_H

/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Data associated with a server */


#ifndef COMMON_H
#include "common.h"
#endif

#include <netinet/in.h>

#ifndef BACKFILL_H
#include "backfill.h"
#endif
#ifndef SCORE_H
#include "score.h"
#endif

typedef enum {
	server_Unresolved,
	server_Resolving,
	server_Resolved,
	server_Connecting,
	server_Connected,			/* Group info might not be valid yet */
	server_AllDone,
	server_Error				/* An error has occurred, don't use this server
								   again */
} server_Status;

struct ServerData {
	ServerData *next;
	int sidx;					/* Server index number, 0-9, used for
								   identification in interprocess comms */
	server_Status status;
	char *identifier;
	char *address;
	char *addr_s;
	int port;
	int proto;
	char *user;
	char *password;
	char *path;
	BOOL mode_reader;
	BOOL enter_group;
	BOOL pipeline;
	BOOL xover;					/* Use XOVER to get lists of Message-IDs (etc)
								   instead of XHDR */
	long int max_threads;
	long int server_weight;
	long int nthreads;			/* Number of threads running */
	long int next_group_stats;	/* Next group to be returned by
								   server_get_next_group_stats */
	long int ngroups;			/* Number of groups in use */
	long int group_size;		/* Total number of elements available in group,
								   not necessarily all occupied */
	GroupData **group;			/* Vector of newsgroups */
	BOOL feed;					/* Whether to check for outgoing articles */
	char *extra_feed;			/* Name of extra feed independent of feed
								   setting above */
	long int noutgoing;			/* Number of outgoing articles */
	long int outgoing_size;		/* Number of elements in outgoing vector */
	long int nuploaded;			/* Number of outgoings uploaded */
	char **outgoing;			/* Vector of outgoing article leaf names */
	long int total_to_download;	/* Number of articles remaining to download;
								   used to help decide whether to start new
								   thread */
	BOOL uploading;				/* Whether any threads are uploading to this
								   server */
	long int min_worth_getting;	/* Minimum number of articles it's worth
								   starting a new download thread for */
	long int default_score;
	long int kill_score;
	score_t *scores;
	long int backfill_expiry;
	BackFillServer *bfserv;
	long int xref_in_xover;		/* See command_FoundXrefInXover */
	BOOL xref_includes_name;
	BOOL xref_is_newsgroups;
	BOOL ignore_436;
	BOOL ssl;					/* Use SSL for this server */
	BOOL connx_msg;				/* Whether we've already printed a "Connected"
								   message in brief mode */
	BOOL locked_upload;			/* One or more uploads was locked when we tried
								   to post it, so we should try again just in
								   case it was rejected for the thread which
								   had the lock. */
	clock_t locked_at;			/* CPU time when upload was found locked */
};

void server_delete(ServerData *);

void server_mark_error(ServerData *);

void server_add_group(ServerData *, GroupData *);

/* Adds all groups from a server's newsrc file */
void server_read_newsrc(ServerData *);

/* Converse of above, writes newsrc file after fetch */
void server_write_newsrc(ServerData *);

/* Reads options from a server-specific cf file */
void server_read_cf(ServerData *);

/* Scans rc_dir to find all newsrc files if serv is NULL. If serv is non-NULL,
 * just processes newsrc file for serv */
void server_scan_rc_dir(const char *serv);

/* Returns most eligible server due to be serviced by a thread. On entry, if
 * *upload_first is true, result will be biased towards uploading. On exit
 * *upload_first indicates whether thread uploads first */
typedef enum {
	server_GFT_OK,				/* Connect to this server */
	server_GFT_AllBusy,			/* All currently busy, or with insufficient
								   info to be connected to, but not yet
								   finished */
	server_GFT_AllDone			/* All done, can exit now */
} server_GFT_Result;
server_GFT_Result server_get_for_threading(ServerData **, BOOL *upload_first);

/* Returns index of server's next group to get stats and Message-IDs for. If no
 * groups are available, returns -1. Any returned group will have status
 * updated to Allocated. */
long int server_get_next_group_stats(ServerData *);

/* Returns index of server's best group to enter and fetch articles from, based
 * on number of threads already fetching from it and then number of articles
 * available. group_ix should be -1 or last group the thread entered. If this
 * group has articles available, it's assumed to be the best. If no groups are
 * available (all AllDone or waiting for stats), returns -1. If stats are still
 * being fetched for some groups, *waiting is set to TRUE.
 */
long int server_get_best_group_articles(ServerData *, long int group_ix,
		BOOL *waiting);

#endif /* ! SERVER_H */
