/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include <limits.h>
#include <stdio.h>

#if USE_PTHREADS
#include <pthread.h>
#endif
#include "deblog.h"
#include "nscurses.h"
#include "stats.h"

static struct timeval stats_start_time;
static long int stats_current_dups = 0;
static long int stats_history_dups = 0;
static long int stats_total_bytes_rcvd = 0;
static long int stats_total_bytes_sent = 0;
static long int stats_total_seen = 0;

static long int stats_total_to_upload = 0;
static long int stats_uploaded = 0;
static long int stats_duplicates = 0;
static long int stats_bad_uploads = 0;
static long int stats_killed = 0;

static BOOL stats_quiet_flag = FALSE;

/* A node on a list for holding last few packets of info about bytes
 * transferred */
typedef struct stats_node {
	struct stats_node *next;
	struct timeval tv;
	long int bytes;
} stats_node;

static stats_node *stats_sent_head = NULL;
static stats_node *stats_rcvd_head = NULL;
static stats_node *stats_sent_tail = NULL;
static stats_node *stats_rcvd_tail = NULL;
static BOOL stats_rcvd_not_first = FALSE;
static BOOL stats_sent_not_first = FALSE;
static struct timeval stats_rcvd_last_time = {0, 0};
static struct timeval stats_sent_last_time = {0, 0};
static long int stats_rcvd_result = 0;
static long int stats_sent_result = 0;

void stats_start(void)
{
	struct timezone tz;

	if (gettimeofday(&stats_start_time, &tz))
		deblog_error(_("Can't read time\n"));
}

/* STATS_INTERVAL is interval between refreshes of info on screen in ms */
#ifdef STATS_INTERVAL
#define STATS_INTERVAL_TV \
	{ (STATS_INTERVAL) / 1000, ((STATS_INTERVAL) * 1000) % 1000000 }
#else
#define STATS_INTERVAL_TV { 0, 350000 }
#endif
struct timeval stats_interval = STATS_INTERVAL_TV; 

/* Get bytes per sec transferred over last 5 seconds */
static long int stats_get_bps(stats_node **head, stats_node **tail,
							  BOOL *not_first, struct timeval *last_time,
							  long int *result)
{
	/* Floating point because intermediate calculations can get out of range
	 * if we try to stick to ints (even long) */
	double total = 0.0;
	struct timeval difference;
	stats_node *node;
	struct timeval now;
	struct timeval since = { 0, 0 };
	struct timezone tz;
	static struct timeval period = { 5, 0 };

	gettimeofday(&now, &tz);
	xtimeval_subtract(&now, last_time, &difference);
	if (!*not_first || timercmp(&difference, &stats_interval, > ))
	{
		BOOL first_node = TRUE;
		*not_first = TRUE;
		*last_time = now;
		for (node = *head; node; )
		{
			stats_node *next = node->next;

			xtimeval_subtract(&now, &node->tv, &difference);
			/* If the difference between now and node's time is more than
			 * period, delete the node without including it in the calculation
			 */
			if (timercmp(&difference, &period, > ))
			{
				/* Items should only be removed from head of list unless we've
				 * gone drastically wrong */
				*head = node->next;
				if (*tail == node)
					*tail = NULL;
				free(node);
			}
			/* Otherwise if the node's time is within the period, include it in
			 * the calculation */
			else
			{
				/* If this is the first node to be included in the calculation,
				 * note its time */
				if (first_node)
				{
					first_node = FALSE;
					since = node->tv;
				}
				total += (double) node->bytes;
			}
			node = next;
		}
		xtimeval_subtract(&now, &since, &difference);
		/* Make sure that there were some nodes counted and time isn't too
		 * small to be measured */
		if (first_node || (difference.tv_sec == 0 && difference.tv_usec == 0))
		{
			*result = 0;
		}
		else
		{
			*result = (long int) (total / ((double) difference.tv_sec + (double)
						difference.tv_usec / 1E6));
		}
		/*
		fprintf(stderr, "ZZ total %f bytes counted in time %f, bps %ld\n",
				total,
				(double) difference.tv_sec + (double) difference.tv_usec / 1E6,
				*result);
		fflush(stderr);
		*/
				
	}
	return *result;
}

static void stats_queue_add(stats_node **head, stats_node **tail,
		const struct timeval *tv, long int bytes)
{
	stats_node *node = XMALLOC(sizeof(stats_node));

	node->next = NULL;
	node->tv = *tv;
	node->bytes = bytes;
	if (!*head)
		*head = node;
	if (*tail)
		(*tail)->next = node;
	*tail = node;
}

static long int stats_total_to_fetch = 0;
static long int stats_fetched = 0;
static long int stats_backfilled = 0;

static void stats_display(void)
{
	char sent[16];
	char rcvd[16];
	long sent_n;
	long rcvd_n;

	sent_n = stats_get_bps(&stats_sent_head, &stats_sent_tail,
			&stats_sent_not_first, &stats_sent_last_time, &stats_sent_result);
	rcvd_n = stats_get_bps(&stats_rcvd_head, &stats_rcvd_tail,
			&stats_rcvd_not_first, &stats_rcvd_last_time, &stats_rcvd_result);
	strcpy(rcvd, XNTOKILOMEG(rcvd_n));
	strcpy(sent, XNTOKILOMEG(sent_n));
	/*
	fprintf(stderr, "rcvd_n %ld rcvd %s sent_n %ld sent %s\n",
			rcvd_n, rcvd, sent_n, sent);
	fflush(stderr);
	*/
	if (nscurses_active())
	{
		nscurses_stats(stats_fetched, stats_total_to_fetch, stats_total_seen,
				rcvd, stats_uploaded, stats_total_to_upload, sent);
	}
	else if (!stats_quiet_flag)
	{
		stats_clear_line();
		printf(_("\r  << %ld/%ld @ %sB/s \t >> %ld/%ld @ %sB/s\t\t"),
				stats_fetched, stats_total_to_fetch, rcvd,
				stats_uploaded, stats_total_to_upload, sent);
		fflush(stdout);
		deblog_need_lf();
	}
}

void stats_clear_line(void)
{
	printf("\r%-74s\r", " ");
	fflush(stdout);
}

void stats_update(BOOL force)
{
	static BOOL first = TRUE;
	static struct timeval last_time = {0, 0};
	struct timeval now;
	struct timeval difference;
	struct timezone tz;
#if USE_PTHREADS
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	pthread_mutex_lock(&mutex);
#endif

	if (gettimeofday(&now, &tz))
	{
#if USE_PTHREADS
		pthread_mutex_unlock(&mutex);
#endif
		deblog_error(_("Can't read time\n"));
	}
	/* Subtract last_time from now (doesn't matter about last_time being
	 * meaningless if first) */
	xtimeval_subtract(&now, &last_time, &difference);
	if (force || first || timercmp(&difference, &stats_interval, > ))
	{
		first = FALSE;
		last_time = now;
		stats_display();
	}
#if USE_PTHREADS
	pthread_mutex_unlock(&mutex);
#endif
}

void stats_inc_total_to_fetch(void)
{
	++stats_total_to_fetch;
	stats_update(!maincf_get_regulate_updates());
}

void stats_inc_total_seen(long int by)
{
	stats_total_seen += by;
	stats_update(!maincf_get_regulate_updates());
}

void stats_dec_total_to_fetch(void)
{
	--stats_total_to_fetch;
	stats_update(!maincf_get_regulate_updates());
}

void stats_inc_fetched(void)
{
	++stats_fetched;
	stats_update(!maincf_get_regulate_updates());
}

void stats_inc_backfilled(void)
{
	++stats_backfilled;
}

void stats_inc_killed(void)
{
	++stats_killed;
}

void stats_update_bytes_rcvd(long int bytes, const struct timeval *tv)
{
	/*deblog_debug(_("%ld bytes rcvd since %.2f\n"), bytes, FLOAT_TV(tv));*/
	stats_queue_add(&stats_rcvd_head, &stats_rcvd_tail, tv, bytes);
	stats_total_bytes_rcvd += bytes;
	stats_update(FALSE);
}

void stats_update_bytes_sent(long int bytes, const struct timeval *tv)
{
	/*deblog_debug(_("%ld bytes sent since %.2f\n"), bytes, FLOAT_TV(tv));*/
	stats_queue_add(&stats_sent_head, &stats_sent_tail, tv, bytes);
	stats_total_bytes_sent += bytes;
	stats_update(FALSE);
}

void stats_inc_current_dups(void)
{
	++stats_current_dups;
}

void stats_inc_history_dups(void)
{
	++stats_history_dups;
}

void stats_inc_total_to_upload(void)
{
	++stats_total_to_upload;
}

void stats_dec_total_to_upload(void)
{
	--stats_total_to_upload;
	stats_update(!maincf_get_regulate_updates());
}

void stats_inc_uploaded(void)
{
	++stats_uploaded;
	stats_update(!maincf_get_regulate_updates());
}

void stats_inc_duplicates(void)
{
	++stats_duplicates;
}

void stats_inc_bad_uploads(void)
{
	++stats_bad_uploads;
}

void stats_summary(void)
{
	struct timeval now, difference;
	struct timezone tz;
	char sent[16];
	char rcvd[16];

	if (gettimeofday(&now, &tz))
		deblog_error(_("Can't read time\n"));
	xtimeval_subtract(&now, &stats_start_time, &difference);

	deblog_message(_("%ld %s downloaded, %ld killed\n"), stats_fetched,
			stats_fetched == 1 ? _("article") : _("articles"),
			stats_killed);
	deblog_message(_("%ld %s downloaded out of order\n"), stats_backfilled,
			stats_backfilled == 1 ? _("article") : _("articles"));
	deblog_message(_("%ld %s skipped, %ld already in history\n"),
			stats_current_dups,
			stats_current_dups == 1 ? _("duplicate") : _("duplicates"),
			stats_history_dups);
	deblog_message(_("%ld %s uploaded\n"), stats_uploaded,
			stats_uploaded == 1 ? _("article") : _("articles"));
	deblog_message(
			_("%ld %s rejected as %s, %ld for other reasons\n"),
			stats_duplicates,
			stats_duplicates == 1 ? _("upload") : _("uploads"),
			stats_duplicates == 1 ? _("a duplicate") : _("duplicates"),
			stats_bad_uploads);
	strcpy(rcvd, XNTOKILOMEG(stats_total_bytes_rcvd));
	strcpy(sent, XNTOKILOMEG(stats_total_bytes_sent));
	deblog_message(_("%sB received, %sB sent in %.2f secs\n"), rcvd, sent,
			FLOAT_TV(&difference));
	strcpy(rcvd, XNTOKILOMEG((long int) ((double) stats_total_bytes_rcvd /
					((double) difference.tv_sec + (double) difference.tv_usec /
					 1E6))));
	strcpy(sent, XNTOKILOMEG((long int) ((double) stats_total_bytes_sent /
					((double) difference.tv_sec + (double) difference.tv_usec /
					 1E6))));
	deblog_message(_("(%sB/s, %sB/s)\n"), rcvd, sent);
}

void stats_quiet(void)
{
	stats_quiet_flag = TRUE;
}

