/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

#include <sys/stat.h>
#include <fcntl.h>


#include "uploads.h"
#include "vector.h"

typedef struct {
	char *pathname;
	char *server;	/* NULL if good */
	char *reason;	/* NULL if good */
} UploadsStatus;

static struct VECTOR(UploadsStatus *) uploads_vector;

static BOOL uploads_initialised = FALSE;

/* Safe to call this repeatedly */
static void uploads_init(void)
{
	if (!uploads_initialised)
	{
		VECTOR_INIT(uploads_vector, UploadsStatus *);
		uploads_initialised = TRUE;
	}
}

static UploadsStatus *uploads_lookup(const char *pathname)
{
	size_t i;

	uploads_init();
	for (i = 0; i < uploads_vector.vec_used_members; ++i)
	{
		UploadsStatus *ustat = uploads_vector.vec_data[i];

		if (!strcmp(pathname, ustat->pathname))
			return ustat;
	}
	return NULL;
}

static void uploads_add(const char *pathname,
		const char *server, const char *reason)
{
	UploadsStatus *ustat = XMALLOC(sizeof(UploadsStatus));

	ustat->pathname = XSTRDUP(pathname);
	ustat->server = XSTRDUP(server);
	ustat->reason = XSTRDUP(reason);
	VECTOR_APPEND(uploads_vector, ustat);
}

void uploads_bad(const char *pathname, const char *server, const char *reason)
{
	UploadsStatus *ustat = uploads_lookup(pathname);

	if (ustat)
	{
		/* Keep old status whatever it was, ignore new failure */
		return;
	}
	uploads_add(pathname, server, reason);
}

void uploads_good(const char *pathname)
{
	UploadsStatus *ustat = uploads_lookup(pathname);

	if (ustat)
	{
		if (ustat->server)
		{
			free(ustat->server);
			ustat->server = NULL;
		}
		if (ustat->reason)
		{
			free(ustat->reason);
			ustat->reason = NULL;
		}
	}
	else
	{
		uploads_add(pathname, NULL, NULL);
	}
}

static void uploads_move_one_bad(const char *pathname,
		const char *server, const char *reason)
{
	char *badname;
	char *baddir = XSTRDUP(pathname);
	int infd = -1;
	int outfd = -1;
	rw_buf *inbuf = NULL;
	rw_buf *outbuf = NULL;
	BOOL result = FALSE;
		
	badname = strrchr(baddir, '/');
	*badname = 0;
	badname = strrchr(baddir, '/');
	*badname = 0;
	badname = baddir;
	baddir = XSTRDUPCAT(badname, "/failed");
	free(badname);
	deblog_info(_("Attempting to move bad upload %s to %s\n"), pathname,
			baddir);
	outfd = XMKTMPINDIR(baddir, _("Bad"), &badname);
	deblog_debug(_("badname %s file handle %d\n"), badname, outfd);
	infd = open(pathname, O_RDONLY);
	if (infd != -1)
		inbuf = rw_create_buf(infd, FALSE, FALSE, -1);
	/*if (inbuf)
		outfd = open(badname, O_WRONLY | O_CREAT | O_EXCL, 0644);*/
	if (inbuf && outfd != -1)
		outbuf = rw_create_buf(outfd, FALSE, FALSE, -1);
	if (outbuf)
	{
		result = rw_writef(outbuf,
				_("X-Rejected-Server: %s\nX-Rejected-Reason: %s\n"),
				server, reason) > 0;
	}
	if (result)
	{
		result = !rw_copy(inbuf, outbuf, NULL, NULL);
	}
	if (result)
	{
		rw_close_and_free(inbuf);
		rw_close_and_free(outbuf);
		XREMOVE(pathname);
	}
	if (!result)
	{
		deblog_critical(_("Couldn't move problem article: %s\n"),
				strerror(errno));
		deblog_error(_("Bad article stuck at %s\n"), pathname);
	}
	deblog_warning(_("Bad article moved to %s\n"), badname);
	free(badname);
}

void uploads_move_any_bad(void)
{
	size_t i;

	if (!uploads_initialised)
		return;

	for (i = 0; i < uploads_vector.vec_used_members; ++i)
	{
		UploadsStatus *ustat = uploads_vector.vec_data[i];

		if (ustat->reason)
			uploads_move_one_bad(ustat->pathname, ustat->server, ustat->reason);
	}
}

